/* 
 * G E M
 * Copyright (C) 2002 Matej Knopp <knopp@users.sf.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __gem_uta_h__
#define __gem_uta_h__

#ifdef __cplusplus
extern "C" {
#endif

#include <SDL/SDL.h>

typedef struct _GemUta GemUta;
typedef struct _GemRect GemRect;

typedef Uint32 GemUtaBBox;

#define GEM_UTA_BBOX_CONS(x0, y0, x1, y1)  (((x0) << 24) | ((y0) << 16) | \
                                            ((x1) << 8) | (y1))

#define GEM_UTILE_EMPTY   0x00000000
#define GEM_UTILE_FULL    0x01012020

#define GEM_UTA_BBOX_X0(ub) ((ub) >> 24)
#define GEM_UTA_BBOX_Y0(ub) (((ub) >> 16) & 0xff)
#define GEM_UTA_BBOX_X1(ub) (((ub) >> 8) & 0xff)
#define GEM_UTA_BBOX_Y1(ub) ((ub) & 0xff)

#define GEM_UTILE_SHIFT     5
#define GEM_UTILE_SIZE      (1 << GEM_UTILE_SHIFT)

#define GEM_MIN(a,b) ((a) < (b) ? (a) : (b))
#define GEM_MAX(a,b) ((a) > (b) ? (a) : (b))

struct _GemUta
{
	int x;
	int y;
	int width;    /* number of tiles in one line */
	int height;   /* number of tiles in one row */
	GemUtaBBox *tiles;
};

struct _GemRect
{
	int x0;
	int x1;
	int y0;
	int y1;
};

GemUta    * gem_uta_new        (int x, int y, int width, int height);

GemUta    * gem_uta_new_coords (int x, int y, int width, int height);

void        gem_uta_free       (GemUta * uta);

void        gem_uta_union      (GemUta * uta1, const GemUta * uta2);

void        gem_uta_copy      (GemUta * uta1, const GemUta * uta2);

void        gem_uta_add_rect   (GemUta * uta, const GemRect * rect);

#define GEM_UTA_QUERY_CONTAIN		0
#define GEM_UTA_QUERY_INTERSECT		1

int         gem_uta_query_rect (const GemUta * uta, 
                                const GemRect * rect, int query);

SDL_Rect  * gem_uta_get_rects  (const GemUta * uta, int * num_rects, 
                                int min_x, int min_y,
                                int max_x, int max_y);

void        gem_uta_clear      (GemUta * uta);

/******<CUT HERE>***********************************************************/
/*
 * Usage:
 * The usage of gem-uta is pretty simple. You first need to allocate 
 * your uta with gem_uta_new or gem_uta_new_coords. The difference 
 * between gem_uta_new and gem_uta_new_coords is that the first takes
 * the size and position as number of tiles while the second takes 
 * considers the given coordinates pixel values.
 * e.g: (if you have screen with resolution 640 * 480)
 *
 *   GemUta * uta = gem_uta_new_coords (0, 0, 640, 480); 
 *
 * Then you should add some rectangles with gem_uta_add_rect. You may
 * need to do a simple conversion between SDL_Rect and GemRect, e.g:
 *
 *   SDL_Rect rect; / * some rectangle * /
 *   GemRect grect;
 *   grect0.x = rect.x; grect.y0 = rect.y;
 *   grect.x1 = rect.x + rect.w - 1; grect.y1 = rect.y + rect.h - 1;
 *  
 * and then add the grect to the uta:
 * 
 *   gem_uta_add_rect (uta, &grect);
 *
 * After you'd added all the rectangles you wanted you can retrieve
 * an array of SDL_Rect's that convers the same or little bigger
 * area than the rectangles you've added. (Without overlapping, of 
 * course):
 *
 *  int num_rects;
 *  SDL_Rect *rects = gem_uta_get_rects (uta, &num_rects, 0, 0, 640, 480);
 *  SDL_UpdateRects (display, num_rects, rects);
 *  free (rects); / * Don't forget to free the rectangles * /
 * 
 * After you'd finished the frame you don't need to free the uta, just
 * clear it and you can continue adding and getting rectangles:
 *
 *  gem_uta_clear (uta);
 *
 */

#ifdef __cplusplus
} /* extern "C" { */
#endif


#endif /* __gem_uta_h__ */
