/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "canvas.h"
#include "graphic.h"

#include "editorcontrol.h"
#include "serializer.h"
#include "bodyfactory.h"
#include "cworld.h"
#include "starship.h"
#include "functionexts.h"
#include "pathdrawer.h"
#include "debugmsg.h"

#include <iostream>
#include <functional>

Canvas::Canvas(EditorControl * ec): angle(0),selected(NULL), level(new cWorld()), zc(level), ec(ec), orbit(0) {
    level->paused=false;
    addMouseListener(this);
    mode=EDIT;
    angle=0;
    draggingObject=false;
    mDragged = false;
    mousePressEvent = false;
    velocityDragged=NULL;
    play=false;
    last_step=0;
}

inline void Canvas::drawOrbit(pair<GameBody*,GameBody*> p) {
    Vector2d center=view_transformation*p.first->getPosition();
    Graphic::drawCircle(center,int((center-view_transformation*p.second->getPosition()).dist()),
                        SDL_MapRGB(Graphic::getScreen()->format, 200, 0,0));
}

inline void updateOrbit(GameBody* orbitCenter,GameBody* orbiting) {
    Vector2d radius=orbiting->getPosition()-orbitCenter->getPosition();
    real r=radius.dist();
    Vector2d vel(-radius.y,radius.x);
    orbiting->setVelocity(vel*sqrt(PhysicEngine::G*orbitCenter->mass/r)/r+orbitCenter->getVelocity());
}

void Canvas::updateOrbitPair(pair<GameBody*,GameBody*> p) {
    updateOrbit(p.first,p.second);
    updateOrbits(p.second);
}

void Canvas::updateOrbits(GameBody* c) {
    pair<multimap<GameBody*,GameBody*>::iterator,multimap<GameBody*,GameBody*>::iterator> p
    = orbits.equal_range(c);
    for_each(p.first,p.second,bind1st(mem_fun(&Canvas::updateOrbitPair),this));
}


bool Canvas::removeThisOrbits(GameBody* start,GameBody* c) {
    pair<multimap<GameBody*,GameBody*>::iterator,multimap<GameBody*,GameBody*>::iterator> 
        p = orbits.equal_range(start);
    for (multimap<GameBody*,GameBody*>::iterator i=p.first;i!=p.second;i++) {
        if (i->second==c) {
            orbits.erase(i);
            return true;
        } else if (removeThisOrbits(i->second,c)) return true;
    }
    return false;
}

void Canvas::setMass(GameBody * b,real m) {
    Starship* ship = dynamic_cast<Starship*>(b);
    if (ship) {
        if (ship->getFuel()>m)
            ship->setFuelNoMass(m);
    }

    b->mass=m;
    level->refitBody(b);
    updateOrbits(b);

    level->retrace();
    level->step(0);
}


void Canvas::initTime() {
    last_step=Graphic::ticks();
    level->set(0,EmptyEvent());
}

void Canvas::refreshViewTransformation() {
    if (level->getTrackCoordinateSystem())
        view_transformation=level->getTrackCoordinateSystem()->get_view_transformation(false,level->getZoom(),level->getEyeShift());
    else
        view_transformation=Matrix2x3(1.0/level->getZoom(),1.0/level->getZoom(),0,-Graphic::screencenter.x,
                                      -Graphic::screencenter.y,level->getEyeShift().x,level->getEyeShift().y);
}

void drawVelocity(const Matrix2x3 & view_transformation,real zoom,GameBody*gb) {
    Vector2d o=view_transformation*gb->getPosition();
    //  Vector2d d=view_transformation*(
    //            (gb->getPosition()+gb->getVelocity()*1000)
    //           );
    Vector2d d=o+gb->getVelocity()*500;
    Graphic::drawArrow(int(o.x),int(o.y),int(d.x),int(d.y),
                       SDL_MapRGB(Graphic::getScreen()->format, 200,0,0));
    Graphic::drawCircle(d,int(2*zoom),SDL_MapRGB(Graphic::getScreen()->format, 255,255,128));
}

inline int max(int x,int y) {
    return x>y?x:y;
}

void Canvas::draw(gcn::Graphics *) {
    SDL_FillRect(Graphic::getScreen(),NULL,SDL_MapRGB(Graphic::getScreen()->format, 0, 0,0));

    // Screen scrolling
    Uint16 delta=Graphic::ticks()-last_step;
    last_step+=delta;
    Uint8 appstate=SDL_GetAppState();
    if ((appstate & SDL_APPMOUSEFOCUS) && (appstate & SDL_APPACTIVE) && (appstate & SDL_APPINPUTFOCUS))
        mscroll.update(delta,scrollDir);
    else mscroll.lower(delta);
    if (mscroll.scrollvel!=Vector2d()) {
        level->setEyeShift(level->getEyeShift()+delta*mscroll.scrollvel/level->getZoom());
        refreshViewTransformation();
    } else if (zc.state!=ZoomControl::Z_NONE) {
        level->zoomStep(delta);
        refreshViewTransformation();
    }
    level->draw();

    if (selected) {
        Vector2d center=view_transformation*selected->getPosition();
        Graphic::drawCircle(center,max(Uint16(selected->getWidth()/1.5*level->getZoom()),10),
                            SDL_MapRGB(Graphic::getScreen()->format, 255, 255, 0));
    }

    for_each(orbits.begin(),orbits.end(),bind1st(mem_fun(&Canvas::drawOrbit),this));

    if (orbit) {
        Vector2d center=view_transformation*orbitCenter->getPosition();
        Graphic::drawCircle(center,Uint16((center-orbiting).dist()),SDL_MapRGB(Graphic::getScreen()->format, 200, 0,0));
    }

    cWorld::iterator end=level->objectsEnd();
    for (cWorld::iterator i=level->objectsBegin();i!=end;i++) {
        drawVelocity(view_transformation,level->getZoom(),*i);
    }
};


inline bool sndIs(GameBody* b,pair<GameBody*,GameBody*> p) {
    return b==p.second;
}

void Canvas::mouseClickDeleteMode (int x, int y, int, int) {
    if (!level->getTrackCoordinateSystem()) return;

    gcn::Rectangle gcnr=getDimension();
    Vector2d c;

    // Erase bodies in the background.

    real f=(level->getZoom()*7+1)/level->getZoom();
    Matrix2x3 view(f,f,-angle, -GameBody::screencenter.x+gcnr.x,
                   -GameBody::screencenter.y+gcnr.y,
                   level->getTrackCoordinateSystem()->getPositionX()+level->getEyeShift().x,
                   level->getTrackCoordinateSystem()->getPositionY()+level->getEyeShift().y);
    c=view*Vector2d(x,y);
    SDL_Rect r={Sint16(c.x-5*f),Sint16(c.y-5*f),Sint16(10*f),Sint16(10*f)};

    level->bg.objects.remove_if(bind2nd(mem_fun(&GameBody::touches),&r));

    // Erase bodies in the foreground.

    c=level->getTrackCoordinateSystem()->get_view2world_transformation(false,level->getZoom(),level->getEyeShift())*Vector2d(x+gcnr.x,y+gcnr.y);

    GameBody * hitted;
    bool hit=level->hittedBodies(Vector2d(x+gcnr.x,y+gcnr.y),&hitted,1);
    if (hit) {
        orbits.erase(hitted);
        multimap<GameBody*,GameBody*>::iterator i=orbits.begin();
        while (i!=orbits.end()) {
            if (i->second==hitted) {
                multimap<GameBody*,GameBody*>::iterator j=i++;
                orbits.erase(j);
            } else i++;
        }
        if (selected==hitted)
            selected=NULL;
        if (hitted==level->getTrackCoordinateSystem()) {
            // set new origin
            cWorld::iterator end=level->objectsEnd();
            cWorld::iterator i=level->objectsBegin();
            if (i!=end) {
                if (*i==hitted)
                    i++;
                if (i!=end)
                    level->setTrackCoordinateSystem(*i);
            }
            level->deleteBody(hitted);

            refreshViewTransformation();
        } else
            level->deleteBody(hitted);
    }
}

bool existOrbit(const multimap<GameBody*,GameBody*> & orbits,GameBody*orbitCenter,GameBody*orbiting) {
    pair<multimap<GameBody*,GameBody*>::const_iterator,multimap<GameBody*,GameBody*>::const_iterator> p
    = orbits.equal_range(orbitCenter);
    for (multimap<GameBody*,GameBody*>::const_iterator i=p.first;i!=p.second;i++)
        if ((*i).second==orbiting) return true;
    return false;
}

void Canvas::mouseClickAddMode (int x, int y, int button, int) {
    gcn::Rectangle gcnr=getDimension();
    Vector2d c;

    // Cancel everything if user pressed left button.
    if (button==gcn::MouseInput::RIGHT && orbit) {
        orbit=false;
        return;
    }

    // The user is trying to do something.
    // It is trying to add an object to the background?
    if (selectedType.find("farstar")!=string::npos || selectedType.find("nebula")!=string::npos
            || selectedType.find("galaxia")!=string::npos) {
        real f=(level->getZoom()*7+1)/level->getZoom();
        Matrix2x3 view;
        if (level->getTrackCoordinateSystem())
            view=Matrix2x3(f,f,-angle, -GameBody::screencenter.x+gcnr.x,
                           -GameBody::screencenter.y+gcnr.y,
                           level->getTrackCoordinateSystem()->getPositionX()+level->getEyeShift().x,
                           level->getTrackCoordinateSystem()->getPositionY()+level->getEyeShift().y);
        else
            view=Matrix2x3(f,f,-angle, -GameBody::screencenter.x+gcnr.x,
                           -GameBody::screencenter.y+gcnr.y,level->getEyeShift().x,level->getEyeShift().y);
        //                               originBody->getPositionX(),originBody->getPositionY());
        c=view*Vector2d(x,y);
        if (selectedType.find("farstargroup")!=string::npos) {
            list<GameBody*> gbl=starlist(selectedType,int(c.x),int(c.y),40);
            for (list<GameBody*>::iterator i=gbl.begin();i!=gbl.end();i++)
                level->bg.objects.push_front(*i);

        } else level->bg.objects.push_front(BodyFactory::instance()->create(selectedType.c_str(),c));
    } else {
        // Well we are trying to put a planet or something else which can orbit.

        bool haseye=level->eye;
        if (haseye) {
            c=level->getTrackCoordinateSystem()->get_view2world_transformation(false,level->getZoom(),level->getEyeShift())*Vector2d(x+gcnr.x,y+gcnr.y);
        } else
            c=Matrix2x3(1.0/level->getZoom(),1.0/level->getZoom(),0,-Graphic::screencenter.x,
                        -Graphic::screencenter.y,level->getEyeShift().x,level->getEyeShift().y)*Vector2d(x+gcnr.x,y+gcnr.y);

        GameBody * hitted;
        bool hit=level->hittedBodies(Vector2d(x+gcnr.x,y+gcnr.y),&hitted,1);

        // If we are placing an object around another we set the orbit.
        if (orbit && hit && orbitCenter != hitted) {
            removeThisOrbits(hitted,orbitCenter);
            if (!existOrbit(orbits,orbitCenter,hitted)) {
                orbits.insert(pair<GameBody*,GameBody*>(orbitCenter,hitted));
                updateOrbit(orbitCenter,hitted);
                updateOrbits(hitted);
            }
            orbit=false;
            selected=hitted;
        } else if (orbit) {
            orbit=false;
        } else if (hit) {
            orbitCenter=hitted;
            orbiting=Vector2d(x+gcnr.x,y+gcnr.y);
            orbit=true;
            selected = hitted;
        } else if (selectedType == "player1") { // Add a new object.
            if (level->getShip()) {
                GameBody * g=level->getShip();
                GameBody * g2=BodyFactory::instance()->create("starship1",c);
                level->addBody(g2);
                level->setShip(dynamic_cast<Starship*>(g2));
                level->other=g2;
                orbits.erase(g);
                multimap<GameBody*,GameBody*>::iterator i=orbits.begin();
                while (i!=orbits.end()) {
                    if (i->second==g) {
                        multimap<GameBody*,GameBody*>::iterator j=i++;
                        orbits.erase(j);
                    } else i++;
                }
                level->deleteBody(g);
                delete g;
            } else {
                level->setShip(dynamic_cast<Starship*>(BodyFactory::instance()->create("starship1",c)));
                level->addBody(level->getShip());
            }
            if (!haseye)
                refreshViewTransformation();
            selected = level->getShip();
        } else {
            GameBody * g=BodyFactory::instance()->create(selectedType.c_str(),c);
            level->addBody(g);
            if (!haseye)
                refreshViewTransformation();
            selected = g;
        }
        if (selected) {
            selectedIsOriginConected=isOriginConected(selected);
            ec->ctlEdit->update(selected);
        }
    }
}


void Canvas::keyPressed (gcn::KeyEvent& e) {
    gcn::Key key = e.getKey();
    if (!level->getTrackCoordinateSystem()) return;

    // do nothing if the focus is in some text component
    if (_getFocusHandler()!=NULL && _getFocusHandler()->getFocused()!=NULL
            && (dynamic_cast<gcn::TextBox*>(_getFocusHandler()->getFocused())!=NULL
                || dynamic_cast<gcn::TextField*>(_getFocusHandler()->getFocused())!=NULL)) {
        return;
    }

    Vector2d center;
    real zoom=level->getZoom();

    switch (key.getValue()) {
    case gcn::Key::UP:
        scrollDir.y = -1;
        break;
    case gcn::Key::DOWN:
        scrollDir.y = 1;
        break;
    case gcn::Key::RIGHT:
        scrollDir.x = 1;
        break;
    case gcn::Key::LEFT:
        scrollDir.x = -1;
        break;
    case '+':
		level->zoomEye.x = LOGICAL_WIDTH/2;
		level->zoomEye.y = LOGICAL_HEIGHT/2;
        zc.pressIn();
        break;
    case '-':
        zc.pressOut();
        break;
    case '*':
        if (play)
            level->zoomTimev=100.0/99.0;
        break;
    default:
        ;
    }

    level->setEyeShift(level->getEyeShift()+center/zoom);
    refreshViewTransformation();
}

void Canvas::keyReleased (gcn::KeyEvent& e) {
    gcn::Key key = e.getKey();
    if (!level->getTrackCoordinateSystem()) return;

    switch (key.getValue()) {
    case gcn::Key::UP:
    case gcn::Key::DOWN:
        scrollDir.y = 0;
        break;
    case gcn::Key::LEFT:
    case gcn::Key::RIGHT:
        scrollDir.x = 0;
        break;
    default:
        zc.release();
    }
}


bool Canvas::isOriginConected(GameBody* c) {
    if (level->getTrackCoordinateSystem()->id==c->id) return true;

    pair<multimap<GameBody*,GameBody*>::iterator,multimap<GameBody*,GameBody*>::iterator> p
    = orbits.equal_range(c);
    for (multimap<GameBody*,GameBody*>::iterator i=p.first;i!=p.second;i++)
        if (isOriginConected((*i).second)) return true;

    return false;
}

void Canvas::mousePressed(gcn::MouseEvent& e) {
    int x;
    int y;
    toAbsolutePos(e.getX(),e.getY(),&x,&y);
    int button = e.getButton();

    mousePressEvent = true;

    if (mode==DELETE)
        mouseClickDeleteMode(x,y,button,1);

    if (button==gcn::MouseInput::RIGHT) {
        if (orbit) // cancel orbit creation
            orbit=false;
        else {
            gcn::Rectangle gcnr=getDimension();

            GameBody* g;
            bool hit=level->hittedBodies(Vector2d(x+gcnr.x,y+gcnr.y),&g,1);
            if (hit) {
                level->setTrackCoordinateSystem(g);
                level->step(0);
            }
        }
    } else if (mode==EDIT && button==gcn::MouseInput::LEFT && level->getTrackCoordinateSystem()) {
        gcn::Rectangle gcnr=getDimension();

        cWorld::iterator i=level->objectsBegin();
        for (;i!=level->objectsEnd();i++) {
            Vector2d v=view_transformation*(*i)->getPosition()+(*i)->getVelocity()*500;
            v.x-=x+gcnr.x;
            v.y-=y+gcnr.y;
            if (v.sqrDist()<6) 
                break;
        }
        if (i!=level->objectsEnd()) {
            velocityDragged=*i;
            selectedIsOriginConected=isOriginConected(velocityDragged);
        } else {
            GameBody* g;
            bool hit=level->hittedBodies(Vector2d(x+gcnr.x,y+gcnr.y),&g,1);
            if (hit) {
                draggingObject=true;
                selected=g;
                selectedIsOriginConected=isOriginConected(g);
                ec->ctlEdit->update(g);
            }
        }
    }
}

void Canvas::toAbsolutePos(int rx,int ry,int* x,int* y) {
    getAbsolutePosition(*x,*y);
    *x = Graphic::fromScreenx(*x+rx);
    *y = Graphic::fromScreeny(*y+ry);

}

void Canvas::mouseReleased(gcn::MouseEvent& e) {
    int x;
    int y;
    toAbsolutePos(e.getX(),e.getY(),&x,&y);
    int button = e.getButton();

    if (!mousePressEvent)
        return;
    mousePressEvent = false;
    if (!(velocityDragged && mDragged) && !mDragged && mode!=DELETE && button==gcn::MouseInput::LEFT)
        mouseClickAddMode(x,y,button,1);

    if (draggingObject && mode==EDIT && button==gcn::MouseInput::LEFT) {
        for (multimap<GameBody*,GameBody*>::iterator i=orbits.begin();i!=orbits.end();i++) {
            if (i->second==selected)
                updateOrbitPair(*i);
        }

        updateOrbits(selected);
        draggingObject=false;

        level->retrace();
        level->step(0);
    } else if (velocityDragged && mDragged) {
        multimap<GameBody*,GameBody*>::iterator i=orbits.begin();
        while (i!=orbits.end()) {
            if (i->second==velocityDragged) {
                multimap<GameBody*,GameBody*>::iterator j=i++;
                orbits.erase(j);
            } else i++;
        }
        updateOrbits(velocityDragged);

        level->retrace();
        level->step(0);

    }

    velocityDragged=NULL;
    mDragged = false;
}


void Canvas::updateOrbitPos(GameBody* c,Vector2d * d) {
    c->setPosition(c->getPosition()+*d);
    pair<multimap<GameBody*,GameBody*>::iterator,multimap<GameBody*,GameBody*>::iterator> 
        p = orbits.equal_range(c);
    for (multimap<GameBody*,GameBody*>::iterator i=p.first;i!=p.second;i++)
        updateOrbitPos((*i).second,d);
}

void Canvas::mouseDragged(gcn::MouseEvent& e) {
    mouseMoved(e);
}

void Canvas::mouseMoved (gcn::MouseEvent& e) {
    int x;
    int y;
    toAbsolutePos(e.getX(),e.getY(),&x,&y);
    if (!level->getTrackCoordinateSystem()) 
        return;

    if (orbit) {
        orbiting.x=x+getX();
        orbiting.y=y+getY();
    } else if (mode==EDIT && draggingObject) {
        if (play) 
            ec->action("play");

        mDragged = true;

        gcn::Rectangle gcnr=getDimension();
        Vector2d c=level->getTrackCoordinateSystem()->get_view2world_transformation(false,level->getZoom(),level->getEyeShift())*
                   Vector2d(x+gcnr.x,y+gcnr.y)-selected->getPosition();

        if (selectedIsOriginConected)
            level->setEyeShift(level->getEyeShift()-c);

        updateOrbitPos(selected,&c);
        refreshViewTransformation();

        if (selectedIsOriginConected || (level->getShip() && selected->id==level->getShip()->id)) {
            level->retrace();
            level->step(0);
        }
    } else if (velocityDragged) {
        mDragged = true;

        gcn::Rectangle gcnr=getDimension();
        Vector2d c=Vector2d(x+gcnr.x,y+gcnr.y)-view_transformation*velocityDragged->getPosition();

        velocityDragged->setVelocity(c/500);
    }
}

void Canvas::setDimension(const gcn::Rectangle &gcnr) {
    if (level->getTrackCoordinateSystem())
        refreshViewTransformation();
    gcn::Widget::setDimension(gcnr);
}

void Canvas::clear() {
    draggingObject=false;
    selected=NULL;
    orbit=0;
    orbiting=0;
    orbitCenter=NULL;
    orbits.clear();
    level->clear();
    velocityDragged=NULL;
    view_transformation=Matrix2x3();
}

ostream & operator << (ostream & o,const Canvas & c) {
    o<<*c.level;
    for (cWorld::iterator i=c.level->objectsBegin();i!=c.level->objectsEnd();i++) {
        pair<multimap<GameBody*,GameBody*>::const_iterator,
        multimap<GameBody*,GameBody*>::const_iterator> p
        = c.orbits.equal_range(*i);
        o<<distance(p.first,p.second)<<'\n';
        for (multimap<GameBody*,GameBody*>::const_iterator i=p.first;i!=p.second;i++)
            o<<indexOf(c.level->objectsBegin(),c.level->objectsEnd(),(*i).second)<<'\n';
    }
    return o;
}

istream & operator >> (istream & is,Canvas & c) {
    c.orbits.clear();
    StreamTokenizer st(is);
    Lexer p(&st);
    PARSEbegin(Lexer,p)
    parseWorld(p,*c.level);

    for (cWorld::iterator i=c.level->objectsBegin();i!=c.level->objectsEnd();i++) {
        int count=0;
        LINT(&count);
        GameBody * orbitCenter=*i;
        for (int j=0;j<count;j++) {
            int orbitIndex;
            LINT(&orbitIndex);
            c.orbits.insert(pair<GameBody*,GameBody*>(orbitCenter,
                            c.level->getBody(orbitIndex)));
        }
        ENDBLOCK;
    }
    ENDBLOCK;
    PARSEend;

    if (p.error()) {
        cerr<<p.errorMessage()<<'\n';
        return is;
    }

    c.level->step(0);
    if (c.level->getTrackCoordinateSystem())
        c.refreshViewTransformation();
    else c.view_transformation=Matrix2x3();
    return is;
}

Canvas::~ Canvas() {
    delete level;
}
