/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CANVAS_H
#define CANVAS_H

#include <guichan.hpp>
#include <string>
#include "vector2d.h"
#include "global.h"
#include "mousescroll.h"
#include "zoomcontrol.h"

using namespace std;
class EditorControl;
class cWorld;
class GameBody;

/**
 * This class is a widget capable of reacting to user input to build a level.
 * It needs to have its mode set in order to know what the user it is trying to do.
 */
class Canvas : public gcn::Widget,public gcn::MouseListener,
            public gcn::KeyListener  {
public:

    /**
     * Whether canvas should animate the world.
     */
    bool play;

    /**
     * The zoom factor of the view.
     */
    //real zoom;

    /**
     * The angle at which the world is seen.
     */
    real angle;

    /**
     * The selected object type to work with.
     */
    string selectedType;

    /**
     * The current editing mode.
     */
    enum { DELETE, EDIT } mode;

    /**
     * This is a selected object. The value is nothing if there is
     * no current selection.
     */
    GameBody * selected;

    /**
     * Constructs a canvas.
     * @param l The level to work with.
     */
    Canvas(EditorControl * ec);

    /**
     * Destroys a canvas.
     */
    ~Canvas();

    /**
      * Starts counting the world time for the animation.
      */
    void initTime();

    /**
     * Sets the mass of the given object. It updates all asociated orbits.
     * @param b The body which mass is changed.
     * @param m The new value for the mass.
     */
    void setMass(GameBody * b,real m);

    /**
     * Draws gui elements associated to the editor.
     * @param g The graphics to use when drawing this widget.
     */
    void draw(gcn::Graphics * g);

    /**
     * Method given by GUICHAN for capturing mouse clicks.
     */
    void mousePressed (gcn::MouseEvent& e);

    /**
     * Method given by GUICHAN for capturing mouse clicks.
     */
    void mouseReleased (gcn::MouseEvent& e);

    /**
     * Method given by GUICHAN for capturing mouse movement.
     */
    void mouseMoved (gcn::MouseEvent& e);
    void mouseDragged(gcn::MouseEvent& e);

    /**
     * Method given by GUICHAN for capturing key presses.
     */
    void keyPressed (gcn::KeyEvent& e);

    /**
     * Method given by GUICHAN for capturing key releases.
     */
    void keyReleased (gcn::KeyEvent& e);

    /**
     * We capture the event of setting dimension for initializing the
     * view transformation.
     */
    void setDimension(const gcn::Rectangle &gcnr);

    void refreshViewTransformation();


    /**
     * Updates the orbits of every body around c to make them circular.
     * @param c
     */
    void updateOrbits(GameBody* c);
    bool removeThisOrbits(GameBody* start,GameBody* c);

    void clear();

    cWorld * level;

    ZoomControl zc;

private:
    EditorControl * ec;
    GameBody * orbitCenter;
    Vector2d orbiting;
    int orbit;
    int last_step;
    Vector2d scrollDir;
    std::multimap<GameBody *,GameBody *> orbits;
    void drawOrbit(pair<GameBody*,GameBody*> p);
    Matrix2x3 view_transformation;
    void mouseClickAddMode (int x, int y, int button, int count);
    void mouseClickDeleteMode (int x, int y, int button, int count);
    void updateOrbitPair(pair<GameBody*,GameBody*> p);
    void updateOrbitPos(GameBody* c,Vector2d * d);
    bool isOriginConected(GameBody* c);
    void toAbsolutePos(int rx,int ry,int* x,int* y);

    bool mDragged, draggingObject,selectedIsOriginConected, mousePressEvent;
    GameBody * velocityDragged;
    MouseScroll mscroll;

    friend istream & operator >> (istream & is,Canvas & c);
    friend ostream & operator << (ostream & o,const Canvas & c);
};

#endif // CANVAS_H
