/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CMASK_H
#define CMASK_H

#include "resourcemanager.h"
#include "parsercombinators.h"
#include <list>
#include "outline.h"

/** 
 * Implementation of a set of non-overlapped convex polygons which define
 * the outline of an object.
 *
 * Each polygon has a minimum enclosing disk which is used to speed the 
 * collision checks.
 * */
class CMask : public std::list<std::pair<OUTLINE_Polygon*,Circle> > {
};

/** Reads an object mask from a Lexer. */
Lexer & CMASK_parse(Lexer & p,CMask & m);

/** Frees the memory used by a cmask. */
void CMASK_free(CMask * c);

/** 
 * Detects collision of two circles which are moving. 
 *
 * @param r1 is the radius of a fixed circle at the origin.
 * @param r2 is the radius of the second circle.
 * @param pos is the position of the second circle.
 * @param vel is the translation performed by the second circle.
 * @param t is the fraction of vel at which collision occurred (between zero and one).  
 * @param normal is the normal in the contact point if ther was a collision.  
 * @return true iff there was a collision.
 */
bool circle_swept_collision(real r1,real r2,const Vector2d & pos,const Vector2d & vel,real * t,Vector2d * normal);

/**
 * Detects a collision between two cmasks m1 and m2.
 * @param m is the transformation of m2 coordinates to m1 coordinates.
 * @param vel is the translation vector in m1 coordinates.
 * @return true iff there was a collision.
 * */
bool CMASK_intersect(const CMask * m1,const Matrix2x3 & m,real angleshift,const CMask * m2,const Vector2d & vel,real * t=NULL);

/**
 * Detects collision of a cmask with a circle which moves from its current location as 
 * indicated by the tarnslation vector vel. 
 */
bool CMASK_intersect(const CMask * m,const Circle & c,const Vector2d & vel,real * t);


/** Resource loader for cmasks. */
class CMaskLoader : public ResourceLoader<CMask> {
private:
    std::map<std::string,std::string> mask_paths;
public:
    CMask * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(CMask *p) {
        CMASK_free(p);
    };
};

typedef Resource<CMaskLoader> RMCMASKRef;


#endif // CMASK_H
