/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "collisionengine.h"
#include "cworld.h"
#include "gamebody.h"
#include <set>

using namespace std;

bool box_swept_collision(GameBody * gb1,GameBody * gb2) {
    // The time unit in these calculations is the time
    // passed between the moment when the bodies where at the
    // last test position and the moment when they reach their
    // current one. So we search a collision in the time interval
    // [0..1].


    // Calculate velocity of gb2 w.r.t. gb1
    // Thus, we can consider now that gb1 is not moving.
    Vector2d vel=gb2->getPosition()-gb2->lastTestPosition()-(gb1->getPosition()-gb1->lastTestPosition());
    // Calculate original position of gb2 w.r.t. gb1. i.e. put gb1 as
    // coordinates origin.
    // This is the left top corner of gb2, and the origin is on  left top
    // corner of gb1. Note we are accounting for the hotspots.
    Vector2d pos=(gb2->lastTestPosition()-gb2->hotspot)-(gb1->lastTestPosition()-gb1->hotspot);


    // This is the time needed for gb2 to reach the right border of gb1.
    real tix=vel.x!=0 ? (gb1->getWidth()-pos.x)/vel.x : 0;
    // And this is the time needed for gb2 to reach the left border of gb1.
    real tfx=vel.x!=0 ? -(pos.x+gb2->getWidth())/vel.x : 0;
    // Sort the times
    if (tix>tfx) {
        real temp=tix;
        tix=tfx;
        tfx=temp;
    }
    // If the interval does not overlap with [0..1] quit now.
    if (tfx<=0 || tix>=1) return false;


    // This is the time needed for gb2 to reach the bottom border of gb1.
    real tiy=vel.y!=0 ? (gb1->getHeight()-pos.y)/vel.y : 0;
    // And this is the time needed for gb2 to reach the top border of gb1.
    real tfy=vel.y!=0 ? -(pos.y+gb2->getHeight())/vel.y : 0;
    // Sort the times
    if (tiy>tfy) {
        real temp=tiy;
        tiy=tfy;
        tfy=temp;
    }
    // Calculate overlap with interval [0..1]
    // If it does not overlap, quit.
    if (tfx<=0 || tix>=1) return false;
    // else clip the interval to the [0..1]
    if (tiy<0) tiy=0;
    if (1<tfy) tfy=1;

    // Now if the time intervals overlap it means there was a colision
    return (tiy<tfx && tix<tfy);
}



bool circle_swept_collision(GameBody * gb1,GameBody * gb2,real * t,Vector2d * normal) {
    // The time unit in these calculations is the time
    // passed between the moment when the bodies where at the
    // last test position and the moment when they reach their
    // current one. So we search a collision in the time interval
    // [0..1].

    // Calculate original position of gb2 w.r.t. gb1. i.e. put gb1 as
    // coordinates origin.
    // This is the hotspot of gb2, and the origin is the hotspot
    // of gb1. We assuming the hotspots are the circle centers.
    Vector2d pos=gb2->lastTestPosition()-gb1->lastTestPosition();

    // Calculate velocity of gb2 w.r.t. gb1
    // Thus, we can consider now that gb1 is not moving.
    Vector2d vel=gb2->getPosition()-gb1->getPosition()-pos;

    return circle_swept_collision(gb1->hotspot.x-1,gb2->hotspot.x-1,pos,vel,t,normal);
}


int mask_collision(GameBody * gb1,GameBody * gb2,real * t,Vector2d * normal) {
    if (&*gb1->mask && &*gb2->mask) {
        Matrix2x3 m1(1,1,-gb1->getAngle(),-gb1->lastTestPosition().x,-gb1->lastTestPosition().y,gb1->hotspot.x,gb1->hotspot.y);
        Matrix2x3 m2=m1*Matrix2x3(1,1,gb2->getAngle(),-gb2->hotspot.x,-gb2->hotspot.y,gb2->lastTestPosition().x,gb2->lastTestPosition().y);
        Vector2d vel=m1*(gb2->getPosition()-gb2->lastTestPosition()-gb1->getPosition()+2*gb1->lastTestPosition());
        vel-=gb1->hotspot;
        return CMASK_intersect(&*gb1->mask,m2,M_PI*(gb2->getAngle()-gb1->getAngle())/180,&*gb2->mask,vel,t);
    } else if (!&*gb1->mask && !&*gb2->mask)
        return circle_swept_collision(gb1,gb2,t,normal);
    else {
        GameBody * g,*g2;
        Circle c;
        if (&*gb1->mask) {
            g=gb1;
            g2=gb2;
            c.center=gb2->lastTestPosition();
            c.r=gb2->hotspot.x-2;
        } else {
            g=gb2;
            g2=gb1;
            c.center=gb1->lastTestPosition();
            c.r=gb1->hotspot.x-2;
        }
        Matrix2x3 m1(1,1,-g->getAngle(),-g->lastTestPosition().x,-g->lastTestPosition().y,g->hotspot.x,g->hotspot.y);
        Vector2d vel=m1*(g2->getPosition()-g2->lastTestPosition()-g->getPosition()+2*g->lastTestPosition());
        c.center=m1*c.center;
        vel-=g->hotspot;
        return CMASK_intersect(&*g->mask,c,vel,t);
        //	else CHERROR<<pos.dist()/(i->second.r+j->second.r)<<ENDL;
    };
}

#define is(x,t) (std::string::npos!=static_cast<GameBody*>(x)->type.find(t))

bool swept_collision(GameBody * gb1,GameBody * gb2,real * t,Vector2d * normal) {
    if (is(gb1,"asteroid") && is(gb2,"asteroid")) return false;
    return mask_collision(gb1,gb2,t,normal);
    //  return circle_swept_collision(gb1,gb2,t,normal);
}
