/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef COLLISIONENGINE_H
#define COLLISIONENGINE_H

#include "collist.h"
#include "gamebody.h"
#include <list>

template<class Item,class Data>
class CollisionEngine;
class cWorld;
template <class Item,class Data>
struct _s {
    CollisionEngine<Item,Data> * ce;
    bool fordelete;
    Item * i;
};

/**
 * This class handles collisions of all the bodies on a given area.
 */
template<class Item,class Data>
class CollisionEngine {
public:

    /**
     * Creates a collision engine.
     * @param min_x left bound of the collision area.
     * @param min_y top bound of the collision area.
     * @param max_x right bound of the collision area.
     * @param max_y bottom bound of the collision area.
     */
    CollisionEngine(int min_x,int min_y,int max_x,int max_y,Data * w);

    /**
     * Destroys a collision engine.
     */
    ~CollisionEngine();

    /**
     * Adds a body to be considered by the collision engine.
     * @param c The body to add.
     */
    void add(Item * c);

    /**
     * Checks for collisions and calls the proper action for each
     * colliding object.
     */
    void step(int (*Handler)(int,void**));

    void deleteBody(Item * c);

private:
    COLLIST_LIST * clist;
    Data * data;
    typedef std::map<Item*,_s<Item,Data>* > ItemMap;
    ItemMap _sl;

    void erase(Item * c) {
        typename ItemMap::iterator i=_sl.find(c);
        if (i!=_sl.end()) {
            _sl.erase(i);
        }
    };

    template <class ItemG,class DataG>
    friend void get_item(COLLIST_ITEM * ci);
};

template <class ItemG,class DataG>
void get_item(COLLIST_ITEM * ci) {
    _s<ItemG,DataG> * d=static_cast<_s<ItemG,DataG>*>(ci->userdata);
    if (d->fordelete) {
        remove_from_collist_during_callback(d->ce->clist,ci);
        d->ce->erase(d->i);
        delete d;
        return;
    }
    GameBody * c=static_cast<GameBody*>(d->i);
    ci->x=fmin(c->getPositionX()-c->hotspot.x,c->lastTestPosition().x-c->hotspot.x);
    ci->y=fmin(c->getPositionY()-c->hotspot.y,c->lastTestPosition().y-c->hotspot.y);
    ci->w=c->getWidth()+fabs(c->getPositionX()-c->lastTestPosition().x);
    ci->h=c->getHeight()+fabs(c->getPositionY()-c->lastTestPosition().y);
}

template<class Item,class Data>
CollisionEngine<Item,Data>::CollisionEngine(int min_x,int min_y,int max_x,int max_y,Data * w) {
    clist=create_collist(min_x,min_y,max_x,max_y,get_item<Item,Data>);
    data=w;
}

template<class Item,class Data>
void CollisionEngine<Item,Data>::add(Item * c) {
    typename ItemMap::iterator i=_sl.find(c);
    _s<Item,Data> * d;
    if (i!=_sl.end()) {
        d=i->second;
    } else {
        d=new _s<Item,Data>;
        _sl[c]=d;
    }
    d->fordelete=false;
    if (i==_sl.end()) {
        d->ce=this;
        d->i=c;
        add_to_collist(clist,(void*)d);
    }
}

template<class Item,class Data>
CollisionEngine<Item,Data>::~CollisionEngine() {
    for (typename ItemMap::iterator i=_sl.begin();i!=_sl.end();i++) delete i->second;
    _sl.clear();
    destroy_collist(clist);
};



bool circle_swept_collision(GameBody * gb1,GameBody * gb2,real*t,Vector2d * normal);
bool box_swept_collision(GameBody * gb1,GameBody * gb2);
bool swept_collision(GameBody * gb1,GameBody * gb2,real*t,Vector2d * normal);

template <class Item,class Data,int (*Handler)(Data*,Item*,Item*,real,const Vector2d &)>
int wrap_collision_handler(int n, void **array) {
    _s<Item,Data> ** arr=(_s<Item,Data>**)array;
    for (int i=0;i<n;i++) {
        _s<Item,Data> * f=arr[i*2+1],*s=arr[i*2+2];
        Vector2d normal;
        real t;
        if (!f->fordelete && !s->fordelete &&
                swept_collision(static_cast<GameBody*>(f->i),static_cast<GameBody*>(s->i),&t,&normal)) {
            Handler((Data*)array[0],f->i,s->i,t,normal);
        }
    }
    return 0;
}

template<class Item,class Data>
void CollisionEngine<Item,Data>::step(int (*Handler)(int,void**)) {
    do_collist_collisions(clist,Handler,(void*)data);
    for (typename ItemMap::iterator i=_sl.begin();i!=_sl.end();i++)
        if (!i->second->fordelete) static_cast<GameBody*>(i->second->i)->testForCollision();
}

template<class Item,class Data>
void CollisionEngine<Item,Data>::deleteBody(Item * c) {
    typename ItemMap::iterator i=_sl.find(c);
    if (i!=_sl.end()) {
        i->second->fordelete=true;
    }
};

#endif // COLLISIONENGINE_H
