/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cworld.h"
#include <functional>
#include "functionexts.h"
#include "starship.h"
#include "pathdrawer.h"
#include "timer.h"
#include "collisionengine.h"
#include "collisionhandler.h"
#include "soundcollection.h"
#include "runtimeopts.h"
#include "serializer.h"
#include "debugmsg.h"

extern SoundCollection * sounds;

#include <math.h>

using namespace std;

#define MIN_ZOOM 0.01125
#define MAX_ZOOM 1.0

#define MIN_ZOOM_TIME 0.5
#define MAX_ZOOM_TIME 60.0

extern int mousePositionX;
extern int mousePositionY;

cWorld::cWorld() : realsim(PlayerState(1000,0,0),onStep),
        shipmarker("shipmarker") {
    init();
}

PlayerState * cWorld::getPlayerState() {
    return realsim.getPlayerState();
};

#define BONUSTOTAL 50

void cWorld::setBonusTime(int seconds) {
    bonusTimeUnit=seconds*(1000/BONUSTOTAL);
    accumBonusTime=0;
    bonus=BONUSTOTAL;
};

void cWorld::addBonus(int seconds) {
    if (bonusTimeUnit!=0) bonus+=seconds*1000/bonusTimeUnit;
}

int cWorld::getBonusTime() {
    return bonusTimeUnit/(1000/BONUSTOTAL);
}

int cWorld::getBonus() {
    return bonus;
}

int cWorld::getPlayerPoints() {
    return realsim.getPlayerState()->points;
};
void cWorld::setPlayerPoints(int points) {
    realsim.getPlayerState()->points=points;
};

void cWorld::init() {
    ch=new CollisionHandler(this);
    description="";
    last_delta=0;
    finish=false;
    paused=true;
    ship=NULL;
    bonusTimeUnit=0;
    bonus=0;
    accumBonusTime=0;
    shipid=-1;
    other=NULL;
    zoomTime=0.5;
    zoomTimev=1;
    eye=NULL;
    trackCoordinateSystem=NULL;
    zoomv=1.0;
    setZoom(1.0);
    shipPos=Graphic::screencenter;
    precision=20;
    realsim.physic.bigStep=160;
    pathDrawer = new PathDrawer(&realsim,onStep);
    pathRange=SHORT_RANGE;
    realsim.deletionListeners.addListener(this);
    realsim.collisionListeners.addListener(ch);
}


void cWorld::clear() {
    zoom=1;
    eyeShift=Vector2d();
    trackCoordinateSystem=eye=NULL;
    ship=NULL;

    realsim.purge();
    for (list<GameBody*>::iterator i=bg.objects.begin();i!=bg.objects.end();i++) {
        delete *i;
    }
    bg.objects.clear();
    if (pathDrawer) pathDrawer->clear();
    ship=NULL;
}
cWorld::~cWorld() {
    clear();
    if (ch) delete ch;
    if (pathDrawer) delete pathDrawer;
}

void cWorld::setZoom(real value) {
    zoom=value;
    if (zoom<MIN_ZOOM) {
        zoom=MIN_ZOOM;
        zoomv=1.0;
    } else if (MAX_ZOOM<zoom) {
        zoom=MAX_ZOOM;
        zoomv=1.0;
    }
}

bool notIsNull(PhysicBody * & p) {
    return p!=NULL;
}

cWorld::iterator cWorld::objectsBegin() {
    return cast_it<GameBody*>(filter_it(realsim.physic.bodies.begin(),realsim.physic.bodies.end(),&notIsNull));
};

cWorld::iterator cWorld::objectsEnd() {
    return cast_it<GameBody*>(filter_it(realsim.physic.bodies.end(),realsim.physic.bodies.end(),&notIsNull));
};

ostream & operator << (ostream &,const GameBody & );

void cWorld::extraTime(int delta) {
    int init=Graphic::ticks();
    if (ship && pathDrawer) {
        if (pathRange==SHORT_RANGE) {
            while (delta>int(Graphic::ticks())-init && pathDrawer->numNodes()<35) {
                if (!pathDrawer->grow(500)) break;
            }
        } else if (pathRange==MEDIUM_RANGE) {
            while (delta>int(Graphic::ticks())-init && pathDrawer->numNodes()<200) {
                if (!pathDrawer->grow(500)) break;
            }
        } else
            while (delta>int(Graphic::ticks())-init && pathDrawer->numNodes()<1000) {
                if (!pathDrawer->grow(500)) break;
            }
    }
}

void cWorld::setEyeShift(const Vector2d & value) {
    eyeShift=value;
    refreshView();
};

void cWorld::zoomStep(int delta) {
    if (zoomv!=1) {
        real oldZoom=zoom;
        zoom*=pow(zoomv,delta);
        setZoom(zoom);

        // precompute mouse position in world coordinates
        // use the center  of screen as zoom origin when zooming out
        Vector2d mouseWorldPosition = !eye ? 0 : trackCoordinateSystem->get_view2world_transformation(false,getZoom(),getEyeShift())
                                                 * (zoom>=oldZoom ? zoomEye : Vector2d(LOGICAL_WIDTH/2,LOGICAL_HEIGHT/2));

        // compute new eyeShift
        if (eye)
            eyeShift = (eye->getPosition()+eyeShift-mouseWorldPosition)*oldZoom/zoom+mouseWorldPosition-eye->getPosition();

        refreshView();
    }
}


int checkShipEngine(cWorld * w,Simulation * sim,Starship * ship,unsigned int delta) {
    if (ship) {
        int temp=(int)ship->fuelToTime(ship->getFuel());
        if (temp<(int)delta) {
            w->set(temp,EmptyEvent());
            sim->step(temp);
            ship->stepShip(temp);
            return delta-temp;
        }
    }
    return delta;
};

void cWorld::step(int delta) {
    Uint32 stepinit=Graphic::ticks();

    last_delta=delta;
    // Handle zoom view.
    zoomStep(delta);

    // Handle zoom time.
    zoomTime*=pow(zoomTimev,delta);
    if (zoomTime<MIN_ZOOM_TIME) 
        zoomTime=MIN_ZOOM_TIME;
    else if (MAX_ZOOM_TIME<zoomTime) 
        zoomTime=MAX_ZOOM_TIME;

    // Handle eyeShift
    if (eyeShiftVel!=Vector2d()) setEyeShift(eyeShift+eyeShiftVel*delta);

    if (ship && pathDrawer)
        pathDrawer->onStep();

    if (paused) {
        if (ship) {
            if (ship->powerOn()) {
              realsim.step(1);
              if (pathDrawer) 
                  pathDrawer->step(1);
			}
            ship->stepShip(delta);
        }
        return;
    }

    // Handle bonus time
    if (bonus>0 && bonusTimeUnit>0) {
        accumBonusTime+=delta;
        if (accumBonusTime>=bonusTimeUnit) {
            int delunits=accumBonusTime/bonusTimeUnit;
            if (bonus<=delunits) 
                delunits=bonus;

            bonus-=delunits;
            accumBonusTime%=bonusTimeUnit;
        };
    }

    // Step bodies.
    int delta2;
    if (zoomTime!=MIN_ZOOM_TIME) {
        delta2=(int)(delta*zoomTime);
        delta=(int)(delta*MIN_ZOOM_TIME);
    } else {
        delta=(int)(delta*MIN_ZOOM_TIME);
        delta2=delta;
    }

    if (zoomTimev>1) {
        int tempdelta=delta2;
        while (tempdelta>=500 && Graphic::ticks()-stepinit<30) {
            int ndelta=checkShipEngine(this,&realsim,ship,500);
            realsim.step(ndelta);
            if (ship) {
                ship->stepShip(ndelta);
                if (pathDrawer) 
                    pathDrawer->step(ndelta);
            }
            tempdelta-=ndelta;
        }
        if (Graphic::ticks()-stepinit<30) {
            int ndelta=checkShipEngine(this,&realsim,ship,tempdelta);
            realsim.step(ndelta);
            if (ship) {
                ship->stepShip(ndelta);
                if (pathDrawer) 
                    pathDrawer->step(ndelta);
            };
        }
    } else {
        int ndelta=checkShipEngine(this,&realsim,ship,delta2);
        realsim.step(ndelta);
        if (ship) {
            ship->stepShip(ndelta);
            if (pathDrawer) 
                pathDrawer->step(ndelta);
        };
    }

    refreshView();
}

int cWorld::getPrecision() {
    return precision;
}

void cWorld::retrace() {
    pathDrawer->retrace();
};

inline void cWorld::drawGraphic(GameBody * g,real theta,const Matrix2x3 * view) {
    g->draw(g==eye ? eyePos : *view*g->getPosition(),zoom,g->getAngle()-theta);
};

inline void cWorld::drawWithSurroundingCircle(GameBody * g,real theta,const Matrix2x3 * view) {
    Vector2d temp=*view*g->getPosition();

    if (ship && g->id==ship->id)
        shipmarker.draw(g==eye ? eyePos :temp,1,ship->getAngle());
    else {
        if (g->type.find("trash")!=std::string::npos)
            g->drawSurroundingCircle(g==eye ? eyePos : temp,zoom,g->cmfColor);
        g->draw(g==eye ? eyePos :temp,zoom,g->getAngle()-theta);
    }
};


void cWorld::draw() {
    PhysicEngine * pe=&realsim.physic;
    const real angle=0;
    if (eye) {

        bg.draw(zoom,angle,eye->getPosition()+eyeShift);

        if (ship) {
            shipPos=view_transformation*ship->getPosition();

            if (pathDrawer) 
                pathDrawer->draw(view_transformation);
        }

        if (zoom<=0.2) {
            for (vector<PhysicBody*>::iterator i=pe->bodies.begin();i!=pe->bodies.end();i++) {
                if (*i) 
                    drawWithSurroundingCircle(static_cast<GameBody*>(*i),angle,&view_transformation);
            }
        } else {
            for (vector<PhysicBody*>::iterator i=pe->bodies.begin();i!=pe->bodies.end();i++) {
                if (*i) 
                    drawGraphic(static_cast<GameBody*>(*i),angle,&view_transformation);
            }
        }

        eye->drawSurroundingCross(eyePos,zoom,SDL_MapRGB(Graphic::getScreen()->format,40,168,55));

    } else 
        bg.draw(zoom,angle,eyeShift);
}


void cWorld::setShip(Starship * s) {
    ship=s;
    if (ship) shipid=ship->id;
}


int cWorld::hittedBodies(const Vector2d & point,GameBody ** bodies,int n) {
    PhysicEngine * pe=&realsim.physic;
    if (eye) {
        Matrix2x3 view_inv(1.0/zoom,1.0/zoom,0,-Graphic::screencenter.x,-Graphic::screencenter.y, eye->getPositionX()+eyeShift.x,eye->getPositionY()+eyeShift.y);
        Vector2d pos=view_inv*point;
        return hits(pe->bodies,zoom,pos,bodies,n);
    } else return 0;
};


bool cWorld::setTrackCoordinateSystem(int x,int y) {
    GameBody * gb;
    if (hittedBodies(Vector2d(x,y),&gb,1)) {
        if (sounds) sounds->viewPointChange.play();

        setTrackCoordinateSystem(gb);
        return true;
    } else {
        trackCoordinateSystem=eye;

        return false;
    }
}

GameBody * cWorld::getBody(int id) {
    PhysicEngine * pe=&realsim.physic;
    return 0<=id && id<int(pe->bodies.size()) ? static_cast<GameBody*>(pe->bodies[id]) : NULL;
};

void cWorld::setTrackCoordinateSystem(GameBody * gb) {
    trackCoordinateSystem=static_cast<GameBody*>(realsim.physic.bodies[gb->id]);
    eyeShift=eye ? (eyeShift+eye->getPosition()-trackCoordinateSystem->getPosition()) : eyeShift-trackCoordinateSystem->getPosition();
    eye=trackCoordinateSystem;
    if (pathDrawer) 
        pathDrawer->setOrigin(eye->id);
    refreshView();
}

void cWorld::refreshView() {
    if (eye) {
        view_transformation=eye->get_view_transformation(false,zoom,eyeShift);
        eyePos=view_transformation*eye->getPosition();
    }
}

void cWorld::addBody(GameBody * c) {
    realsim.addBody(c);
    if (pathDrawer) pathDrawer->addBody(c);

    if (c->isTrash())
        getPlayerState()->remainingtrash++;

    if (!eye)
        setTrackCoordinateSystem(c);
}

void cWorld::refitBody(GameBody * b) {
    realsim.physic.refitBody(b);
};

void cWorld::deleteBody(GameBody * c) {
    if (pathDrawer) pathDrawer->deleteBody(c->id);
    realsim.deleteBody(c->id);
}

bool cWorld::win() {
    if (ship) {
        PlayerState * st=realsim.getPlayerState();
        return ship->getTrashCount() >= st->mustCollect;
    } else return false;
}

bool cWorld::unreachable() {
    if (ship) {
        PlayerState * st=getPlayerState();
        return ship->getTrashCount()+st->remainingtrash < st->mustCollect;
    } else return false;
}

void cWorld::evtDeleteBody(Simulation * sim,int id) {
    GameBody* oldOrigin = getTrackCoordinateSystem();
    if (oldOrigin && id==oldOrigin->id) {
        if (other) {
            setTrackCoordinateSystem(other);
            other=NULL;
        } else {
            eye=NULL; // don't draw, bad things might happen
            eyeShift += oldOrigin->getPosition();
        }
    }

    GameBody* g;
    if (sim==&realsim && (g=dynamic_cast<GameBody*>(sim->physic.bodies[id])) && g->isTrash()) {
        getPlayerState()->remainingtrash--;
        if (win() || unreachable()) endWorld();
    } else if (ship && id==ship->id) {
        if (sim==&realsim) {
            if (sounds) sounds->engine.stop();
            setShip(NULL);
            endWorld();
        }
    }
}


void cWorld::endWorld(int delta) {
    finish=true;
};


void cWorld::readFrom(const char * path) {
    FILE * f=fopen(path,"r");
    if (!f) {
        CHERROR<<"Error reading level "<<path<<": "<<SDL_GetError()<<ENDL;
        exit(1);
    }
    StreamTokenizer st(f);
    Lexer p(&st);
    if (parseWorld(p,*this).error()) {
        CHERROR<<"Error reading level: "<<p.errorMessage()<<ENDL;
        fclose(f);
        exit(1);
    };
    fclose(f);
};
