/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CWORLD_H
#define CWORLD_H

#include "timer.h"
#include "background.h"
#include "functionexts.h"
#include "simulation.h"
#include <vector>

class Starship;
class GameBody;
class PhysicBody;
class PhysicEngine;
class PathDrawer;
class CollisionHandler;

template <class Item,class Data>
class CollisionEngine;


/**
 * This class represents the world wich must be updated and
 * drawed in each frame.
 */
class cWorld : public SimDeletionListener {
public:
    typedef cast_iterator<GameBody*,filter_iterator<std::vector<PhysicBody*>::iterator> > iterator;

	/** Used when scrolling the view. */
    Vector2d eyeShiftVel;
	/** World description. */
    std::string description;

    PathDrawer * pathDrawer;

    typedef enum {SHORT_RANGE, MEDIUM_RANGE, LONG_RANGE} PathRange;
    PathRange pathRange;

    /** Signals that the world has reach a terminal state */
    bool finish;

    /**
     * The background.
     */
    Background bg;

    /**
     * Zoom time factor.
     */
    real zoomTime;

    /**
     * Velocity of change of zoom time factor.
     */
    real zoomTimev;

    /** Whether the world is paused.  */
    bool paused;

    /**
     * Builds the world.
     */
    cWorld();

    /**
     * Sets the zoom of the world view.
     * @param value a scale value between 0.0 and 1.0.
     */
    void setZoom(real value);
    real getZoom() {
        return zoom;
    };
    void setEyeShift(const Vector2d & value);
    Vector2d getEyeShift() {
        return eyeShift;
    };

    /**
     * Gets the begining of a list of bodies in the foreground.
     * @return an iterator to the first bodie in the foreground world.
     */
    iterator objectsBegin();

    /**
     * Gets the end of a list of bodies in the foreground.
     * @return an iterator to the end of the objects in the foreground world.
     */
    iterator objectsEnd();

    /**
     *  Returns the list of objects containing a given point.
     * @param point the queried point.
     * @param bodies the buffer to put the found objects.
     * @param n the length of the bodies buffer.
     * @return the amount of collected objects
     */
    int hittedBodies(const Vector2d & point,GameBody ** bodies,int n);

    /**
     * Makes the world change taking into account the time that has
     * passed since the last step.
     */
    void step(int delta);
    void zoomStep(int delta);

    /**
     * Tells the world it can make some computations and return the
     * flow control to the caller when elapsed a time interval or before.
     * @param delta the amount of milliseconds of the interval.
     */
    void extraTime(int delta);

    /**
     * Draws the world with all in it.
     */
    void draw();

    /**
     * Retraces the future route of the ship.
     */
    void retrace();

    /**
     * @return the time precision (in miliseconds) the world is working with.
     *         No timer events will occur in the middle of an interval of this
     *         length.
     */
    int getPrecision();

    /**
     * Sets a body which intersects current point for giving the
     * coordinate system for drawing the tracks.
     * @param x Horizontal coordinate in screen coordinates.
     * @param y Vertical coordinate in screen coordinates.
     * @return true if an object was selected.
     */
    bool setTrackCoordinateSystem(int x,int y);

    /**
     * Sets a body for giving the coordinate system for drawing the tracks.
     * @param gb The body to set.
     */
    void setTrackCoordinateSystem(GameBody * gb);

    GameBody * getBody(int id);
    GameBody * getTrackCoordinateSystem() {
        return trackCoordinateSystem;
    };

    /**
     * Sets a ship for the player.
     * @param s The ship to set.
     */
    void setShip(Starship * s);

    /**
     * Gets the body representing the player ship.
     * @returns The body representing the ship.
     */
    Starship * getShip() const {
        return ship;
    };


    /**
     * Adds the body to the world.
     * @param c The body to add.
     */
    void addBody(GameBody * c);

    /**
     * Tells the world that an object mass has changed so it must be refitted
     * in the internal data structures accordingly.
     * @param c The body to refit.
     */
    void refitBody(GameBody * b);

    /**
     * Erases a body from the world.
     * @param c The body to erase.
     */
    void deleteBody(GameBody * c);
    void evtDeleteBody(Simulation * s,int id);

    /**
     * Frees al elements on the world.
     */
    ~cWorld();

    /**
     * Empties the world of objects.
     */
    void clear();

    /**
     * The shipPos in screenCoordinates.
     */
    Vector2d shipPos;

	/** Origin for zooming in and out. */
    Vector2d zoomEye;

    PlayerState * getPlayerState();
    int getPlayerPoints();
    void setPlayerPoints(int points);
    void setBonusTime(int seconds);
    int getBonusTime();
    void addBonus(int seconds);
    int getBonus();
    bool win();
    bool unreachable();

    /**
     * Sets a timer.
     * @param delta Period of time in miliseconds to wait before calling the event.
     */
    template <class Event>
    TimerId set(int delta,const Event & evt) {
        return realsim.tevents.set(delta,evt);
    };

    /** Object whose eyes are used to see the world. */
    GameBody * eye;
    GameBody * other;

    /**
     * Gives the zoom change per milisecond.
     */
    real zoomv;

    void readFrom(const char * path);

private:

    /**
     * Zoom factor of the view.
     */
    real zoom;

    CollisionHandler * ch;

    void endWorld(int delta=500);
    void refreshView();

    int bonus;
    int bonusTimeUnit;
    int accumBonusTime;

    /** ship identifier */
    int shipid;

    /** Amount of trash in scene. */
    int remainingtrash;

    /** Duration of frames in milliseconds. */
    int precision;

    /**
     * The ship the player is going to control.
     */
    Starship * ship;
    /** Object whose coordinate system is used for tracing tracks. */
    GameBody * trackCoordinateSystem;

    void init();
    std::list<int> onStep;

    void drawGraphic(GameBody * g,real theta,const Matrix2x3 * view);
    void drawWithSurroundingCircle(GameBody * g,real theta,const Matrix2x3 * view);

    Simulation realsim;
    int last_delta;
    Matrix2x3 view_transformation;
    /**
     * An offset from the eye, useful for seeing portions of
     * the screen not centered on the eye.
     */
    Vector2d eyeShift;
    Vector2d eyePos;

    Graphic shipmarker;

    friend Lexer & parseWorld (Lexer & p,cWorld & w);
};

#endif // CWORLD_H
