/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filecoding.h"
#include <errno.h>
#include "global.h"
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include "debugmsg.h"
#include <iostream>

using namespace std;

basic_string<wchar_t> stows(const string& s) {
    basic_string<wchar_t> ws;
    ws.reserve(s.length());
    for (string::const_iterator i=s.begin();i!=s.end();i++) 
		ws+=(wchar_t)((unsigned char) *i);
    return ws;
}

basic_string<wchar_t> stows(const char * s) {
    basic_string<wchar_t> ws;
    ws.reserve(strlen(s));
    for (const char * i=s;*i!='\0';i++) 
		ws+=(wchar_t)((unsigned char) *i);
    return ws;
}

string wstos(const basic_string<wchar_t> & ws) {
    string s;
    s.reserve(ws.length());
    for (basic_string<wchar_t>::const_iterator i=ws.begin();i!=ws.end();i++)
        s+=char(*i);
    return s;
}



STDStreamFiller::STDStreamFiller(istream & i) : i(i) {
    i>>noskipws;
}

size_t STDStreamFiller::fill(char * buffer,size_t atmost) {
    return i.rdbuf()->sgetn(buffer,atmost);
}


STDStreamPrinter::STDStreamPrinter(std::ostream & o) : o(o) {};
size_t STDStreamPrinter::put(char * buffer,size_t bufsize) {
    return o.rdbuf()->sputn(buffer,bufsize);
};



StringFiller::StringFiller(const string &str) : s(str) {};

StringFiller::StringFiller(const basic_string<wchar_t> &str) {
    for (int i=0;i<int(str.length());i++)
        for (int j=0;i<int(sizeof(wchar_t));i++)
            s+=((char*)&str[i])[j];
};

size_t StringFiller::fill(char * buffer,size_t atmost) {
    if (atmost<s.length()) {
        memcpy(buffer,s.data(),atmost);
        s.erase(0,atmost);
        return atmost;
    } else {
        size_t l=s.length();
        memcpy(buffer,s.data(),l);
        s.clear();
        return l;
    }
};


IStreamUTF8Decoder::IStreamUTF8Decoder(Filler * f)
        : f(f),_error(false), state(0) {
    inbytesleft=0;
    cbufpos=cbuffer;
    refill_inputbuffer();
};


void IStreamUTF8Decoder::refill_inputbuffer() {
    memmove(cbuffer,cbufpos,inbytesleft);
    readchars=f->fill(cbuffer+inbytesleft,cbufsize-inbytesleft);
    inbytesleft+=readchars;
    cbufpos=cbuffer;
}

#define B1100_0000 0xC0
#define B1110_0000 0xE0
#define B1111_0000 0xF0
#define B1111_1000 0xF8
#define B1111_1100 0xFC
#define B1111_1110 0xFE
#define B1000_0000 0x80
#define B0011_1111 0x3F
#define B0001_1111 0x1F
#define B0000_1111 0x0F
#define B0000_0111 0x07
#define B0000_0011 0x03
#define B0000_0001 0x01


/**
 * @return the state. -1 if it is an ilegal sequence, 0 if all bytes were read,
 *         or any other value if more bytes are needed.
 * */
int utf8_decode(int* state,uint32_t* codepoint,char byte) {
    if (*state==0) {
        if (!(byte & B1000_0000))
            *codepoint=byte;
        else if ((byte & B1110_0000) == B1100_0000) {
            *codepoint= byte & B0001_1111;
            *state= *codepoint>B0000_0001? 1 : -1;
        } else if ((byte & B1111_0000) == B1110_0000) {
            *codepoint=byte & B0000_1111;
            *state= *codepoint>0? 2 : ((B0001_1111<<3) | 2);
        } else if ((byte & B1111_1000) == B1111_0000) {
            *codepoint=byte & B0000_0111;
            *state= *codepoint>0? 3 : ((B0000_1111<<3) | 3);
        } else if ((byte & B1111_1100) == B1111_1000) {
            *codepoint=byte & B0000_0011;
            *state= *codepoint>0? 4 : ((B0000_0111<<3) | 4);
        } else if ((byte & B1111_1110) == B1111_1100) {
            *codepoint=byte & B0000_0001;
            *state= *codepoint>0? 5 : ((B0000_0011<<3) | 5);
        } else
            *state=-1;
    } else if ((byte & B1100_0000) == B1000_0000) {
        *codepoint = (*codepoint<<6) | (byte & B0011_1111);
        *state = *codepoint>(unsigned int)(*state>>3) ? (*state & B0000_0111) - 1 : -1;
    } else
        *state=-1;

    return *state;
}

/**
 * @return -1 if there is not enough space in the output buffer, -2 on error,
 *            otherwise the amount of written bytes.
 * */
int utf8_encode(uint32_t c,char* obuffer,size_t obufsize) {
    if (c<0x7F) {
        if (obufsize<=0)
            return -1;
        *obuffer=(char)c;
        return 1;
    } else if (c<0x07FF) {
        if (obufsize<2)
            return -1;
        *obuffer    =(char)(  c>>6                 | B1100_0000);
        *(obuffer+1)=(char)(( c      & B0011_1111) | B1000_0000);
        return 2;
    } else if (c<0xFFFF) {
        if (obufsize<3)
            return -1;
        *obuffer    =(char)(  c>>12                | B1110_0000);
        *(obuffer+1)=(char)(((c>> 6) & B0011_1111) | B1000_0000);
        *(obuffer+2)=(char)(( c      & B0011_1111) | B1000_0000);
        return 3;
    } else if (c<0x1FFFFF) {
        if (obufsize<4)
            return -1;
        *obuffer    =(char)( (c>>18)               | B1111_0000);
        *(obuffer+1)=(char)(((c>>12) & B0011_1111) | B1000_0000);
        *(obuffer+2)=(char)(((c>> 6) & B0011_1111) | B1000_0000);
        *(obuffer+3)=(char)(( c      & B0011_1111) | B1000_0000);
        return 4;
    } else if (c<0x3FFFFFF) {
        if (obufsize<5)
            return -1;
        *obuffer    =(char)( (c>>24)               | B1111_1000);
        *(obuffer+1)=(char)(((c>>18) & B0011_1111) | B1000_0000);
        *(obuffer+2)=(char)(((c>>12) & B0011_1111) | B1000_0000);
        *(obuffer+3)=(char)(((c>> 6) & B0011_1111) | B1000_0000);
        *(obuffer+4)=(char)(( c      & B0011_1111) | B1000_0000);
        return 5;
    } else if (c<0x7FFFFFFF) {
        if (obufsize<6)
            return -1;
        *obuffer    =(char)( (c>>30)               | B1111_1100);
        *(obuffer+1)=(char)(((c>>24) & B0011_1111) | B1000_0000);
        *(obuffer+2)=(char)(((c>>18) & B0011_1111) | B1000_0000);
        *(obuffer+3)=(char)(((c>>12) & B0011_1111) | B1000_0000);
        *(obuffer+4)=(char)(((c>> 6) & B0011_1111) | B1000_0000);
        *(obuffer+5)=(char)(( c      & B0011_1111) | B1000_0000);
        return 6;
    } else
        return -2;
}


size_t IStreamUTF8Decoder::read(wchar_t* buffer,size_t bufsize) {
    if (eof() || error()) {
        _error=true;
        return 0;
    }

    size_t chars=0;
    uint32_t codepoint;
    do {
        while (inbytesleft>0 && chars<bufsize && state!=-1) {
            switch (utf8_decode(&state,&codepoint,*cbufpos)) {
            case 0:
                buffer[chars]=codepoint;
                chars++;
                break;
            case -1:
                if (chars==0)
                    _error=true;
                break;
            default:
                ;
            }
            cbufpos++;
            inbytesleft--;
        }
        if (state!=-1 && inbytesleft==0)
            refill_inputbuffer();
    } while (inbytesleft>0 && chars<bufsize && state!=-1);
    if (state==-1) {
        CHERROR<<"IStreamUTF8Decoder: invalid UTF-8 data!"<<ENDL;
        exit(1);
    } else {
    };

    return chars;
};

bool IStreamUTF8Decoder::eof() {
    return !readchars;
};

bool IStreamUTF8Decoder::error() {
    return _error;
};


OStreamUTF8Encoder::OStreamUTF8Encoder(Printer * p)
        : p(p),myerrno(0) {
};

size_t OStreamUTF8Encoder::write(const wchar_t * buffer,size_t bufsize) {
    if (error()) return 0;
    size_t chars=0;
    char * obufpos;
    while (chars<bufsize && !error()) {
        size_t outbytesleft=obufsize;
        obufpos=obuffer;

        bool enough_room=true;
        while (enough_room && chars<bufsize && !error()) {
            int res = utf8_encode(buffer[chars],obufpos,outbytesleft);
            if (res>=0) {
                obufpos+=res;
                outbytesleft-=res;
                chars++;
            } else if (res==-1)
                enough_room = false;
            else if (res==-2)
                myerrno=-2;

        }
        p->put(obuffer,obufpos-obuffer);
    }
    return chars;
};

int OStreamUTF8Encoder::error() {
    return myerrno;
};


