/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GAMEBODY_H
#define GAMEBODY_H

#include "graphic.h"
#include "physicbody.h"
#include <functional>
#include <algorithm>
#include <list>
#include <string>
#include "cmask.h"


class GameBody : public Graphic, public PhysicBody {
public:
    /**
     * A string to tell what kind of bodie is this.
     */
    std::string type;
    RMCMASKRef mask;

    /**
     * Color used to draw the surrounding circumference of this body.
     */
    Uint32 cmfColor;

    GameBody(real x0=0,real y0=0,real vx0=0,real vy0=0,real m=1e3,real w=0);

    /** Builds a body with mass.
     * @param m The mass of the body.
     * @param x0 Initial horizontal coordinate.
     * @param y0 Initial vertical coordinate.
     * @param vx0 Initial horizontal velocity.
     * @param vy0 Initial vertical velocity.
     * @param graphicid Identifier of the image for this body.
     */
    GameBody(const std::string & graphicid,real x0=0,real y0=0,real vx0=0,real vy0=0,real m=1e3,real w=0,bool usecache=true,bool prerotate=false);

    /** Builds a body with mass.
     * @param m The mass of the body.
     * @param pos Initial position.
     * @param vel Initial velocity.
     * @param graphicid Identifier of the image for this body.
     */
    GameBody(const std::string & graphicid,Vector2d pos,Vector2d vel=Vector2d(0,0),real m=1e16,real w=0,bool usecache=true,bool prerotate=false);

    /**
     * This method is called whenever the world is going to execute a
     * step and the object is registered in order to have this method called.
     */
    virtual void onStep() {};

    /**
     * Gets the view transformation for this body.
     * @param zoom The zoom factor of the view.
     * @param rotate Should the view rotate with the body orientation.
     * @param shift the translation from the body to get the view.
     * @return The matrix representing the transformation.
     */
    Matrix2x3 get_view_transformation(bool rotate,real zoom,const Vector2d & shift=0) const;

    /**
     * Gets the world to object transformation for this body.
     * @return The matrix representing the transformation.
     */
    Matrix2x3 get_view2world_transformation(bool rotate,real zoom,const Vector2d & shift=0) const;

    /**
     * Decides whether an object contains a given point or not.
     * @param point the queried point.
     * @return true if the body contains the point, false otherwise.
     */
    bool contains(const Vector2d * point) const;

    /**
     * Polymorphic copy constructor.
     */
    PhysicBody * clone();
    void copy(PhysicBody * dest) const {
        *dynamic_cast<GameBody*>(dest)=*this;
    }

    /**
     * Decides whether an object resized contains a given point or not.
     * @param zoom a scaling factor for resizing the object before the check.
     * @param point the queried point.
     * @return true if the resized body contains the point, false otherwise.
     */
    bool contains(real zoom,const Vector2d * point) const;

    /**
     * Decides whether an object touches a given rectanglepoint or not.
     * @param r the queried rectangle.
     * @return true if the body touches the rectangle, false otherwise.
     */
    bool touches(const SDL_Rect * r);

    /** Tells if a body is a trash or not. */
    bool isTrash() const;

    virtual std::ostream & serialize(std::ostream & o) const;

};

/**
 *  Returns the list of objects containing a given point.
 * @param s the sequence in which to search for the objects.
 * @param point the queried point.
 * @param bodies the buffer to put the found objects.
 * @param n the length of the bodies buffer.
 * @return the amount of collected objects
 */
template <class Sequence>
int hits(const Sequence & s,const Vector2d & point,GameBody ** bodies,int n) {
    int pos=0;
    typename Sequence::const_iterator i=s.begin();
    while (pos<n && i!=s.end()) {
        if ((*i) && static_cast<GameBody*>(*i)->contains(&point)) bodies[pos++]=static_cast<GameBody*>(*i);
        i++;
    }
    return pos;
}

/**
 *  Returns the list of objects containing a given point.
 * @param s the sequence in which to search for the objects.
 * @param zoom a scaling factor for resizing the objects before the check.
 * @param point the queried point.
 * @param bodies the buffer to put the found objects.
 * @param n the length of the bodies buffer.
 * @return the amount of collected objects
 */
template <class Sequence>
int hits(const Sequence & s,real zoom,const Vector2d & point,GameBody ** bodies,int n) {
    int pos=0;
    typename Sequence::const_iterator i=s.begin();
    while (pos<n && i!=s.end()) {
        if ((*i) && static_cast<GameBody*>(*i)->contains(zoom,&point)) bodies[pos++]=static_cast<GameBody*>(*i);
        i++;
    }
    return pos;
}

#endif // GAMEBODY_H
