/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "gamecontrols.h"
#include "cworld.h"
#include "starship.h"
#include <SDL_image.h>
#include <SFont.h>
#include "global.h"
#include <sstream>
#include "soundcollection.h"
#include "runtimeopts.h"
#include <iostream>
#include "fade.h"
#include "music.h"
#include "debugmsg.h"

extern SoundCollection * sounds;

using namespace std;

int mousePositionX=0;
int mousePositionY=0;
bool shiftPressed=false;
Vector2d scrollDir;
bool mousePressedFlag=false;

/** 
 * Abstract interface for timed effects like render text on top of the scene
 * for a period of time, fade ins, fade outs, timers or moving the view towards
 * a given object. 
 */
class Effect {
public:
    virtual void draw()=0;
    virtual bool step(int delta)=0;
    virtual void process_event(const SDL_Event & e)=0;
    virtual ~Effect() {};
};

class FadeInEffect : public Effect {
private:
    Fade f;
public:
    FadeInEffect() {
        f.setMode(Fade::IN,1000);
    }
    void draw() {
        Graphic::setAlwaysUpdate(f.step());
    };
    bool step(int delta) {
        if (f.getMode()==Fade::NONE) return true;
        else return false;
    };
    void process_event(const SDL_Event &) {};
};

/** Draws text on top of the screen. */
class TextEffect : public Effect {
public:
    TextEffect(const std::string & text,cWorld * w,int delay=-1) : text(text), w(w), delay(delay),font("24p") {}
    void draw() {
        font->drawL(int(Graphic::screencenter.x),int(Graphic::screencenter.y)>>2,
                   text,FontRenderer::CENTER);
    };
    bool step(int delta) {
        if (!w->paused) 
			delay=0;

        if (delay>delta) {
            delay-=delta;
            return false;
        } else if (delay>=0) {
            return true;
        } else return false;
    }
    void process_event(const SDL_Event & e) {
        switch(e.type) {
            case SDL_KEYDOWN:
            case SDL_KEYUP:
                switch (e.key.keysym.sym) {
                case SDLK_p:
                    delay=0;
                    break;
                default:;
                };
                break;
        default:;
        }
    };
private:
    std::string text;
    cWorld * w;
    int delay;
    RMFONTRef font;
};

/** A timer to finish the world simulation. */ 
class EndWorld : public Effect {
private:
    GameControls  * g;
    int delay;
public:
    EndWorld(GameControls  * g):g(g),delay(2000) {};
    void draw() {};
    bool step(int delta) {
        if (delay>delta) {
            delay-=delta;
            return false;
        } else {
            // finish the loop
            g->done=true;
            return true;
        }
    }
    void process_event(const SDL_Event & e) {
    };
};

/** Timer to play music when it fires. */
class PlayMusic : public Effect {
private:
    int delay;
    std::string musid;
public:
    PlayMusic(int delay,const std::string & musid):delay(delay),musid(musid) {};
    void draw() {};
    bool step(int delta) {
        if (delay>delta) {
            delay-=delta;
            return false;
        } else {
            Music::play(musid);
            return true;
        }
    }
    void process_event(const SDL_Event & e) {
    };
};


/** Has the view look for a given object in the world. */
class FollowBody : public Effect {
private:
    cWorld  * w;
    int follow;
    Vector2d scrollvel;
public:
    FollowBody(cWorld * w,int follow):w(w),follow(follow), scrollvel(w->getZoom(),0) {};
    void draw() {};

    bool step(int delta) {

        GameBody * g=w->getBody(follow);
        if (g && w->getTrackCoordinateSystem()) {

            Vector2d v=g->getPosition()-(w->getTrackCoordinateSystem()->getPosition()+w->getEyeShift());
            real d=v.dist();
            if (d*w->getZoom()<10) 
                return true;
            scrollvel=v*scrollvel.dist()/d;

            if (v.dist()-scrollvel.sqrDist()*1.5/FASTDECREASERATE/w->getZoom()>60) {
                fastenScroll(delta,&scrollvel.y,sgn(scrollvel.y));
                fastenScroll(delta,&scrollvel.x,sgn(scrollvel.x));
            } else {
                lowerScroll(delta,&scrollvel.y);
                lowerScroll(delta,&scrollvel.x);
            }
            w->setEyeShift(w->getEyeShift()+delta*scrollvel/w->getZoom());

            return scrollvel==Vector2d();
        }

        return true;
    }
    void process_event(const SDL_Event & e) {
    };
};

GameControls * gc;
SDL_mutex * effectsm=NULL;

int nextm=1;

/** Called whenever a music ends playing during a level. */
void cycleMusic() {
    std::string current="intro";
    int delay=30000;
    nextm++;
    nextm%=3;
    if (Music::getCurrent()=="intro") 
        current="space1";
    else if (Music::getCurrent()=="space1") 
        current="space2";
    else {
        current="intro";
        nextm=1;
    }

    if (effectsm) {
        SDL_LockMutex(effectsm);
        if (gc) gc->effects.push_back(new PlayMusic(delay,current));
        SDL_UnlockMutex(effectsm);
    }
}


GameControls::GameControls(cWorld *w) :
        ic(w),
        w(w),
        zc(w), shipviewbt(18,Graphic::getScreen()->h-84,"","shipviewbt"),
        viewpointbt(45,Graphic::getScreen()->h-98,"","viewpointbt"),
        orbitlengthbt(74,Graphic::getScreen()->h-83,"","orbitlengthbt"),
        timebt(122,Graphic::getScreen()->h-55,"","timebt"),
        quitbt("quitbt",false),
        helpbt("helpbt",false),
        scoredisplay("scoredisplay",false),
        active(true), focus(true),ending(false),
        font("24p"),font12("12p"), font24pgray("24pgray") {
    wheelcount=0;
    wheel=0;

    effectsm=SDL_CreateMutex();
    if (!effectsm) {
        fprintf(stderr,"Could not create effects mutex: %s\n", SDL_GetError());
        exit(1);
    }

    lastBonus=0;
    lastCollected=0;
	pauseMode=false;
    begin();
    done=false;
    chTimeForward=-1;

    scorelabel=getText("scorelabel");
    bonuslabel=getText("bonuslabel");
    collectlabel=getText("collectlabel");

    scorex = (LOGICAL_WIDTH-scoredisplay.getWidth())/2;
    scorelabelx = scorex + 59;
    bonuslabelx = scorex + 200;
    collectlabelx = scorex + 343;
}

#define NEXTMUSIC (nextm==0? "intro" : (nextm==1 ? "space1" : "space2" ))

void GameControls::begin() {
    stringstream temp;
    temp<<w->description<<"\n\n";

    std::string msgcollect=getText("msgcollect");
    std::string msgpiece=getText("msgpiece");
    std::string msgpieces=getText("msgpieces");

    if (w->getPlayerState()->mustCollect>0) {
        temp<<msgcollect<<w->getPlayerState()->mustCollect;
        if (w->getPlayerState()->mustCollect>1) temp<<msgpieces;
        else temp<<msgpiece;
    }
    if (w->getPlayerState()->mustCollect>0)
        temp<<'.';

    helpstr=temp.str();
    SDL_LockMutex(effectsm);
    effects.push_back(new TextEffect(helpstr,w));
    effects.push_back(new FadeInEffect());
    SDL_UnlockMutex(effectsm);

    gc=this;
    if (!Music::isPlaying()) Music::play(NEXTMUSIC);
    Music::whenFinished(&cycleMusic);

}


bool keyon=false;
int shipid=0;

void GameControls::step(int delta) {
    w->step(delta);
    ic.step(delta);
    zc.step(delta);

    SDL_LockMutex(effectsm);
    list<Effect*>::iterator i=effects.begin();
    while (i!=effects.end())
        if ((*i)->step(delta)) {
            list<Effect*>::iterator j=i++;
            delete *j;
            effects.erase(j);
        } else i++;
    SDL_UnlockMutex(effectsm);

    // Screen scrolling
    if (focus && active) 
        mscroll.update(delta,scrollDir);
    else 
        mscroll.lower(delta);
    w->eyeShiftVel=mscroll.scrollvel/w->getZoom();
};

#define ZOOMCONTROLX 47
#define ZOOMCONTROLY(screen) (screen->h-48)


void GameControls::draw() {
    SDL_Surface * screen=Graphic::getScreen();
    Graphic::clearScreen(SDL_MapRGB(screen->format, 0, 0,0));
    w->draw();
    if (w->getShip())
        fuelbar.draw(w->getShip()->getFuel()/w->getShip()->maxFuel);
    else
        fuelbar.draw(-1);

    /****************************************************************
       Draw GUI images
    *****************************************************************/

    ic.draw();
    zc.draw(ZOOMCONTROLX,ZOOMCONTROLY(screen));
    shipviewbt.draw();
    viewpointbt.draw();
    orbitlengthbt.draw();
    timebt.draw();
    UPDATEOFF(quitbt.draw(LOGICAL_WIDTH-quitbt.getWidth(),0));
    UPDATEOFF(helpbt.draw(LOGICAL_WIDTH-quitbt.getWidth(),LOGICAL_HEIGHT-helpbt.getHeight()));
    UPDATEOFF(scoredisplay.draw(scorex,0));
    ic.drawCursor();

    /****************************************************************
        Draw TEXTS
    *****************************************************************/

    // player stats
    ostringstream ss;
    ss<<w->getPlayerPoints();
    UPDATEOFF(font12->drawL(240,5,ss.str(),FontRenderer::RIGHT));
    UPDATEOFF(font24pgray->drawL(scorelabelx,1,scorelabel));
    UPDATEOFF(font24pgray->drawL(bonuslabelx,1,bonuslabel));
    UPDATEOFF(font24pgray->drawL(collectlabelx,1,collectlabel));

    ss.str("");
    ss<<w->getBonus();
    if (lastBonus!=w->getBonus()) {
        lastBonus=w->getBonus();
        UPDATEOFF(font12->drawL(380,5,ss.str(),FontRenderer::RIGHT));
    } else 
        font12->drawL(380,5,ss.str(),FontRenderer::RIGHT);

    int c=lastCollected;
    PlayerState * st=w->getPlayerState();
    if (w->getShip())
        c=st->mustCollect-w->getShip()->getTrashCount();

    ss.str("");
    ss<<(c>0?c:0);

    if (lastCollected!=c)
        font12->drawL(520,5,ss.str(),FontRenderer::RIGHT);
    else 
        UPDATEOFF(font12->drawL(520,5,ss.str(),FontRenderer::RIGHT));
    lastCollected=c;

    SDL_LockMutex(effectsm);
    map_container(list<Effect*>,effects,(*i)->draw());
    SDL_UnlockMutex(effectsm);
}

bool checkQuitBt(const Graphic & quitbt,int x,int y) {
    return ((Vector2d(x,y)-Vector2d(Graphic::getScreen()->w-quitbt.getWidth()/2,quitbt.getHeight()/2)).sqrDist()<100);
}

bool checkHelpBt(const Graphic & helpbt,int x,int y) {
    return ((Vector2d(x,y)-Vector2d(Graphic::getScreen()->w-helpbt.getWidth()/2,
                                    Graphic::getScreen()->h-helpbt.getHeight()/2)).sqrDist()<100);
}

bool GameControls::finish() {
    return done;
};

class GCShipEngineOn : public Event {
private:
    int id;
public:
    GCShipEngineOn(int id):id(id) {};
    void setQueue(Simulation *,const TimerId &) {}
    int fire(Simulation * sim) {
        Starship*ship=static_cast<Starship*>(sim->physic.bodies[id]);
        //ship->setAngle(angle);
        ship->turn_on_engine();
        keyon=true;
        return 0;
    }
    void cancel(Simulation *) {}
};

class GCShipEngineOff : public Event {
private:
    int id;
    cWorld * w;
public:
    GCShipEngineOff(cWorld *w,int id):id(id), w(w) {};
    void setQueue(Simulation *,const TimerId &) {}
    int fire(Simulation * sim) {
        Starship*ship=dynamic_cast<Starship*>(sim->physic.bodies[id]);
        if (ship) {
            ship->turn_off_engine();
            keyon=false;
            w->retrace();
        }
        return 0;
    }
    void cancel(Simulation *) {}
};


// Tells scrollDir from mouse position.
Vector2d findScrollDir(int x, int y) {
    Vector2d mv = Vector2d(x,y)-Graphic::screencenter;
    real d = mv.dist();
    if (d==0)
        return 0;
    else
        return mv/d;
}

int snapshotcounter=0;

#define WHEELDELAY (120)

bool GameControls::process_events(Uint32 delta) {

    bool mouseMoved=false;

    SDL_Event e;
    Vector2d acc;
    ostringstream ss;
    int i=0;

    if (w->finish && !ending) {
        SDL_LockMutex(effectsm);
        effects.push_back(new EndWorld(this));
        SDL_UnlockMutex(effectsm);
    }

    if (finish()) {
        if (sounds) 
			sounds->timeForward.stop(chTimeForward);
        return false;
    }

    while (SDL_PollEvent(&e) && i++<30) {

        switch (e.type) {
        case SDL_KEYDOWN:
            switch (e.key.keysym.sym) {
            case SDLK_p:
				pauseMode = !pauseMode;
                w->paused = pauseMode;
                break;
            case SDLK_LEFT:
                if (shiftPressed) {
                    scrollDir.x = -1;
                    break;
                }
                if (w->paused)
                    w->paused=false;
                if (w->getShip()) 
                    w->getShip()->turn_left();
                break;
            case SDLK_RIGHT:
                if (shiftPressed) {
                    scrollDir.x = 1;
                    break;
                }
                if (w->paused) 
                    w->paused=false;
                if (w->getShip()) 
                    w->getShip()->turn_right();
                break;
            case SDLK_UP:
                if (shiftPressed) {
                    scrollDir.y = -1;
                    break;
                }
                if (w->paused) 
                    w->paused=false;
                if (w->getShip()) {
                    shipid=w->getShip()->id;
                    // We need to push this event to have a smooth path drawer.
                    // This also makes switching repeteadly the engine to work the
                    // same than holding the engine on.
                    w->set(0,GCShipEngineOn(w->getShip()->id));
                }
                break;
            case SDLK_RSHIFT:
            case SDLK_LSHIFT:
                shiftPressed = true;
                break;
            case SDLK_RCTRL:
            case SDLK_LCTRL:
                w->pathRange=cWorld::PathRange((w->pathRange+1)%3);
                ic.evtResetPath();
                break;
            case SDLK_KP_PLUS:
            case SDLK_PLUS:
				w->zoomEye.x = LOGICAL_WIDTH/2;
				w->zoomEye.y = LOGICAL_HEIGHT/2;
                zc.pressIn();
                break;
            case SDLK_KP_MINUS:
            case SDLK_MINUS:
                zc.pressOut();
                break;
            case SDLK_s:
                ss.str("");
                ss<<snapshotcounter++;
                SDL_SaveBMP(Graphic::getScreen(),(ss.str()+".bmp").c_str());
                break;
            case SDLK_KP_MULTIPLY:
            case SDLK_ASTERISK:
            case SDLK_DOWN:
                if (shiftPressed) {
                    scrollDir.y = 1;
                    break;
                }
                if (sounds) chTimeForward=sounds->timeForward.play();

                if (w->paused) 
                    w->paused=false;
                w->zoomTimev=100.0/99.0;
                timebt.setState(Button::PRESSED);
                break;
            case SDLK_ESCAPE:
                return false;
            default:;
            }
            break;
        case SDL_KEYUP:
			if (pauseMode)
				w->paused = pauseMode;
            switch (e.key.keysym.sym) {
            case SDLK_RSHIFT:
            case SDLK_LSHIFT:
                shiftPressed = false;
                scrollDir = 0;
                break;
            case SDLK_KP_PLUS:
            case SDLK_PLUS:
                zc.releaseIn();
                break;
            case SDLK_KP_MINUS:
            case SDLK_MINUS:
                zc.releaseOut();
                break;
            case SDLK_KP_MULTIPLY:
            case SDLK_ASTERISK:
            case SDLK_DOWN:
                if (shiftPressed && scrollDir.y!=0) {
                    scrollDir.y = 0;
                    break;
                }
                if (sounds) sounds->timeForward.stop(chTimeForward);
                w->zoomTime=0.5;
                w->zoomTimev=1;
                timebt.setState(Button::RELEASED);
                break;
            case SDLK_UP:
                if (shiftPressed && scrollDir.y!=0) {
                    scrollDir.y = 0;
                    break;
                }
                if (w->getShip()) {
                    shipid=w->getShip()->id;
                    // We need to push this event to have a smooth path drawer.
                    // This also makes switching repeteadly the engine to work the
                    // same than holding the engine on.
                    w->set(1,GCShipEngineOff(w,w->getShip()->id));
                }
                break;
            case SDLK_LEFT:
                if (shiftPressed && scrollDir.x!=0) {
                    scrollDir.x = 0;
                    break;
                }
                if (w->getShip()) {
                    w->getShip()->stop_turning();
                    w->retrace();
                }
                break;
            case SDLK_RIGHT:
                if (shiftPressed && scrollDir.x!=0) {
                    scrollDir.x = 0;
                    break;
                }
                if (w->getShip()) {
                    w->getShip()->stop_turning();
                    w->retrace();
                }
                break;
            default:
                ;
            }
            break;
        case SDL_MOUSEBUTTONDOWN:
            mousePressedFlag=true;
            mouseMoved=false;
            switch (e.button.button) {
            case SDL_BUTTON_RIGHT:
                ic.mouseClick(mousePositionX,mousePositionY,false);
                break;
            case SDL_BUTTON_LEFT:
                zc.mousePress(ZOOMCONTROLX,ZOOMCONTROLY(Graphic::getScreen()),mousePositionX,mousePositionY);
                if (zc.state==ZoomControl::Z_NONE) {
                    if (!shipviewbt.mousePressed(mousePositionX,mousePositionY,true)) {
                        if (!viewpointbt.mousePressed(mousePositionX,mousePositionY,true)) {
                            if (timebt.mousePressed(mousePositionX,mousePositionY,true)) {
                                if (sounds) 
                                    chTimeForward=sounds->timeForward.play();
                                if (w->paused) 
                                    w->paused=false;
                                w->zoomTimev=100.0/99.0;
                            } else if (!orbitlengthbt.mousePressed(mousePositionX,mousePositionY,true)) {
                                if (checkHelpBt(helpbt,mousePositionX,mousePositionY)) {
                                    w->paused=true;
                                    SDL_LockMutex(effectsm);
                                    effects.push_back(new TextEffect(helpstr,w));
                                    SDL_UnlockMutex(effectsm);
                                } else if (checkQuitBt(quitbt,mousePositionX,mousePositionY))
                                    return false;
                                else {
                                    if (SDL_GetModState() & KMOD_SHIFT)
                                        scrollDir=findScrollDir(mousePositionX,mousePositionY);
                                    ic.mouseClick(mousePositionX,mousePositionY,true);
                                }
                            }
                        }
                    }
                };
                break;
            case SDL_BUTTON_WHEELUP:
                zc.pressOut();
                if (wheel) 
					wheelcount+=WHEELDELAY;
                else {
                    wheelcount=delta+WHEELDELAY;
                    wheel=1;
                }
                break;
            case SDL_BUTTON_WHEELDOWN:
				w->zoomEye.x = mousePositionX;
				w->zoomEye.y = mousePositionY;
                zc.pressIn();
                if (wheel) 
					wheelcount+=WHEELDELAY;
                else {
                    wheelcount=delta+WHEELDELAY;
                    wheel=-1;
                }
                break;
            default:
                ;
            }
            break;
        case SDL_MOUSEBUTTONUP:
            mousePressedFlag=false;
            scrollDir=0;
            if (e.button.button!=SDL_BUTTON_WHEELDOWN && e.button.button!=SDL_BUTTON_WHEELUP) 
                zc.release();
            if (shipviewbt.mouseReleased(mousePositionX,mousePositionY,e.button.button==SDL_BUTTON_LEFT)) {
                if (w->getShip()) {
                    SDL_LockMutex(effectsm);
                    effects.push_back(new FollowBody(w,w->getShip()->id));
                    SDL_UnlockMutex(effectsm);
                }
                if (sounds) sounds->shipbuttons.play();
            } else if (viewpointbt.mouseReleased(mousePositionX,mousePositionY,true)) {
                if (w->getTrackCoordinateSystem()) {
                    SDL_LockMutex(effectsm);
                    effects.push_back(new FollowBody(w,w->getTrackCoordinateSystem()->id));
                    SDL_UnlockMutex(effectsm);
                }
                if (sounds) sounds->shipbuttons.play();
            } else if (orbitlengthbt.mouseReleased(mousePositionX,mousePositionY,true)) {
                w->pathRange=cWorld::PathRange((w->pathRange+1)%3);
                if (sounds) sounds->shipbuttons.play();
                ic.evtResetPath();
            } else if (timebt.mouseReleased(mousePositionX,mousePositionY,true)) {
                if (sounds) sounds->timeForward.stop(chTimeForward);
                w->zoomTime=0.5;
                w->zoomTimev=1;
            };
            break;
        case SDL_MOUSEMOTION:
            mousePositionX=Graphic::fromScreenx(e.motion.x);
            mousePositionY=Graphic::fromScreeny(e.motion.y);
            mouseMoved=true;
            if (mousePressedFlag && (SDL_GetModState() & KMOD_SHIFT))
                scrollDir = findScrollDir(e.motion.x, e.motion.y);
            break;
        case SDL_QUIT:
			// throw generates most likely a call to GameControls destructor 
			// in the calling code 
            throw 0;
            return false;
        case SDL_ACTIVEEVENT: {
            Uint8 st=SDL_GetAppState();
            if (st & SDL_APPMOUSEFOCUS) 
				focus=e.active.gain;
            if (st & (SDL_APPINPUTFOCUS | SDL_APPACTIVE)) {
                w->paused=pauseMode || !e.active.gain;
				if (!e.active.gain)
					Music::pause();
				else
					Music::resume();
			}
        }
        if (!w->paused) 
			Graphic::fullUpdate();
        break;
        default:
            ;
        }
        SDL_LockMutex(effectsm);
        map_container(list<Effect*>,effects,(*i)->process_event(e));
        SDL_UnlockMutex(effectsm);
    }

    if (wheel) {
        if (delta<=wheelcount) 
			wheelcount-=delta;
        else wheelcount=0;
        if (!wheelcount) {
            zc.release();
            wheel=0;
        };
    }

    if (mouseMoved) {
        mouseMotion(mousePositionX,mousePositionY);
        shipviewbt.mouseMotion(mousePositionX,mousePositionY);
        viewpointbt.mouseMotion(mousePositionX,mousePositionY);
        orbitlengthbt.mouseMotion(mousePositionX,mousePositionY);
    }

    return true;
};


void GameControls::mouseMotion(int x, int y) {
    ic.mouseMotion(x,y);
};

GameControls::~GameControls() {
    Music::whenFinished(NULL);
    gc=NULL;
    delete_container_elements(list<Effect*>,effects);
    SDL_DestroyMutex(effectsm);
    effectsm=NULL;
}


void GameControls::playLevel() {
    Music::load("intro");
    Music::load("space1");
    Music::load("space2");

    int last_step=Graphic::ticks();
    int timeDraw=0,timeStep=0, timeOther=0,timeProc=0;
    int time = last_step;
    int init = last_step;
    int fs=0;
    int delta=0;

    UPDATEbegin;
    Graphic::fullUpdate();

    while (process_events(delta)) {
        fs++;
        timeProc+=Graphic::ticks()-time;
        time = Graphic::ticks();

        step(delta);
        timeStep+=Graphic::ticks()-time;
        time = Graphic::ticks();

        // If we still have some free time, wait.
        delta=Graphic::ticks()-last_step;
        if (delta<30)
            w->extraTime(30-delta);
        else w->extraTime(10);

        timeOther+=Graphic::ticks()-time;
        if (Graphic::ticks()-last_step<20) SDL_Delay(10);
        time = Graphic::ticks();
        delta=time-last_step;
        last_step+=delta;

        draw();
        Graphic::flip(false);
        timeDraw+=Graphic::ticks()-time;
        time = Graphic::ticks();
    }
    time-=init;

    bool playinggameover=false;
    if (!w->getShip() || w->unreachable()) {
        playinggameover=true;
        Music::play("GameOverMus");
    }

    Fade f;
    f.setMode(Fade::OUT,1000);
    while (f.getMode()!=Fade::NONE) {
        step(delta);
        // If we still have some free time, wait.
        delta=Graphic::ticks()-last_step;
        if (delta<30)
            w->extraTime(30-delta);
        else w->extraTime(10);

        if (Graphic::ticks()-last_step<20) SDL_Delay(10);

        delta=Graphic::ticks()-last_step;
        last_step+=delta;

        draw();
        Graphic::setAlwaysUpdate(f.step());
        Graphic::flip(false);
    }
    if (sounds) 
        sounds->engine.stop();
    if (!playinggameover && (!w->getShip() || w->unreachable())) Music::play("GameOverMus");

    UPDATEend;

    if (time>0) {
        CHPROFILE<<"process events: "<<timeProc/double(time)<<"  step: "<<timeStep/double(time);
        CHPROFILE<<"  draw: "<<timeDraw/double(time)<<"  other: "<<timeOther/double(time)<<"\n";
        CHPROFILE<<"fps: "<<fs*1000/double(time)<<ENDL;
    }
}
