/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GRAPHIC_H
#define GRAPHIC_H

#include <SDL.h>
#include "reference.h"
#include "resourcemanager.h"

#include "vector2d.h"

#define LOGICAL_WIDTH 640
#define LOGICAL_HEIGHT 480


class ImageLoader : public ResourceLoader<SDL_Surface> {
    std::set<std::string> img_paths;
public:
    SDL_Surface * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(SDL_Surface *p) {
        SDL_FreeSurface(p);
    };
};

typedef Resource<ImageLoader> RMIMGRef;

class ImageSetLoader : public ResourceLoaderNonRef<std::pair<Ref<RMIMGRef,delarray>,int> > {
    std::map<std::string,std::vector<std::string>*> img_sets;
public:
    std::pair<Ref<RMIMGRef,delarray>,int> create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(std::pair<Ref<RMIMGRef,delarray>,int> & p) {
        p.first=Ref<RMIMGRef,delarray>();
    };
    ~ImageSetLoader();
};

typedef ImageSetLoader::ResourceT RMIMGSETRef;

inline RMIMGSETRef getImageSet(const std::string & id) {
    return Singleton<ImageSetLoader>::getInstance()->create(id);
}


/**
 * This class holds all the necesary graphic code for the game.
 * Classes wich wants to draw themselves onto screen must inherit this
 * class.
 *
 * A Graphic object has an asociated image. This image is rendered on coordinate
 * system acording to its position (use setPosition to set it).
 * The postion of an image is the position of its hotspot. The hotspot
 * is a point specified relative to the upper left corner of the image.
 *  ___________________________________
 * |                                   |
 * | X                                 |
 * |                                   |
 * |                                   |
 * |___________________________________|
 *
 *  We could say here that the hotspot X is at coordinates (1,1) relative
 *  to the upper left corner (0,0).
 *  So if the hotspot is at (30,29) in global coordinates, then the upper
 *  left corner of the image is at (29,28) at global coordinates.
 *
 *  This class also provides some static methods for managing the graphic system.
 */
class Graphic {
private:
    Ref<SDL_Surface,SDL_FreeSurface> cache;
    real cache_angle,cache_zoom;
    SDL_Rect cache_rect;
    RMIMGSETRef images;
    RMROTRef rot;
private:
    friend class ZoomRotator;
    friend class FontRenderer;
    static float base_scalefactor;
    static bool updates;
    static SDL_Surface * screen;
    static bool fullupdate, alwaysupdate;
    static void mUpdateRect(SDL_Surface * s,SDL_Rect * r);
    Uint8 ROT_FLAG;

protected:

    /** What to draw. */
    RMIMGRef bmp;

    /**
     * Creates a graphic without an associated image;
     * Do not use any drawing method prior to asigning an image to this graphic an its cache.
     * This constructor is provided for subclases which cannot associate an image at the
     * begining of constructor execution.
     */
    Graphic();

public:
    /**
     * Flags to pass to sdl when setting a video mode.
     */
    static Uint32 sdlflags;

    /** Position of the image handle relative to its uper left corner. */
    Vector2d hotspot;

    /** screen for sdl */
    static SDL_Surface * getScreen() {
        return screen;
    };

    static void setScaleFactors();
    static float getBaseScaleFactorX();
    static float getBaseScaleFactorY();

    /** screen center, this is used for centering the view of the game
      * and the level editor.
      */
    static Vector2d screencenter;
    static int screenx;
    static int screeny;
    static int screenw;
    static int screenh;

    /** Converts logical coordinates and lengths to screen coordinates. */
    static int toScreenx(int x);
    static int toScreeny(int y);
    static int toScreenw(int w);
    static int toScreenh(int h);
    static int fromScreenx(int x);
    static int fromScreeny(int y);

    /**
     * Sets video mode for displaying graphics to a suitable
     * bitdepth for a resolution of w x h.
     * If it fails it calls exit(1);
     */
    static void set_video_mode(int w, int h);

    /** Sets the palete for the video mode, if it is indexed.
     */
    static void setPalette(const std::string & id,int flags=SDL_LOGPAL | SDL_PHYSPAL);
    static void setPalette(const SDL_Palette * pal);
    SDL_Palette * getPalette();

    /**
     * Toggles fullscreen mode.
     */
    static void toggleFullscreen();
    static void updateRect(SDL_Surface * s,SDL_Rect * r);
    static void updateRect(SDL_Rect * r) {
        updateRect(getScreen(),r);
    }

    static void fullUpdate();
    static bool isUpdating() {
        return updates;
    };
    static void setAlwaysUpdate(bool always);

    /**
     * Clears the screen taking into account whether it is double buffered.
     * @param color used to clear the screen.
     */
    static void clearScreen(Uint32 color);

    /**
     * This method indicates we have finish drawing and
     * we want to flip the buffers.
     */
    static void flip(bool full=true);

    /**
      * toggles updates on or off.
      */
    static void toggle_updates(bool on);

    /** Sets the full screen flag. Call set_video_mode after this. */
    static void setFullscreenFlag(bool full);

    /**
     * Gets the time since start of program.
     * @return the amount of miliseconds since program start.
     */
    static unsigned int ticks() {
        return SDL_GetTicks();
    }

    /**
     * Sleeps current thread during the period specified in miliseconds.
     * @param t the time to sleep.
     */
    static void sleep(int t) {
        SDL_Delay(t);
    }

    /**
     * Draws an arrow in screen coordinates.
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     */
    static void drawArrow(int x1,int y1,int x2,int y2);
    static void drawArrow(int x1,int y1,int x2,int y2,Uint32 color);

    /**
     * Draws a circumference.
     * @param r The radius.
     * @param center The center in screen coordinates.
     * @param color The color to use in screen format.
     */
    static void drawCircle(Vector2d center,int r,Uint32 color);
    static void drawRect(Sint16 x1,Sint16 y1,Sint16 x2,Sint16 y2,Uint32 color);
    static void drawFilledRect(Sint16 x1,Sint16 y1,Sint16 x2,Sint16 y2,Uint32 color);

    /**
     * Draws a circumference around the given graphic.
     * @param zoom The scaling factor to use over the graphic size.
     * @param pos The graphic position in screen coordinates.
     * @param color The color to use in screen format.
     */
    void drawSurroundingCircle(Vector2d center,real zoom,Uint32 color);

    /**
     * Draws a cross around the given graphic.
     * @param zoom The scaling factor to use over the graphic size.
     * @param pos The graphic position in screen coordinates.
     * @param color The color to use in screen format.
     */
    void drawSurroundingCross(Vector2d center,real zoom,Uint32 color);

    /**
     * Loads a graphic from a file.
     * The hotspot is taken at center of the image.
     * @param graphicid An identifier of the image to use.
     * @param cache Use a cache for rotating and zomming the image.
     * @param isopaque Don't make the image transparent on black regions.
     */
    Graphic(const std::string & filepath,bool cache=true,bool isopaque=false,bool prerotate=false);

    /**
     * This method draws the graphic on to the screen, rotated over its hotspot and
     * in the given position.
    * @param pos The position where to put the graphic hotspot in screen coordinates.
    * @param zoom The zoom factor in which the viewer is looking at.
    * @param angle How much rotated is the graphic seen from the view point.
     */
    void draw(Vector2d pos,real zoom=1,real angle=0);
    virtual void draw(int x,int y,real zoom=1,real angle=0);
    virtual void draw(int x,int y,SDL_Rect * r);

    /**
     * Returns the graphic width.
     * @return the graphic width.
     */
    int getWidth() const;

    /**
     * Returns the graphic height.
     * @return the graphic height.
     */
    int getHeight() const;

    void setFrame(int i);
    int getImageCount() {
        return images.second;
    };

    /**
     * Tells whether the pixel at (x,y) has the colorkey color.
     */
    bool insideImage(int x,int y) const;

    static SDL_Surface * createSurface(Uint16 w,Uint16 h);

    virtual ~Graphic();
};

class GraphicSeq {
private:
    int sstart,send,remaining,current;
    int sstep;
    bool sloop;
public:
    Graphic * seqg;
    int framestep;
    bool hold;
    GraphicSeq(Graphic * g);
    void setFrameSequence(int start,int end,bool loop,bool forward);
    void stepSeq(int delta);
    void nextFrame();
};

class ZoomRotator {
private:
    RMIMGRef img;
    bool * updated;
    SDL_Surface ** rot;
    float zoom;
    Uint8 ROT_FLAG;
public:
    ZoomRotator(const std::string & id);
    void draw(int x,int y,float angle,float zoom);
    ~ZoomRotator();
};


/** A loader for Fonts. */
class FontLoader : public ResourceLoader<FontRenderer> {
private:
    std::map<std::string,std::string> font_paths;
public:
    FontRenderer * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(FontRenderer * p) {
        delete p;
    };
};

typedef Resource<FontLoader> RMFONTRef;


#define UPDATEbegin {Graphic::toggle_updates(true);
#define UPDATEend Graphic::toggle_updates(false);}
#define UPDATEOFFbegin {{Graphic::toggle_updates(false);
#define UPDATEOFFend Graphic::toggle_updates(true);}}
#define UPDATEOFF(code) UPDATEOFFbegin code; UPDATEOFFend

Uint32 getpixel(SDL_Surface *surface, int x, int y);
bool SDLinsideImage(SDL_Surface * surface,int x,int y);

#endif // GRAPHIC_H
