/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "outline.h"
#include <assert.h>
#include <algorithm>


using namespace std;

#define mod(x,m) ((x)<(m) ? (x) : (x-m))
#define modmod(x,m) ((x)<0 ? (x+m) : mod(x,m))
#define prom_mod(i,j,m) ((i)<=(j) ? ((i)+(j))/2 : mod(((i)+(j)+(m))/2,m))
#define increase(i,pol) ((i)==int((pol).size()-1)?0:(i)+1)
#define decrease(i,pol) ((i)?(i)-1:int((pol).size()-1))


OUTLINE_Polygon::OUTLINE_Polygon(int size) : mysize(size) {
    assert(size>1);
    points=new Point[mysize];
    byangle=new Uint16[mysize];
    angles=new real[mysize];
    internal=new Uint8[(mysize+7)/8];
    memset((void*)internal,0,(mysize+7)/8);
}

OUTLINE_Polygon::~OUTLINE_Polygon() {
    delete[] points;
    delete[] byangle;
    delete[] angles;
    delete[] internal;
}

Uint16 OUTLINE_Polygon::getAngleIndex(real angle) const {
    while (angle<0) angle+=2*M_PI;
    while (angle>=2*M_PI) angle-=2*M_PI;
    return Uint16(trunc(mysize*angle/(2*M_PI)));
}

inline bool between(int x,int y,int z) {
    if (x<=z) return x<y && y<=z;
    else return y<=z || x<y;
}

void OUTLINE_Polygon::compute() {
    // fill angles
    for (int i=1;i<mysize;i++) {
        Point p=points[i]-points[i-1];
        if (p==0) {
            CHERROR<<"poly: ";
            for (int i=0;i<size();i++) CHERROR<<" "<<(*this)[i];
            CHERROR<<ENDL;
            assert(p!=0);
        }
        real a=atan2(p.y,p.x);
        if (a<0) a+=2*M_PI;
        angles[i-1]=a;
    };
    Point p=points[0]-points[mysize-1];
    assert(p!=0);
    real a=atan2(p.y,p.x);
    if (a<0) a+=2*M_PI;
    angles[mysize-1]=a;

    // fill byangle
    int j=0;
    int aindex=getAngleIndex(angles[j]);
    int prev=aindex;
    for (int i=0;i<mysize;i++) {
        while (!between(prev,i,aindex)) {
            j=mod(j+1,mysize);
            prev=aindex;
            aindex=getAngleIndex(angles[j]);
        }
        byangle[i]=j;
    }
};

// Walk.first the observer point,
// Walk.second the wall point.
typedef std::pair<Point,Point> Walk;
typedef std::list<Point> Poly;

Walk find_wall(SDL_Surface*img,const Point & p=Point(0,0)) {
    int x,y;

    for (x=p.x;x<img->w;x++)
        if (SDLinsideImage(img,x,p.y))
            return Walk(Point(x-1,p.y),Point(1,0));

    for (y=p.y+1;y<img->h;y++)
        for (x=0;x<img->w;x++)
            if (SDLinsideImage(img,x,y))
                return Walk(Point(x-1,y),Point(1,0));

    return Walk(Point(p.x-2,p.y),Point());
};

std::pair<Point,Point> follow_wall(SDL_Surface *img,const Walk & walk) {
    Point next=walk.first-walk.second.rot90();
    if (SDLinsideImage(img,next.x,next.y)) // hit a wall
        if (SDLinsideImage(img,next.x+walk.second.x,next.y+walk.second.y)) // not on a corner
            return Walk(walk.first+walk.second,-walk.second.rot90()); // turn -90 degrees
        else return Walk(walk.first,-walk.second.rot90()); // turn -90 degrees
    else if (SDLinsideImage(img,next.x+walk.second.x,next.y+walk.second.y)) // not on a corner
        return Walk(next,walk.second); // go on in the same direction
    else return Walk(next+walk.second,walk.second.rot90()); // turn 90 degrees on a corner
};

bool is_corner(const Graphic & g,const Walk & walk) {
    Point next=walk.first-walk.second.rot90();
    return !g.insideImage(next.x,next.y) && !g.insideImage(next.x+walk.second.x,next.y+walk.second.y);
}

Poly * getOutline(SDL_Surface*img) {
    Poly *l=new Poly();
    Walk init=find_wall(img);
    if (init.first.x==-2) return l;
    Walk current=init;
    l->push_back(current.first+current.second);

    do {
        Walk next=follow_wall(img,current);
        while (next!=init && current.second==next.second) {
            current=next;
            next=follow_wall(img,current);
        }
        if (l->back()!=current.first+current.second) l->push_back(current.first+current.second);
        current=next;

        // While the observer does not walk throug the same point
        // looking at the same wall.
    } while (init!=current);

    if (++l->begin()!=l->end() && l->back()==l->front()) l->pop_back();
    return l;
};


real polygonArea(const Poly & poly) {
    if (poly.empty()) return 0;

    Poly::const_iterator i=poly.begin();
    Point prev=*i++;
    real area=0;
    while (i!=poly.end()) {
        area+=(prev.y+i->y)*(prev.x-i->x)/2.0;
        prev=*i++;
    }
    area+=(prev.y+poly.front().y)*(prev.x-poly.front().x)/2.0;

    return area;
};


class PPoint {
public:
    Point p;
    bool internal;
    PPoint(const Point & p=0,bool internal=false) : p(p), internal(internal) {};
    operator Point() const {
        return p;
    }
};

typedef std::list<PPoint> Poly2;

real polygonArea(const Poly2 & poly) {
    if (poly.empty()) return 0;

    Poly2::const_iterator i=poly.begin();
    Point prev=*i++;
    real area=0;
    while (i!=poly.end()) {
        area+=(prev.y+i->p.y)*(prev.x-i->p.x)/2.0;
        prev=*i++;
    }
    area+=(prev.y+poly.front().p.y)*(prev.x-poly.front().p.x)/2.0;

    return area;
};


template <class List>
typename List::iterator nextit(typename List::iterator i,List & p) {
    if (++i==p.end()) return p.begin();
    else return i;
};

template <class List>
typename List::iterator previt(typename List::iterator i,List & p) {
    if (i==p.begin()) return --p.end();
    else return --i;
};


template <class P>
void polygon_intersect(list<typename P::iterator> & ps,P & nodes,typename P::iterator begin,typename P::iterator end,
                       const Point & p0,const Point & p1) {
    if (begin==end) return;
    typename P::iterator i=begin;
    typename P::iterator prev;
    do {
        prev=i;
        i=nextit(i,nodes);
        if (segment_segment_intersect(*prev,*i,p0,p1)) ps.push_front(prev);
    } while (i!=end);
}


real messure(Poly::iterator i,Poly & nodes) {
    Point prev=i!=nodes.begin() ? (*(--i)++) : nodes.back();
    Point next=++i!=nodes.end() ? (*i) : nodes.front();
    i--;
    Vector2d v=next-prev;
    Vector2d oldsegment=*i-prev;
    if (v==0) return oldsegment.sqrDist();
    real d=dot(v.rot90(),oldsegment);
    if (d) return fabs(d*d/v.sqrDist());
    else if (dot(v,oldsegment)<=0) return oldsegment.sqrDist();
    else if (dot(oldsegment,v-oldsegment)<=0) return (oldsegment-v).sqrDist();
    else return 0;
}

class cmp {
public:
    bool operator()(const pair<real,Poly::iterator> & p0,const pair<real,Poly::iterator> & p1) {
        return p0.first<p1.first;
    }
};

int optimizePolygonVertex(Poly & poly,real threshold) {
    if (poly.empty()) return 0;
    vector<pair<real,Poly::iterator> > nodes(poly.size());
    nodes.clear();
    for (Poly::iterator i=poly.begin();i!=poly.end();i++)
        nodes.push_back(make_pair(messure(i,poly),i));

    //nth_element(nodes.begin(),nodes.begin()+int(threshold),nodes.end(),cmp());
    sort(nodes.begin(),nodes.end(),cmp());

    for (vector<pair<real,Poly::iterator> >::iterator i=nodes.begin();i!=nodes.end() && i->first<=threshold;i++) {
        list<Poly::iterator> ps;
        polygon_intersect(ps,poly,nextit(i->second,poly),previt(i->second,poly),
                          *previt(i->second,poly),*nextit(i->second,poly));
        if (ps.empty()) {
            poly.erase(i->second);
            return 1;
        }
    }
    return 0;
};

void optimize(Poly & poly,real threshold) {
    while (optimizePolygonVertex(poly,threshold));
}

bool segment_segment_intersect(const Point & p0,const Point & p1,const Point & q0,const Point & q1) {
    real s=dot((p1-p0).rot90(),q1-q0);
    if (s) s=dot(p1-p0,(q0-p0).rot90())/s;
    else return false;
    if (0<s && s<1) {
        real t=dot((q1-q0).rot90(),p1-p0);
        if (t) t=dot(q1-q0,(p0-q0).rot90())/t;
        else return false;
        if (0<t && t<1) {
            //n=Vector2d(q0)+Vector2d(q1-q0)*s;
            return true;
        }
    }

    return false;
}


bool line_line_intersect(const Vector2d & p,const Vector2d & dp,
                         const Vector2d & q,const Vector2d & dq,
                         real & s,real & t) {
    s=dot(dp,dq.rot90());
    if (s) {
        real r=dot(q-p,dq.rot90());
        //qpos=r>0?1:(r<0?-1:0);
        s=r/s;
        t=dot(dp.rot90(),dq);
        r=dot(p-q,dp.rot90());
        //    ppos=r>0?1:(r<0?-1:0);
        t=r/t;
    } else {
        if (dot(dp.rot90(),q-p)==0) {
            if (dp.x) s=(q.x-p.x)/dp.x;
            else if (dp.y) s=(q.x-p.x)/dp.y;
            if (dq.x) t=(p.x-q.x)/dq.x;
            else if (dq.y) t=(p.x-q.x)/dq.y;
            else t=0;
            return true;
        } else {
            s=t=0;
            return false;
        }
    }
    return true;
}

bool line_line_intersect(const Vector2d & p0,const Vector2d & d0,const Vector2d & p1,const Vector2d & d1,Vector2d & n) {
    real s=dot(d0.rot90(),d1);
    if (s) n=p1+d1*dot(d0,(p1-p0).rot90())/s;
    return s!=0;
}


Circle find_circle(const list<Vector2d> & border) {
    list<Vector2d>::const_iterator first=border.begin();
    if (first==border.end()) return Circle();
    list<Vector2d>::const_iterator second=first;
    second++;
    if (second==border.end()) return Circle(*first,0);
    list<Vector2d>::const_iterator third=second;
    third++;
    Vector2d a0=Vector2d(*second-*first)/2.0;
    if (third==border.end()) return Circle(a0+*first,a0.dist());

    Vector2d a1=Vector2d(*third-*second)/2.0;
    Vector2d n;

    if ((a1+*second-*first).sqrDist()<=a1.sqrDist())
        return Circle(a1+*second,a1.dist());
    else if ((a0+*first-*third).sqrDist()<=a0.sqrDist())
        return Circle(a0+*first,a0.dist());
    else if ((Vector2d(*first+*third)/2.0-*second).sqrDist()<=(*first-*third).sqrDist()/4.0)
        return Circle(Vector2d(*first+*third)/2.0,(*first-*third).dist()/2.0);

    if (line_line_intersect(a1+*second,a1.rot90(),a0+*first,a0.rot90(),n))
        return Circle(n,(n-*first).dist());
    else {
        CHERROR<<"find_circle: Something impossible happened: "<<*first<<" "<<*second<<" "<<*third<<ENDL;
        exit(1);
    }
}


real circle_intersection_area(const Circle & c0,const Circle & c1) {
    // Test if one circumference surrounds the other.
    real r=fabs(c1.r-c0.r);
    if ((c1.center-c0.center).sqrDist()<=r*r) {
        if (c1.r>c0.r) return c0.area();
        else return c1.area();
    }

    // Test if they intersect at all
    r=c0.r+c1.r;
    if (r*r<=(c1.center-c0.center).sqrDist()) return 0;

    // Intersections
    real cdist=(c1.center-c0.center).dist();
    real xi0=(c1.r*c1.r-c0.r*c0.r-cdist*cdist)/(2*cdist);
    real xi1=(c1.r*c1.r-c0.r*c0.r+cdist*cdist)/(2*cdist);

    real ai0=c0.r*c0.r*(M_PI-asin(fabs(xi0)/c0.r))+fabs(xi0*sqrt(c0.r*c0.r-xi0*xi0));
    real ai1=c1.r*c1.r*(M_PI-asin(fabs(xi1)/c1.r))+fabs(xi1*sqrt(c1.r*c1.r-xi1*xi1));

    if (xi0*xi1>=0) {
        if (c1.r>c0.r) return ai0+c1.area()-ai1;
        else return ai1+c0.area()-ai0;
    } else return c0.area()-ai0+c1.area()-ai1;
}

real messureDivision(Poly2 & nodes,Poly2::iterator pos,Poly2::iterator current) {
    list<Point> component;
    for (Poly2::iterator i=pos;i!=current;i=nextit(i,nodes))
        component.push_back(*i);
    component.push_back(*current);
    Circle c0=minimumDisk(component.begin(),component.end());

    component.clear();
    for (Poly2::iterator i=current;i!=pos;i=nextit(i,nodes))
        component.push_back(*i);
    component.push_back(*pos);
    Circle c1=minimumDisk(component.begin(),component.end());

    real a=c0.area()+c1.area()-circle_intersection_area(c0,c1);
    if (!a) {
        CHERROR<<"Circled area == 0"<<ENDL;
        exit(1);
    } else return polygonArea(nodes)/a;
};

inline bool convex(const Vector2d & p0,const Vector2d & p1,const Vector2d & p2) {
    real r=dot((p1-p0).rot90(),p2-p1);
    if (r>0) return true;
    else if (!r) return dot(p1-p0,p2-p1)<=0;
    else return false;
}

/**
 * How to break a polygon into convex components:
 *    Among all the posible convex subdivisions,
 *    choose that which minimizes the difference of
 *    the sum of the area of minimum enclosing disks and
 *    the area of the original poligon.
 */
Poly2::iterator chooseSubdivision(Poly2::iterator pos,Poly2 & nodes,bool relax) {
    if (nodes.size()<=3) return nodes.end();

    Poly2::iterator maxpos=nodes.end();
    real max=0;

    list<Poly2::iterator> pinters;

    for (Poly2::iterator i=nextit(pos,nodes);i!=pos; i=nextit(i,nodes) ) {

        pinters.clear();
        polygon_intersect(pinters,nodes,nodes.begin(),--nodes.end(),*i,*pos);
        if (pinters.empty()) {
            if (// have a convex angle where there was nonconvexity
                convex(i->p,pos->p,nextit(pos,nodes)->p) &&
                // unless relax, have a convex angle on the remainig component
                (relax || convex(previt(pos,nodes)->p,pos->p,i->p))&&
                // if new edge is colinear with the first edge, it has opposite direction
                (dot((nextit(pos,nodes)->p-pos->p).rot90(),i->p-pos->p) || dot((nextit(pos,nodes)->p-pos->p),i->p-pos->p)<=0)) {

                real temp=messureDivision(nodes,pos,i);
                if (max<temp) {
                    max=temp;
                    maxpos=i;
                }
            }
        }
    };
    if (max) return maxpos;
    else return nodes.end();
}

void convexComponent(Poly2::iterator pos,Poly2 & nodes,Poly2 & newcomponent) {
    if (nodes.size()<=3) 
        return;

    Poly2::iterator maxpos=chooseSubdivision(pos,nodes,false);
    if (maxpos==nodes.end())
        maxpos=chooseSubdivision(pos,nodes,true);

    if (maxpos==nodes.end()) {
        CHERROR<<" polyconvex: ";
        for (Poly2::iterator i=newcomponent.begin();i!=newcomponent.end();i++)
            CHERROR<<"("<<i->p.x<<" "<<i->p.y<<") ";
        CHERROR<<ENDL;

        CHERROR<<"polyconvex: ";
        for (Poly2::iterator i=nodes.begin();i!=nodes.end();i++)
            CHERROR<<"("<<i->p.x<<" "<<i->p.y<<") ";
        CHERROR<<ENDL;
        CHERROR<<"pos: "<<pos->p<<ENDL;;
        CHERROR<<"next pos: "<<nextit(pos,nodes)->p<<ENDL;;

        CHERROR<<"Polygon decomposition not found!"<<ENDL;
        return;
    }

    newcomponent.push_back(*pos);
    pos->internal=true;
    Poly2::iterator j=pos;
    j=nextit(j,nodes);
    while (j!=maxpos) {
        Poly2::iterator k=j;
        j=nextit(j,nodes);
        newcomponent.push_back(*k);
        nodes.erase(k);
    }
    if (maxpos->p!=pos->p) 
        newcomponent.push_back(PPoint(maxpos->p,true));

    if (maxpos->p==pos->p)
        nodes.erase(maxpos);
}

void breakinconvex(list<Poly2*> & ps,const Poly & p) {
    bool done;
    Poly2 * p2=new Poly2();
    for (Poly::const_iterator i=p.begin();i!=p.end();i++) p2->push_back(*i);
    ps.push_back(p2);

    for (list<Poly2*>::iterator j=ps.begin();j!=ps.end();j++) {
        do {
            done=true;
            for (Poly2::iterator i=(*j)->begin();i!=(*j)->end();i++)
                if (!convex(previt(i,**j)->p,i->p,nextit(i,**j)->p)) {
                    Poly2 * t=new Poly2();
                    convexComponent(i,**j,*t);
                    if (!t->empty()) {
                        ps.push_back(t);
                        done=false;
                    } else {
                        delete t;
                        done=true;
                    }
                    break;
                }
        } while (!done);
    }
}

void deleteinternaledges(list<Poly2*> & ps) {
    list<Poly2*>::iterator j=ps.begin();
    while (j!=ps.end()) {
        Poly2::iterator i=(*j)->begin();
        while (i!=(*j)->end()) {
            if (previt(i,**j)->internal && i->internal) {
                Poly2::iterator k=i++;
                (*j)->erase(k);
            } else i++;
        }
        if ((*j)->empty()) {
            list<Poly2*>::iterator k=j++;
            ps.erase(k);
        } else j++;
    }
}


void makePolygons(OUTLINE_Polygons & lp,const list<Poly2*> & lpp) {
    for (list<Poly2*>::const_iterator i=lpp.begin();i!=lpp.end();i++) {
        OUTLINE_Polygon * mkp=new OUTLINE_Polygon((*i)->size());
        int count=0;
        for (Poly2::iterator j=(*i)->begin();j!=(*i)->end();j++) {
            mkp->setInternal(count,j->internal);
            (*mkp)[count++]=j->p;
        }
        mkp->compute();
        lp.push_back(mkp);
    }
};


/*bool testPixelGraphic(void*bmp,int x,int y){
  return ((Graphic*)bmp)->insideImage(x,y);
}*/

OUTLINE_Polygons * OUTLINE_create(SDL_Surface * bmp,real threshold) {
    Poly * pol=getOutline(bmp);
    optimize(*pol,threshold);
    list<Poly2*> ps;
    breakinconvex(ps,*pol);
    deleteinternaledges(ps);
    OUTLINE_Polygons * lp=new OUTLINE_Polygons();
    makePolygons(*lp,ps);
    return lp;
}

/*OUTLINE_Polygons * OUTLINE_create(const Graphic & bmp,real threshold){
  return OUTLINE_create(bmp.getWidth(),bmp.getHeight(),(void*)&bmp,testPixelGraphic,threshold);
};


OUTLINE_Polygons * OUTLINE_create(SDL_Surface * bmp,real threshold){
  return OUTLINE_create(bmp->w,bmp->h,(void*)bmp,(bool(*)(void*,int,int))SDLinsideImage,threshold);
};*/


OUTLINE_Polygon * OUTLINE_PolygonTransformer::getPolygon() const {
    OUTLINE_Polygon * mkp=new OUTLINE_Polygon(size());
    for (int i=0;i<size();i++)
        (*mkp)[i]=(*this)[i];
    mkp->compute();

    return mkp;
};



inline int cmp_angles(real angle0,real angle1) {
    real diff=angle0-angle1;
    if ((0<diff && diff<M_PI) || diff<=-M_PI) return -1;
    else if (diff) return 1;
    else return 0;
}

int search_PolygonAngle(const OUTLINE_PolygonTransformer & pol,real angle,int left,int right) {
    while (left!=right) {
        int mid=prom_mod(left,right,pol.size());
        switch (cmp_angles(pol.getAngle(mid),angle)) {
        case 1:
            left=increase(mid,pol);
            break;
        case 0:
            return mid;
        default:
            right=mid;
        }
    }
    return left;
}

OUTLINE_MKSum::Index OUTLINE_MKSum::pickNode(int i0) const {
    real a=modmod(pol0.getAngle(i0)+angleshift,2*M_PI);
    int angle=pol1.getAngleIndex(a);
    return OUTLINE_MKSum::Index(i0,search_PolygonAngle(pol1,a,pol1.getIndexI(angle),pol1.getIndexI(increase(angle,pol1))));
};

OUTLINE_MKSum::Index OUTLINE_MKSum::pickNode1(int i1) const {
    real a=modmod(pol1.getAngle(i1)-angleshift,2*M_PI);
    int angle=pol0.getAngleIndex(a);
    OUTLINE_PolygonTransformer pt(&pol0);
    return OUTLINE_MKSum::Index(search_PolygonAngle(pt,a,pol0.getIndexI(angle),pol0.getIndexI(increase(angle,pol0))),i1);
};

OUTLINE_MKSum::Index OUTLINE_MKSum::next(const OUTLINE_MKSum::Index & p) const {
    switch (cmp_angles(modmod(pol0.getAngle(p.i0)+angleshift,2*M_PI),pol1.getAngle(p.i1))) {
    case -1:
        return OUTLINE_MKSum::Index(p.i0,increase(p.i1,pol1));
        break;
    case 1:
        return OUTLINE_MKSum::Index(increase(p.i0,pol0),p.i1);
        break;
    default:
        return OUTLINE_MKSum::Index(increase(p.i0,pol0),increase(p.i1,pol1));
    }
}

OUTLINE_MKSum::Index OUTLINE_MKSum::prev(const OUTLINE_MKSum::Index & p) const {
    int np0=decrease(p.i0,pol0);
    int np1=decrease(p.i1,pol1);
    switch (cmp_angles(modmod(pol0.getAngle(np0)+angleshift,2*M_PI),pol1.getAngle(np1))) {
    case -1:
        return OUTLINE_MKSum::Index(np0,p.i1);
        break;
    case 1:
        return OUTLINE_MKSum::Index(p.i0,np1);
        break;
    default:
        return OUTLINE_MKSum::Index(np0,np1);
    }
}

OUTLINE_MKSum::Index OUTLINE_MKSum::average(const OUTLINE_MKSum::Index & p0,const OUTLINE_MKSum::Index & p1) const {
    if (p0.i0==p1.i0) {
        int n=prom_mod(p0.i1,p1.i1,pol1.size());
        if (n==p0.i1) return p0;
        else return pickNode1(n);
    } else {
        int n=prom_mod(p0.i0,p1.i0,pol0.size());
        if (n==p0.i0) {
            OUTLINE_MKSum::Index temp=pickNode(n);
            if (temp!=p0) return temp;
            else {
                temp=next(temp);
                if (temp!=p1) return temp;
                else return p0;
            }
        } else return pickNode(n);
    }
}

class SumEvaluator {
public:
    typedef OUTLINE_MKSum::Index Index;
    const OUTLINE_MKSum & sum;
    SumEvaluator(const OUTLINE_MKSum * sum) : sum(*sum) {}
    Index next(const Index & i) const {
        return sum.next(i);
    }
    Index prev(const Index & i) const {
        return sum.prev(i);
    }
    Index average(const Index & i,const Index & j) const {
        return sum.average(i,j);
    }
};

class PolygonEvaluator {
public:
    typedef int Index;
    const OUTLINE_Polygon & pol;
    PolygonEvaluator(const OUTLINE_Polygon * pol) : pol(*pol) {}
    Index next(const Index & i) const {
        return increase(i,pol);
    }
    Index prev(const Index & i) const {
        return decrease(i,pol);
    }
    Index average(const Index & i,const Index & j) const {
        return prom_mod(i,j,pol.size());
    }
};

class linedist : public SumEvaluator {
public:
    Vector2d p,d;
    linedist(const Vector2d &p,const Vector2d &d,const OUTLINE_MKSum * sum) : SumEvaluator(sum), p(p), d(d) {};
    real operator()(const Index & i) const {
        return dot(sum[i]-p,d.rot90());
    };
};

class linedistP : public PolygonEvaluator {
public:
    Vector2d p,d;
    linedistP(const Vector2d &p,const Vector2d &d,const OUTLINE_Polygon * pol) : PolygonEvaluator(pol), p(p), d(d) {};
    real operator()(const Index & i) const {
        return dot(Vector2d(pol[i]-p),d.rot90());
    };
};

// Returns a node which is on the line or has an edge that touches the line.
template <class Fun>
bool find_midle_0(const Fun & F,typename Fun::Index left,typename Fun::Index right,typename Fun::Index & res) {
    real dleft=F(left);
    real dright=F(right);
    bool ascending;
    if (dright<dleft) ascending=true;
    else if (dleft<dright) ascending=false;
    else ascending=dleft<F(F.next(left));

    typename Fun::Index mid;
    real dmid;
    typename Fun::Index pnext;
    real dnext;
    if (ascending) {
        do {
            mid=F.average(left,right);
            dmid=F(mid);
            pnext=F.next(mid);
            dnext=F(pnext);
            if (dmid<=dnext)
                if (dleft<=dnext) left=pnext;
                else right=mid;
            else if (dnext<0) right=mid;
            else left=pnext;
        } while (left!=right);
    } else {
        do {
            mid=F.average(left,right);
            dmid=F(mid);
            pnext=F.next(mid);
            dnext=F(pnext);
            if (dmid>=dnext)
                if (dnext<=dleft) left=pnext;
                else right=mid;
            else if (0<dnext) right=mid;
            else left=pnext;
        } while (left!=right);
    }
    pnext=F.next(left);
    if (F(left)*F(pnext)<=0) {
        res=left;
        return true;
    }

    return false;
}



bool OUTLINE_MKSum::intersect_line(const Vector2d &p,const Vector2d &d,real & t0,real & t1) const {
    OUTLINE_MKSum::Index left=pickNode(0);
    OUTLINE_MKSum::Index ip0;
    linedist ld(p,d,this);
    if (find_midle_0(ld,left,prev(left),ip0)) {
        left=next(ip0);
        OUTLINE_MKSum::Index ip1;
        if (find_midle_0(ld,left,ip0,ip1)) {
            Vector2d p0=(*this)[ip0];
            OUTLINE_MKSum::Index pnext=next(ip0);
            real s;
            line_line_intersect(p,d,p0,(*this)[pnext]-p0,t0,s);
            Vector2d q0=(*this)[ip1];
            OUTLINE_MKSum::Index qnext=next(ip1);
            line_line_intersect(p,d,q0,(*this)[qnext]-q0,t1,s);
            if (t1<t0) swap(t0,t1);
            return true;
        };
    }
    return false;
}


OUTLINE_Polygon * OUTLINE_MKSum::getPolygon() {
    list<OUTLINE_Polygon::Point> l;
    OUTLINE_MKSum::Index first=pickNode(0);
    OUTLINE_MKSum::Index pnext=first,pprev;
    //int cc=0;
    do {
        l.push_back((*this)[pnext]);
        pprev=pnext;
        pnext=next(pnext);
        assert(pprev.i0==prev(pnext).i0);
        assert(pprev.i1==prev(pnext).i1);
    } while (pnext!=first);
    OUTLINE_Polygon * mkp=new OUTLINE_Polygon(l.size());
    int count=0;
    for (list<OUTLINE_Polygon::Point>::iterator j=l.begin();j!=l.end();j++)
        (*mkp)[count++]=*j;
    mkp->compute();

    return mkp;
}


template <class Function>
OUTLINE_MKSum::Index find_min_cmp_extreme(const Function & F,const OUTLINE_MKSum & pol) {
    OUTLINE_MKSum::Index left=pol.pickNode(0);
    OUTLINE_MKSum::Index right=pol.prev(left);

    bool ascending;
    OUTLINE_MKSum::Index midnext;
    if (F(right,left,pol)==1) ascending=true;
    else if (F(left,right,pol)==1) ascending=false;
    else {
        midnext=pol.next(left);
        if (F(left,midnext,pol)==1) return left;
        else ascending=false;
    }

    /*  OUTLINE_MKSum::Index a=pol.pickNode(0),b=a,c=a,d=a;
      CHERROR<<"pol:";
      do {
        CHERROR<<" "<<pol[b]<<" "<<b.i0<<" "<<b.i1<<"\n";
        b=pol.next(b);
      } while (a!=b);
      CHERROR<<ENDL;*/

    OUTLINE_MKSum::Index mid;
    if (ascending) {
        do {
            mid=pol.average(left,right);
            midnext=pol.next(mid);

            if (F(mid,midnext,pol)>=1)
                if (F(midnext,right,pol)>=1) right=mid;
                else left=midnext;
            else left=midnext;
        } while (left!=right);
    } else {
        do {
            mid=pol.average(left,right);
            midnext=pol.next(mid);

            if (F(mid,midnext,pol)==-1)
                if (F(midnext,right,pol)<=0) right=pol.prev(mid);
                else left=midnext;
            else right=mid;
        } while (left!=right);
    }
    return left;
}

int cmp_angles(const Vector2d & v0,const Vector2d & v1) {
    real d=dot(v0,v1);
    real n=dot(v0,v1.rot90());
    if (d>0)
        if (n>0) return -1;
        else if (n==0) return 0;
        else return 1;
    else if (d==0)
        if (n>0) return -1;
        else if (n==0) return 0;
        else return 1;
    else if (n>0) return -1;
    else if (n==0)
        if (v1.x>=0) return -1;
        else return 1;
    else return 1;
}

template <class Function>
typename Function::Index find_min_extreme(const Function & F,typename Function::Index left,typename Function::Index right) {
    real ini=F(left);
    real end=F(right);
    real pnext;
    bool ascending;
    typename Function::Index midnext;
    if (end<ini) ascending=true;
    else if (ini<end) ascending=false;
    else {
        midnext=F.next(left);
        pnext=F(midnext);

        if (ini<pnext) return left;
        else ascending=false;
    }

    /*  OUTLINE_MKSum::Index a=pol.pickNode(0),b=a,c=a,d=a;
      CHERROR<<"pol:";
      do {
        CHERROR<<" "<<pol[b]<<" "<<b.i0<<" "<<b.i1<<"\n";
        b=pol.next(b);
      } while (a!=b);
      CHERROR<<ENDL;*/

    typename Function::Index mid;
    if (ascending) {
        do {
            mid=F.average(left,right);
            midnext=F.next(mid);
            real pmid=F(mid);
            real pmidnext=F(midnext);

            if (pmid<=pmidnext)
                if (pmidnext<=end) right=mid;
                else left=midnext;
            else left=midnext;
        } while (left!=right);
    } else {
        do {
            mid=F.average(left,right);
            midnext=F.next(mid);
            real pmid=F(mid);
            real pmidnext=F(midnext);

            if (pmid>pmidnext)
                if (pmidnext>=end) right=F.prev(mid);
                else left=midnext;
            else right=mid;
        } while (left!=right);
    }
    return left;
}


template <class Function>
typename Function::Index simple_find_max(const Function & F,typename Function::Index left,typename Function::Index right) {

    while (left!=right) {
        typename Function::Index mid=F.average(left,right);
        typename Function::Index midnext=F.next(mid);
        real pmid=F(mid);
        real pmidnext=F(midnext);
        if (pmid<=pmidnext) left=midnext;
        else right=mid;
    }
    return left;
}

template <class Function>
typename Function::Index simple_find_min(const Function & F,typename Function::Index left,typename Function::Index right) {

    while (left!=right) {
        typename Function::Index mid=F.average(left,right);
        typename Function::Index midnext=F.next(mid);
        real pmid=F(mid);
        real pmidnext=F(midnext);
        if (pmid>=pmidnext) left=midnext;
        else right=mid;
    }
    return left;
}


class pointdist : public SumEvaluator {
public:
    Vector2d p;
    pointdist(const Vector2d & p,const OUTLINE_MKSum * sum) : SumEvaluator(sum), p(p) {};
    real operator()(const Index & i) const {
        Vector2d p0=sum[i];
        Vector2d e=sum[sum.next(i)]-p0;
        real d=e.sqrDist();
        real sin=dot(e.rot90(),p-p0);
        return (sin<0?-1:1)*sin*sin/d;
    }
};

class pointdistP : public PolygonEvaluator {
public:
    Vector2d p;
    pointdistP(const Vector2d & p,const OUTLINE_Polygon * pol) : PolygonEvaluator(pol), p(p) {};
    real operator()(const Index & i) const {
        Vector2d p0=pol[i];
        Vector2d e=pol[next(i)]-p0;
        real d=e.sqrDist();
        real sin=dot(e.rot90(),p-p0);
        return (sin<0?-1:1)*sin*sin/d;
    }
};

template <class Function>
inline void find_extremes(const Function & F,
                          const typename Function::Index &left,const typename Function::Index &right,
                          typename Function::Index &ext0,typename Function::Index &ext1) {
    ext0=find_min_extreme(F,left,right);
    ext1=simple_find_max(F,F.next(ext0),ext0);
}


bool OUTLINE_MKSum::closest_inside(const Vector2d & p,OUTLINE_MKSum::Index & ext0) const {
    OUTLINE_MKSum::Index ext1,first=pickNode(0),second=next(first);
    Vector2d p0=(*this)[first];
    Vector2d p1=(*this)[second];
    if (p1==p) return true;

    linedist ld(p,p0-p,this);
    find_extremes(ld,second,first,ext0,ext1);
    p0=(*this)[ext0];
    p1=(*this)[ext1];
    if (dot((p1-p0).rot90(),p-p0)>0)
        ext0=simple_find_min(pointdist(p,this),ext1,prev(ext0));
    else ext0=simple_find_min(pointdist(p,this),ext0,prev(ext1));

    p0=(*this)[ext0];
    p1=(*this)[next(ext0)];
    return dot((p1-p0).rot90(),p-p0)>=0;
}

bool OUTLINE_Polygon::closest_inside(const Vector2d & p,int & ext0) const {
    int ext1,first=0,second=increase(first,*this);
    Vector2d p0=(*this)[first];
    Vector2d p1=(*this)[second];
    if (p1==p) return true;

    linedistP ld(p,p0-p,this);
    find_extremes(ld,second,first,ext0,ext1);
    p0=(*this)[ext0];
    p1=(*this)[ext1];
    if (dot((p1-p0).rot90(),p-p0)>0)
        ext0=simple_find_min(pointdistP(p,this),ext1,decrease(ext0,*this));
    else ext0=simple_find_min(pointdistP(p,this),ext0,decrease(ext1,*this));

    p0=(*this)[ext0];
    p1=(*this)[increase(ext0,*this)];
    return dot((p1-p0).rot90(),p-p0)>=0;
}


bool OUTLINE_MKSum::inside(const Vector2d & p) const {
    OUTLINE_MKSum::Index e;
    return closest_inside(p,e);
}

typedef std::vector<PPoint> Polygon;

void Poly2Polygon(const Poly2 & pin,Polygon & pout) {
    pout.resize(pin.size());
    int count=0;
    for (Poly2::const_iterator i=pin.begin();i!=pin.end();i++)
        pout[count++]=*i;
}

ostream & operator << (ostream & o,const Circle & c) {
    o<<"c "<<c.center<<" "<<c.r;
    return o;
};

ostream & operator << (ostream & o,const OUTLINE_Polygon & p) {
    o<<"p "<<p.size();
    for (int i=0;i<p.size();i++)
        o<<" "<<p[i]<<" "<<p.isInternal(i);
    return o;
};

ostream & operator << (ostream & o,const OUTLINE_Polygons & ps) {
    for (OUTLINE_Polygons::const_iterator i=ps.begin();i!=ps.end();i++) {
        o<<**i<<"\n";
        o<<minimumDisk((*i)->begin(),(*i)->end())<<"\n";
    }
    return o;
};

Lexer & OUTLINE_parseCircle(Lexer & p,Circle * c) {
    PARSEbegin(Lexer,p);
    LS(L"c") LVECTOR(&c->center) LFLOAT(&c->r);
    PARSEend;
    return p;
};

Lexer & OUTLINE_parsePolygon(Lexer & p,OUTLINE_Polygon ** pol) {
    PARSEbegin(Lexer,p);
    Vector2d v;
    int i;
    LS(L"p");
    LINT(&i);
    *pol=new OUTLINE_Polygon(i);
    int count=0;
    MANYbegin;
    LVECTOR(&v);
    LINT(&i);
    (**pol)[count]=v;
    (*pol)->setInternal(count++,i);
    MANYend;
    (*pol)->compute();
    PARSEend;
    return p;
};



/*

template <class Function>
OUTLINE_MKSum::Index find_min_extreme(const Function & F,const OUTLINE_MKSum & pol){
  OUTLINE_MKSum::Index left=pol.pickNode(0);

  real ini=F(left,pol);
  OUTLINE_MKSum::Index right=pol.prev(left);

  real end=F(right,pol);
  real pnext;
  bool ascending;
  OUTLINE_MKSum::Index midnext;
  if (end<ini) ascending=true;
  else if (ini<end) ascending=false;
  else {
    midnext=pol.next(left);
    pnext=F(midnext,pol);

    if (ini<pnext) return left;
    else ascending=false;
  }

  OUTLINE_MKSum::Index a=pol.pickNode(0),b=a,c=a,d=a;
  CHERROR<<"pol:";
  do {
    CHERROR<<" "<<pol[b]<<" "<<b.i0<<" "<<b.i1<<"\n";
    b=pol.next(b);
  } while (a!=b);
  CHERROR<<ENDL;

  OUTLINE_MKSum::Index mid;
  if (ascending) {
    do {
      mid=pol.average(left,right);
      midnext=pol.next(mid);
      real pmid=F(mid,pol);
      real pmidnext=F(midnext,pol);

      if (pmid<=pmidnext)
	if (pmidnext<=end) right=mid;
        else left=midnext;
      else left=midnext;
    } while(left!=right);
  } else {
    do {
      mid=pol.average(left,right);
      midnext=pol.next(mid);
      real pmid=F(mid,pol);
      real pmidnext=F(midnext,pol);

      if (pmid>pmidnext)
	if (pmidnext>=end) right=pol.prev(mid);
        else left=midnext;
      else right=mid;
    } while(left!=right);
  }
  return left;
}

class pointangle {
public:
  Vector2d p;
  mutable real min;
  mutable Vector2d p1,p0;
  pointangle(const Vector2d & p) : p(p), min(0) {}
  real operator ()(const OUTLINE_MKSum::Index & i,const OUTLINE_MKSum &pol) const {
    p1=pol[pol.next(i)];
    p0=pol[i];
    Vector2d e=p1-p0;
    Vector2d v=p-p0;
    // Calculate tangent as: dot(v,e.rot90())/dot(v,e)
    real d=dot(v,e);
    if (d<0) // change sign of tangent
      d*=-1;
    real n=dot(v,e.rot90());
    if (d==0)
      if (n>0) return numeric_limits<real>::infinity();
      else if (n==0) {
	min=-numeric_limits<real>::infinity();
	if (v==0) return -numeric_limits<real>::infinity();
	else {
	  assert(false);
	  return 0;
	}
      } else { // stop search
	min=-numeric_limits<real>::infinity();
	return -numeric_limits<real>::infinity();
      }

    n/=d;
    if (n<min) min=n;
    return n;
  };
  bool stop() const { return false; };
};
*/
