/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef OUTLINE_H
#define OUTLINE_H

#include "graphic.h"
#include "SDL.h"
#include "vector2d.h"
#include <list>
#include "debugmsg.h"
#include "parsercombinators.h"

typedef vector2d<Sint16> Point;


/**
 * This class is used to represent convex polygons.
 */
class OUTLINE_Polygon {
public:
    typedef Vector2d Point;
    typedef Vector2d * iterator;
    OUTLINE_Polygon(int size);
    ~OUTLINE_Polygon();

    /** @returns the amount of vertexes of the polygon. */
    Uint16 size() const {
        return mysize;
    }

    /**
     * @param i index of the vertex to retrieve.
     * @returns a reference to the ith vertex.
     */
    Point & operator[] (Uint16 index) const {
        return points[index];
    }

    /**
     * Call this to fill in internal data structures when you have filled
     * the polygon with all its vertexes.
     */
    void compute();

    /** @returns the iterator pointing at the first node of the polygon. */
    iterator begin() const {
        return points;
    }

    /** @returns the iterator pointing past the last node of the polygon. */
    iterator end() const {
        return points+mysize;
    }

    /** @returns the first vertex of the polygon. */
    Point & front() const {
        return points[0];
    }

    /** getAngle(i) is the angle of the edge starting at vertex i w.r.t. the x axis. */
    real getAngle(Uint16 index) const {
        return angles[index];
    }

    /** getIndex(a)==i iff getAngle[i]>=a and getAngle(i-1)<a. */
    Uint16 getIndex(real angle) const {
        return byangle[getAngleIndex(angle)];
    }

    /** getIndexI(a)==getIndex(a*(2*M_PI)/size()). */
    Uint16 getIndexI(Uint16 angleindex) const {
        return byangle[angleindex];
    }

    /**  getAngleIndex(a)==j iff  a>=j*(2*M_PI)/size() and  (j+1)*(2*M_PI)/size()>a */
    Uint16 getAngleIndex(real angle) const;

    /** isInternal(i) iff the edge starting at vertex i is internal. */
    bool isInternal(Uint16 v) const {
        Uint8 b=internal[v/8];
        return (b>>(v%8)) & 0x0001;
    };

    /** 
     * Perhaps the internal trait should be somewhere else. An edge is internal 
     * if it is shared by more than one polygon in the outline of an object. 
     * */
    void setInternal(Uint16 v,bool i) {
        Uint8 b = i ? internal[v/8] | (1<<(v%8)) : (internal[v/8] & ~(1<<v%8));
        internal[v/8]=b;
    };

    bool closest_inside(const Vector2d & p,int & ext0) const;

private:
    /** amount of vertexes of the polygon */
    Uint16 mysize;
    /** array with the vertexes of the polygon, they are listed in counter-clockwise order. */
    Point * points;
    /** angles[i] is the angle of the edge starting at vertex i w.r.t. the x axis. */
    real * angles;
    /** byangle[a]==i  iff angle[i]>=a and angle[i-1]<a. */
    Uint16 * byangle;
    /** internal[i] == 1 iff the edge starting at vertex i is internal to the polygon. */
    Uint8 * internal;
};

typedef std::list<OUTLINE_Polygon*> OUTLINE_Polygons;

/**
 * Creates an outline polygon from a bitmap. It is a set of adjacent convex polygons
 * that more or less cover the opaque pixels of the bitmap.
 * @param g object containing the bitmap to outline.
 * @param threshold used to round a little the outline. Reasonable values are between 0 and 10,
 *        from not rounded to fully rounded.
 * @ returns the list of convex polygons representing the outline.
 */
//OUTLINE_Polygons * OUTLINE_create(const Graphic & g,real threshold);
OUTLINE_Polygons * OUTLINE_create(SDL_Surface * bmp,real threshold);


/**
 * A wrapper for a transformed polygon. It transforms the polygon on the fly
 * as it is traversed. Thus, all operations should be O(1) in time.
 * This is the base class of a hierarchy. It applies the identity
 * transformation.
 * Subclass this class to override the behaviors.
 */
class OUTLINE_PolygonTransformer {
protected:
    const OUTLINE_Polygon*pol;
public:
    OUTLINE_PolygonTransformer(const OUTLINE_Polygon*pol): pol(pol) {};
    virtual Uint16 getIndex(real angle) const {
        return pol->getIndex(angle);
    };
    virtual real getAngle(Uint16 index) const {
        return pol->getAngle(index);
    };
    virtual Uint16 getAngleIndex(real angle) const {
        return pol->getAngleIndex(angle);
    };
    virtual Vector2d operator[] (Uint16 index) const {
        return (*pol)[index];
    };
    Uint16 getIndexI(Uint16 angleindex) const {
        return pol->getIndexI(angleindex);
    }
    virtual ~OUTLINE_PolygonTransformer() {};
    Uint16 size() const {
        return pol->size();
    }
    OUTLINE_Polygon * getPolygon() const;
};


/**
 * Writes an OUTLINE_Polygon to an ostream.
 */
std::ostream & operator << (std::ostream & o,const OUTLINE_Polygons & ps);

Lexer & OUTLINE_parsePolygon(Lexer & p,OUTLINE_Polygon ** pol);


/**
 * A transformer which returns the point symetrical image of the polygon through the
 * origin.
 */
class OUTLINE_PolygonReverser : public OUTLINE_PolygonTransformer {
private:
    Matrix2x3 m;
public:
    OUTLINE_PolygonReverser(const Matrix2x3 & m,const OUTLINE_Polygon*pol) : OUTLINE_PolygonTransformer(pol), m(m) {};
    Uint16 getIndex(real angle) const {
        return pol->getIndex(M_PI+angle);
    };
    real getAngle(Uint16 index) const {
        real angle=M_PI+pol->getAngle(index);
        if (angle<2*M_PI) return angle;
        else return angle-2*M_PI;
    };
    Uint16 getAngleIndex(real angle) const {
        return pol->getAngleIndex(M_PI+angle);
    };
    Vector2d operator[] (Uint16 index) const {
        return -(m*(*pol)[index]);
    };
    void print() {
        CHERROR<<"reverser:"<<ENDL;
        for (int i=0;i<size();i++) {
            CHERROR<<i<<" "<<(*this)[i]<<" "<<getAngle(i)<<" "<<pol->getAngle(i)<<" "<<getIndex(getAngle(i))<<ENDL;
        }
    }
};

/**
 * Minkowski sum of polygons.
 */
class OUTLINE_MKSum {
public:
    class Index {
    public:
        Uint16 i0,i1;
        Index(Uint16 i0,Uint16 i1) : i0(i0), i1(i1) {};
        Index() : i0(0), i1(0) {}
        bool operator != (const Index & p) const {
            return i0!=p.i0 || i1!=p.i1;
        }
        bool operator == (const Index & p) const {
            return i0==p.i0 && i1==p.i1;
        }
    };


    OUTLINE_MKSum(const OUTLINE_Polygon * pol0,real angleshift,const OUTLINE_PolygonTransformer * pol1) :
            pol0(*pol0),pol1(*pol1),angleshift(angleshift) {};

    Index next(const Index & p) const;
    Index prev(const Index & p) const;
    Index average(const Index & p0,const Index & p1) const;
    Vector2d operator[] (const Index & p) const {
        return Vector2d(pol0[p.i0])+pol1[p.i1];
    };
    Index pickNode(int i0) const;
    Index pickNode1(int i1) const;

    /**
     * @returns true iff p falls inside the sum.
     */
    bool inside(const Vector2d & p) const;

    /**
     * @param e the closest edge when p is outside the polygon.
     * @returns true iff p falls inside the sum.
     */
    bool closest_inside(const Vector2d & p,Index & e) const;

    /**
     * @returns true if the line intersects the sum.
     */
    bool intersect_line(const Vector2d &p,const Vector2d &d,real & t0,real & t1) const;

    OUTLINE_Polygon * getPolygon();
private:
    const OUTLINE_Polygon & pol0;
    const OUTLINE_PolygonTransformer & pol1;
    real angleshift;
};


/**
 * A class for representing circles.
 */
class Circle {
public:
    real r;
    Vector2d center;
    Circle() {
        r=0;
    }
    Circle(Vector2d center,real r) : r(r), center(center) {}
    template <class T>
    bool contains(const vector2d<T> & p) const {
        return (center-p).sqrDist()<=r*r;
    };
    real area() const {
        return r*r*M_PI;
    }
};

/**
 * @returns true if the two lines intersect.
 */
bool line_line_intersect(const Vector2d & p,const Vector2d & dp,const Vector2d & q,const Vector2d & dq,
                         real & s,real & t);


Circle find_circle(const std::list<Vector2d> & border);

template <class Iterator>
Circle minimumDisk(Iterator begin,Iterator end) {
    std::list<Vector2d> border;
    return minimumDiskWithBorder(begin,end,border);
};

template <class Iterator>
Circle minimumDiskWithBorder(Iterator begin,Iterator end,std::list<Vector2d> & border) {
    if (begin==end || border.size()==3)
        return find_circle(border);
    else {
        Iterator second=begin;
        second++;
        Circle c=minimumDiskWithBorder(second,end,border);
        if (c.contains(*begin)) return c;

        border.push_front(*begin);
        c=minimumDiskWithBorder(second,end,border);
        border.pop_front();
        return c;
    }
};

Lexer & OUTLINE_parseCircle(Lexer & p,Circle * c);


#endif  //  OUTLINE_H
