/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * This file contains a C/C++ library of parser compbinators inspired in the Parsec library
 * written in Haskell.
 */

#ifndef PARSERCOMBINATORS_H
#define PARSERCOMBINATORS_H

#include <iosfwd>
#include <string>
#include <list>
#include "reference.h"
#include "global.h"
#include "filecoding.h"

/** 
 * Interface to read characters from. 
 * Subclass this to enable parsing files, strings, streams, etc.  
 * */
class CharTokenizer {
public:
    virtual wchar_t readNext()=0;
    virtual bool eof()=0;
    virtual bool error()=0;
    virtual ~CharTokenizer() {};
};

/**
 * Interface to read tokens from.
 * This is the interface to a lexicographic analyzer. 
 */
class Tokenizer {
public:
    typedef struct {
        std::basic_string<wchar_t> text;
        int code;
        int line;
        int column;
    } TokenST;
    virtual bool getNext(TokenST*)=0;
    virtual bool eof()=0;
    virtual bool error()=0;
    virtual ~Tokenizer() {};
};

/** 
 * Implementation of a parser.
 * The library user must instantiate this class with some tokenizer,
 * but must use the macros provided below when parsing instead of directly
 * calling the methods in this class.
 */
class Parser {
private:
    bool errorflag;
    int eofCount;
    Tokenizer * tok;
    struct ParserST {
        int tokCount;
        std::list<std::basic_string<wchar_t> > * expected;
    };
    ParserST st;
    Tokenizer::TokenST tokst;
    std::list<std::list<Tokenizer::TokenST> *> recorded_tokens;
    typedef std::list<Tokenizer::TokenST > KeptTokens;
    std::list<std::pair<KeptTokens*,ParserST> > keep_tokens;
    void clear();
public:

    Parser(Tokenizer * tok);
    Parser & operator = (Tokenizer * tok);
    virtual ~Parser();
    void getNextToken();
    Parser & readstring(std::basic_string<wchar_t> seq);
    Parser & readstring(const char * seq);
    Parser & readChar(wchar_t c);
    Parser & readAnyToken(int * code,std::basic_string<wchar_t> * text);
    Parser & readToken(int code,std::basic_string<wchar_t> * text);
    Parser & readToken(int code,std::string * text);
    Parser & readAny(wchar_t * c);
    Parser & expecting(const std::basic_string<wchar_t> & desc);
    Parser & expecting(const char * desc);

    // Error handling

    /**
     * @returns true if a parsing error has occurred.
     */
    bool error();

    /**
     * Signals that an error has occurred.
     */
    void setError();
    /**
     * Tells an error state has been handled.
     * It must be called just when this->error().
     */
    void clearError();
    /**
     * A description of the error.
     * It must be called just when this->error().
     */
    std::string errorMessage();

    // State Stack operations

    /**
     * Saves the state of the parser on a state stack.
     * This includes also the current input.
     */
    void pushState();
    /**
     * Restores the topmost state on the stack, and pops it.
     */
    void restoreState();
    /**
     * Pops the topmost state on the stack.
     */
    void popState();

    Parser & eof();
    bool reachedEof();
    long getTokenCount() {
        return st.tokCount;
    };
    wchar_t getCurrentChar() {
        return (tokst.text.length()>0?tokst.text[0]:'\0');
    };
};

/** Enclose your parsing code with PARSEbegin(Parser,p); PARSEend; */
#define PARSEbegin(Parser,p) {Parser * __parsercomb_p=&p;switch(0) {default:ENDBLOCK;
#define PARSEend }}

/** A shorthand for enclosing parsing code. */
#define PARSE(Parser,p,parse) PARSEbegin(Parser,p) parse PARSEend

/** 
 * If a parsing code is enclosed in OPTbegin; OPTend; the resulting code will behave the same.
 * Additional OR macros can be appended as separators of different parsing codes:
 * OPTbegin;
 *   p. code 1
 * OR
 *   p. code 2
 * OR
 *   ...
 * OPTend;
 * If p. code i succeds the whole OPTbegin;OPTend block succeds.
 * If p. code i fails without consuming any input, then p. code i+1
 * is attempted.
 * If p. code i fails consuming input then the whole OPTbegin;OPTend; block
 * fails.
 */
#define OPTbegin \
   {int __parsercomb_tc=__parsercomb_p->getTokenCount();\
      switch(0) {default: {{{{switch(0) {default:
#define OPTend }}} }}}; if (__parsercomb_p->error()) break;}
#define OR }}}}};\
    if (!__parsercomb_p->error() || __parsercomb_tc!=__parsercomb_p->getTokenCount()) \
      break;\
    else switch(0) {default: {{{{ __parsercomb_p->clearError();
/** Shorthand for OPTbegin; OPTend;. */
#define OPT(parse) OPTbegin parse  OPTend


/** 
 * If a parsing code is enclosed with MANYbegin(Parser,p); MANYend; 
 * This will cause the parsing code to be run repeteadly until it fails.
 * If the parsing code fails during the last repetition without consuming
 * input, then the parsing code "MANYbegin parsing code MANYend" succeeds.
 * If the parsing code fails during the last repetition consuming
 * input, then the parsing code "MANYbegin parsing code MANYend" fails.
 * */
#define MANYbegin \
  {{{ int __parsercomb_tc; \
    do { __parsercomb_tc=__parsercomb_p->getTokenCount();
#define MANYend \
      } while(true); \
    if (__parsercomb_tc!=__parsercomb_p->getTokenCount()) \
        break;\
    else \
        __parsercomb_p->clearError();\
  }}}

/** Shorthand for enclosing parsing code with MANYbegin; MANYend; */
#define MANY(parse) MANYbegin parse MANYend

/** Same than MANY but the enclosed parsing code must succeed at least 
 * once for MANY1 to succeed.
 */
#define MANY1(parse) {parse;MANY(parse)}

/**
 * PATTTEXT(&str) parses text enclosed in double quotes (") and writes the
 * text into str.
 * */
#define PATTTEXT(str) RT(TOKATTTEXT,str)

/** A parser that succeds without consuming any input. */
#define POK
/** 
 * A parser that reads a whitespace if any. Depends on the Tokenizer 
 * not filtering whitespaces. 
 */
#define PWS OPT(S(" ") OR POK)

/** 
 * TRYbegin p. code TRYend succeds if p. code succeds, and fails without consuming input
 * if p. code fails (consuming input or not).
 * It works somewhat like restoring the input state prior to parsing if p. code fails.
 */
#define TRYbegin \
  {{{__parsercomb_p->pushState();\
switch(0) {default:
#define TRYend }}}} \
  if (__parsercomb_p->error()) {__parsercomb_p->restoreState();break;} else __parsercomb_p->popState();
#define TRY(parse) TRYbegin parse TRYend

/** 
 * Auxiliary method to call methods in a parser instance. The methods must return a 
 * reference to the parser instance.
 */
#define PO(parse) {if (__parsercomb_p->parse.error()) break;}
#define P(parse) {if (parse(*__parsercomb_p).error()) break;}
#define P1(parse,x) {if (parse(*__parsercomb_p,x).error()) break;}

/** Parser one or more tokens matching the string str. */
#define S(str) {T(str)PO(readstring(str))}

/** Like S but consumes whitespaces afterwards. */
#define LS(str) {T(str)PO(readstring(str))WS}

/** RTK(k,&str) matches a token k, and reads token data into str. */
#define RT(k,str) PO(readToken(k,str))

/** Succeeds if the end of input has been reached. */
#define PEOF PO(eof())

/** Succeeds if the input is at an end of line. */
#define PEOLN CHAR('\n')

/** A parser that always fails. */
#define PERROR {__parsercomb_p->setError();break;}

/** ID(&str) succeeds on identifier-like strings and writes the identifier to str. */
#define ID(str) { INPUT(CHARCHECK(iswalpha)MANY(OPT(CHARCHECK(iswalnum) OR CHAR(L':') ) ),str) }
/** Same as above but consumes whitespaces afterwards. */
#define LID(str) {ID(str)WS}
/** T(str) p. code  causes p. code to report that it was expecting str when it fails.  */
#define T(str) __parsercomb_p->expecting(str);
/** NOTFOLLOWBY(p. code) succeds if p. code fails, and fails otherwise. */
#define NOTFOLLOWBY(p) OPT(TRY(p) PERROR OR POK)
/** Matches a given character. */
#define CHAR(c) PO(readChar(btowc(c)))
/** 
 * CHARCHECK(f) evaluates function f over the next character of the input and succeds if
 * f returns true, otherwise it fails without consuming any input.
 */
#define CHARCHECK(c) \
   {if (!c(__parsercomb_p->getCurrentChar())) PERROR \
else PO(readAny(NULL))}

/** Matches any character appearing in str. */
#define ONEOF(str) PO(oneOf(str))

/** Matches any character not appearing in str. */
#define NOTONEOF(str) PO(notOneOf(str))

/**
 * ATTTEXT(&str) parses text enclosed in double quotes (") and writes the
 * text into str.
 * */
#define ATTTEXT(str) {CHAR('\"')INPUT(MANY(NOTONEOF("\"")),str)CHAR('\"')}
/** Same as above but reads one or more whitespaces afterwards. */
#define LATTTEXT(str) {CHAR('\"')INPUT(MANY(NOTONEOF("\"")),str)CHAR('\"')WS}

/** 
 * INPUTbegin; p. code INPUTend(&str); records the input consumed by p. code. It succeeds if 
 * p. code succeeds. In such a case the input read is written into str.
 * */
#define INPUTbegin {__parsercomb_p->pushState();switch(0) { default:
#define INPUTend(str) } \
   if (__parsercomb_p->error()) { \
     __parsercomb_p->popState();\
     break; \
   } else { \
     *str=__parsercomb_p->getString();\
     __parsercomb_p->popState();\
   }}
/** Shorthand for INPUTbegin; INPUTend. */
#define INPUT(p,str) INPUTbegin p INPUTend(str)

/** Matches one ore more whitespaces. */
#define WS1 MANY1(CHARCHECK(iswspace));
/** Matches zero ore more whitespaces. */
#define WS MANY(CHARCHECK(iswspace));

/** Reads a number into string str. */
#define LNUM(str) {INPUT(MANY1(CHARCHECK(iswdigit)),str)WS}
/** Internal macro that checks whether the parser has failed or not. */
#define ENDBLOCK {if (__parsercomb_p->error()) break;}
/** Internal macro that checks whether a parser has failed or not. */
#define CALL(p) { if (p.error()) break; }

/** LFLOAT(&f) reads a float value from the input and whitespaces afterwards. */
#define LFLOAT(f) {T("real number")PO(readFloat(f))WS}
/** LINT(&i) reads an int value from the input and whitespaces afterwards. */
#define LINT(n) {T("integer number")PO(readInt(n))WS}
/** LVECTOR(&p) reads a pair of float values from the input and whitespaces afterwards. */
#define LVECTOR(p) {CHAR('(')WS;LFLOAT(&(p)->x)CHAR(',')WS;LFLOAT(&(p)->y)CHAR(')')WS}

/** A parser specialized for reading character tokens. */
class Lexer {
public:
    struct LexerST {
        int line, column,charCount;
        std::list<std::basic_string<wchar_t> > * expected;
    };
    LexerST st;
private:
    bool errorflag;
    int eofCount;
    CharTokenizer * tok;
    wchar_t currentChar;
    std::list<std::basic_string<wchar_t> > recorded_tokens;
    std::list<std::pair<std::basic_string<wchar_t>,LexerST > > keep_tokens;
    void clear();
public:

    Lexer(CharTokenizer * tok);
    Lexer & operator = (CharTokenizer * tok);
    virtual ~Lexer();
    wchar_t getNextToken();
    Lexer & readSequence(std::basic_string<wchar_t> seq);
    Lexer & readChar(wchar_t c);
    Lexer & readAny(wchar_t * c);
    Lexer & oneOf(const wchar_t * cs);
    Lexer & oneOf(const char * cs);
    Lexer & notOneOf(const wchar_t * cs);
    Lexer & notOneOf(const char * cs);
    Lexer & readstring(const wchar_t * cs);
    Lexer & readstring(const char * cs);
    Lexer & expecting(const std::basic_string<wchar_t> & desc);
    Lexer & expecting(const char * desc);
    bool reachedEof();
    Lexer & readFloat(double * d);
    Lexer & readInt(int * d);

    // Error handling

    /**
     * @returns true if a parsing error has occurred.
     */
    bool error();

    /**
     * Signals that an error has occurred.
     */
    void setError();
    /**
     * Tells an error state has been handled.
     * It must be called just when this->error().
     */
    void clearError();
    /**
     * A description of the error.
     * It must be called just when this->error().
     */
    std::string errorMessage();

    // State Stack operations

    /**
     * Saves the state of the parser on a state stack.
     * This includes also the current input.
     */
    void pushState();
    /**
     * Restores the topmost state on the stack, and pops it.
     */
    void restoreState();
    /**
     * Pops the topmost state on the stack.
     */
    void popState();

    Lexer & eof();

    std::basic_string<wchar_t> getString();
    wchar_t getCurrentChar() {
        return currentChar;
    };

    int getTokenCount() {
        return st.charCount;
    };
};

/** Consumes characters from input until finding an end of line which is not consumed. */
#define NEOLN MANY(NOTONEOF(basic_string<wchar_t>(stows("\n")).c_str()))

/** Provides input from a stream. */
class StreamTokenizer : public CharTokenizer {
private:
Filler * f;
IStreamUTF8Decoder sc;
public:
StreamTokenizer(std::istream & i);
StreamTokenizer(FILE * i);
~StreamTokenizer();
/** if eof() then error
 *    else next character
 */
wchar_t readNext();
bool eof();
bool error();
};


#define TOKATTTEXT 1
#define TOKUNKNOWN 0
#define TOKID 2

/** Provides tokens for a stripped xml-like format. */
class ConfigTokenizer : public Tokenizer {
private:
bool _error;
Lexer l;
public:
ConfigTokenizer(CharTokenizer*);
bool getNext(TokenST*);
bool eof();
bool error();
};


#endif // PARSERCOMBINATORS_H
