/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "path.h"

#include <sge.h>
#include "graphic.h"
#include "clipping.h"
#include <iostream>

using namespace std;

real Path::addThreshold=25;
real Path::deleteThreshold=0.8;

Path::Path() : numNodes(0) {
    drawphase=0;
    laststep=Graphic::ticks();
}

void Path::clear() {
    nodes.clear();
    numNodes=0;
    distancebefore=0;
};

template <class Iterator,class PopOp,class PushOp>
bool Path::addNode(Vector2d n,Iterator i,Iterator limit,PopOp pop,PushOp push) {
    bool replaced=false;
    if (i!=limit) {
        // Calculate angle and distance of original segment with respect to
        // the resulting segment of poping the last node and adding the
        // one passed as parameter.
        if (++i!=limit) {
            Vector2d prelast=*i;
            Vector2d lastsegmentOrig=original-prelast;
            Vector2d newsegment=n-prelast;
            // Calculate cos of the angle * lastsegmentOrig.dist() * newsegment.dist()
            real sinangle= dot(newsegment.rot90(),lastsegmentOrig);
            // Discard last node if difference is small enough.
            if (lastsegmentOrig!=Vector2d() && (sinangle==0 || sinangle*sinangle<addThreshold*lastsegmentOrig.sqrDist())) {
                if (++i==limit) 
                    distancebefore=(n-nodes.front()).sqrDist();
                (nodes.*pop)();
                numNodes--;
                replaced=true;
            } else original = n;
        } else {
            original = n;
            distancebefore=(n-nodes.front()).sqrDist();
        }
    } else original=n;
    (nodes.*push)(n);
    numNodes++;
    return replaced;
}

bool Path::addNode(const Vector2d & n) {
    return addNode<list<Vector2d>::reverse_iterator,void (list<Vector2d>::*)(),void (list<Vector2d>::*)(const Vector2d &)>
           (n,nodes.rbegin(),nodes.rend(),
            &list<Vector2d>::pop_back,&list<Vector2d>::push_back);
};

bool Path::addFrontNode(const Vector2d & n) {
    return addNode<list<Vector2d>::iterator,void (list<Vector2d>::*)(),void (list<Vector2d>::*)(const Vector2d &)>
           (n,nodes.begin(),nodes.end(),
            &list<Vector2d>::pop_front,&list<Vector2d>::push_front);
};

bool Path::deleteNode(Vector2d n) {
    bool replaced=false;
    nodes.pop_front();
    numNodes--;
    if (!nodes.empty()) {
        // Calculate distance of new node to second one.
        real dist=(nodes.front()-n).sqrDist();
        if (dist>=distancebefore+5) {
            nodes.pop_front();
            numNodes--;
            replaced=true;
            if (!nodes.empty()) 
                distancebefore=(nodes.front()-n).sqrDist();
        } else 
            distancebefore=dist;
    }
    nodes.push_front(n);
    numNodes++;
    return replaced;
};

void Path::pop_front() {
    numNodes--;
    nodes.pop_front();
}

void Path::pop_back() {
    numNodes--;
    nodes.pop_back();
}

void Path::draw(const Matrix2x3 & view,const Vector2d & pos,Uint32 color) {
    list<Vector2d>::iterator i=nodes.begin();
    Vector2d current,before;
    if (i!=nodes.end()) {
        before=view*((*i++)+pos);
        for (;i!=nodes.end();i++) {
            current=view*((*i)+pos);
            sge_Line(Graphic::getScreen(),Sint16(before.x),Sint16(before.y),
                     Sint16(current.x),Sint16(current.y),
                     color);
            before=current;
        }
    }
};

#define LINELENGTH 5

inline void findNextClipped(const Matrix2x3 & view,const Vector2d & pos,
                            list<Vector2d>::iterator & i,const list<Vector2d>::iterator & end,
                            Vector2d & before,Vector2d & clipCurrent,Vector2d & current,
                            int w,int h,int & drawphase) {
    clipCurrent=current;
    Vector2d clipBefore=before;

    while (i!=end && !clipSegment(w,h,&clipBefore,&clipCurrent)) {
        real dist=(before-current).dist();
        if (dist>drawphase) drawphase=3*LINELENGTH-int(dist-drawphase)%(3*LINELENGTH);
        else drawphase=int(drawphase-dist);
        clipBefore=before=current;
        i++;
        if (i!=end) clipCurrent=current=view*((*i)+pos);
    }
    if (before!=clipBefore) {
        real dist=(before-clipBefore).dist();
        if (dist>drawphase) drawphase=3*LINELENGTH-int(dist-drawphase)%(3*LINELENGTH);
        else drawphase=int(drawphase-dist);
    }
    before=clipBefore;
}


void Path::drawDashed(const Matrix2x3 & view,const Vector2d & pos,Uint32 color) {
    list<Vector2d>::iterator i=nodes.begin();
    if (i!=nodes.end()) {
        Vector2d current,before,clipCurrent;

        before=view*((*i++)+pos);
        if (i!=nodes.end()) current=view*((*i)+pos);

        int temp=Graphic::ticks();
        if (50<temp-laststep) {
            laststep=temp;
            drawphase=(drawphase+1)%(3*LINELENGTH);
        }
        int w=Graphic::getScreen()->w;
        int h=Graphic::getScreen()->h;
        int clipDrawphase=drawphase;
        real drawLength=0;

        real dist=0;
        findNextClipped(view,pos,i,nodes.end(),before,clipCurrent,current,w,h,clipDrawphase);
        dist=(clipCurrent-before).dist();
        while (i!=nodes.end()) {
            if (clipDrawphase<2*LINELENGTH) {
                // Skip this part
                drawLength=clipDrawphase+1;
                if (dist<=drawLength) {
                    drawLength-=dist;
                    before=current;
                    i++;
                    if (i!=nodes.end()) {
                        current=view*((*i)+pos);
                        clipDrawphase-=int(dist);

                        if (point_inside(w,h,current.x,current.y))
                            clipCurrent=current;
                        else findNextClipped(view,pos,i,nodes.end(),before,clipCurrent,current,w,h,clipDrawphase);

                        dist=(clipCurrent-before).dist();
                    } else break;
                } else {
                    before+=(clipCurrent-before)*drawLength/dist;
                    dist-=drawLength;
                    clipDrawphase=(3*LINELENGTH+clipDrawphase-int(drawLength))%(3*LINELENGTH);
                }
            } else {
                // Draw this part
                drawLength=clipDrawphase-2*LINELENGTH+1;
                dist=(clipCurrent-before).dist();
                if (dist<=drawLength) {
                    drawLength-=dist;
                    Graphic::drawArrow(Sint16(before.x),Sint16(before.y),
                                       Sint16(clipCurrent.x),Sint16(clipCurrent.y),
                                       color);
                    before=current;
                    i++;
                    if (i!=nodes.end()) {
                        current=view*((*i)+pos);
                        clipDrawphase-=int(dist);

                        if (point_inside(w,h,current.x,current.y))
                            clipCurrent=current;
                        else findNextClipped(view,pos,i,nodes.end(),before,clipCurrent,current,w,h,clipDrawphase);

                        dist=(clipCurrent-before).dist();
                    } else break;
                } else {
                    Vector2d temp=before+(clipCurrent-before)*drawLength/dist;
                    Graphic::drawArrow(Sint16(before.x),Sint16(before.y),
                                       Sint16(temp.x),Sint16(temp.y),
                                       color);
                    dist-=drawLength;
                    before=temp;
                    clipDrawphase=(3*LINELENGTH+clipDrawphase-int(drawLength))%(3*LINELENGTH);
                }
            }
        }
    }
};

Vector2d & Path::front() {
    return nodes.front();
};

Vector2d & Path::back() {
    return nodes.back();
};

ostream & operator << (ostream& o,const Path& p) {
    o<<"[";
    for (std::list<Vector2d>::const_iterator i=p.nodes.begin();i!=p.nodes.end();i++)
        o<<*i;
    o<<"]";
    return o;
};
