/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iterator>
#include "pathdrawer.h"
#include "starship.h"
#include "functionexts.h"
#include "path.h"
#include "collisionhandler.h"
#include "soundfx.h"

using namespace std;

PathDrawer::PathDrawer(const Simulation * sim,const list<int> &onStep) :
        DELTA(500), ch(NULL),
        need_retrace(false), sim(*sim->getPlayerState(),onStep), realsim(sim), originErased(false) {

    this->sim=*realsim;
    origin=-1;
    unstep=0;
    realsim->deletionListeners.addListener(this);
    this->sim.deletionListeners.addListener(this);
    this->sim.collisionListeners.addListener(this);
}

PathDrawer::PathDrawer(const Simulation * sim,int origin,const list<int> &onStep) :
        DELTA(500), ch(NULL),
        need_retrace(false),sim(*sim->getPlayerState(),onStep), realsim(sim), originErased(false) {
    this->sim=*realsim;
    setOrigin(origin);
    unstep=0;
    realsim->deletionListeners.addListener(this);
    this->sim.deletionListeners.addListener(this);
    this->sim.collisionListeners.addListener(this);
};

void PathDrawer::clear() {

    sim.purge();
    for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++) {
        delete i->path;
        delete i->times;
    }
    paths.clear();

    unstep=0;
}

PathDrawer::~PathDrawer() {
    realsim->deletionListeners.deleteListener(this);
    this->sim.deletionListeners.deleteListener(this);
    clear();
};


void PathDrawer::addBody(GameBody*b) {
    PhysicBody * p=b->clone();
    sim.addBody(p);
    b->id=p->id;
};

void PathDrawer::deleteBody(int id) {
    sim.deleteBody(id);
};

#define is(x,t) (string::npos!=static_cast<GameBody*>(x)->type.find(t))

void PathDrawer::evtCollision(Simulation * sim,PhysicBody * c1,PhysicBody * c2,const Vector2d & normal) {
    if (sim==realsim) return;
    ch.evtCollision(sim,c1,c2,normal);
};

void PathDrawer::evtDeleteBody(Simulation * sim,int id) {
    if (sim==&this->sim) {
        if (id==origin) originErased=true;
        else for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++)
                if (id==i->id) i->erased=true;
    } else {
        if (id==origin) origin=-1;
        else deleteFollow(id);
    };
};

void PathDrawer::retrace() {
    need_retrace=true;
};

void PathDrawer::setOrigin(int origin) {
    this->origin=origin;
    need_retrace=true;
};

void PathDrawer::addFollow(int follow) {
    paths.push_back(FollowItem(follow));
    list<FollowItem>::iterator i=paths.begin();
    i->times->clear();
    i->path->clear();
    i->erased=false;
    if (!originErased) 
        i->path->addNode(sim.physic.bodies[i->id]->getPosition()-sim.physic.bodies[origin]->getPosition());
};
void PathDrawer::deleteFollow(int follow) {
    for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++)
        if (i->id==follow) {
            delete i->path;
            delete i->times;
            paths.erase(i);
            break;
        }
};

void PathDrawer::emptyEventQueue() {
    sim.tevents.clear();
    sim.tevents.clock=realsim->tevents.clock;
};

extern bool keyon;
extern int shipid;

class _GCShipEngineOff : public Event {
private:
    int id;
public:
    _GCShipEngineOff(int id):id(id) {};
    void setQueue(Simulation *,const TimerId &) {}
    int fire(Simulation * sim) {
        Starship*ship=dynamic_cast<Starship*>(sim->physic.bodies[id]);
        if (ship)
            ship->turn_off_engine();
        return 0;
    }
    void cancel(Simulation *) {}
};

void PathDrawer::onStep() {
    Soundfx::setSilence(true);
    if (need_retrace || keyon) {
        sim=*realsim;
        if (keyon) 
            sim.tevents.set(0,_GCShipEngineOff(shipid));

        _trace();
        need_retrace=false;
    }
    Soundfx::setSilence(false);
}

int PathDrawer::numNodes() {
    int count=0;
    for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++)
        if (i->path->numNodes>count) count=i->path->numNodes;
    return count;
};

void PathDrawer::samplePath(int delta) {
    if (delta<0)
        unstep=delta;
    else if (!originErased) {
        delta+=unstep;
        for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++) {
            if (!i->erased) {
                if (i->path->addNode(sim.physic.bodies[i->id]->getPosition()-sim.physic.bodies[origin]->getPosition()))
                    i->times->back()+=delta;
                else 
                    i->times->push_back(delta);
            }
        }
        unstep=0;
    }
}


void PathDrawer::_trace() {
    originErased=false;
    for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++) {
        i->times->clear();
        i->path->clear();
        i->erased=false;
        i->path->addNode(sim.physic.bodies[i->id]->getPosition()-sim.physic.bodies[origin]->getPosition());
    }
    unstep=0;
    sim.step(DELTA,bind1st(mem_fun(&PathDrawer::samplePath),this));
};

#define NUMPATHCOLORS 2
#define PATHCOLOR1 SDL_MapRGB(Graphic::getScreen()->format, 40,168,55)
#define PATHCOLOR2 SDL_MapRGB(Graphic::getScreen()->format, 128,32,32)

void PathDrawer::draw(const Matrix2x3 & view) {
    Uint32 pathColors[]={PATHCOLOR1,PATHCOLOR2};
    int counter=0;
    for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++)
        i->path->drawDashed(view,realsim->physic.bodies[origin]->getPosition(),pathColors[counter++%NUMPATHCOLORS]);
};

bool PathDrawer::grow(int delta) {
    if (need_retrace)
        return false;

    if (!originErased) {
        list<FollowItem>::iterator i;
        for (i=paths.begin();i!=paths.end();i++)
            if (!i->erased) break;
        if (i!=paths.end()) {
            Vector2d last=i->path->numNodes>0?i->path->back():Vector2d();
            Soundfx::setSilence(true);
            sim.step(delta,bind1st(mem_fun(&PathDrawer::samplePath),this));
            Soundfx::setSilence(false);
            return last!=i->path->back();
        }
    }
    return false;
};

void PathDrawer::step(int delta) {
    if (need_retrace)
        return;
    
    grow(delta);

    if (origin>=0) {
        for (list<FollowItem>::iterator i=paths.begin();i!=paths.end();i++) {
            int t=delta;
            while (t>=i->times->front() && i->path->numNodes>=2) {
                t-=i->times->front();
                i->path->pop_front();
                i->times->pop_front();
            }
            if (t && i->path->numNodes>=2) {
                i->times->front()-=t;
                i->path->front()=realsim->physic.bodies[i->id]->getPosition()-realsim->physic.bodies[origin]->getPosition();
            }
        }
    }
}
