/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PATHDRAWER_H
#define PATHDRAWER_H

#include <list>
#include "vector2d.h"
#include "path.h"
#include "timer.h"
#include "simulation.h"
#include "collisionhandler.h"

class PhysicBody;
class Starship;
class PhysicEngine;
class GameBody;

/**
 * This class draws a path from an object and updates the drawing
 * on each step.
 * It has redrawing methods for handle different situations in which
 * updates may be nedded.
 */
class PathDrawer : public SimDeletionListener, public SimCollisionListener {
public:

    PathDrawer(const Simulation *,const std::list<int> &onStep);

    PathDrawer(const Simulation *,int origin,const std::list<int> &onStep);

    ~PathDrawer();

    void clear();
    void addBody(GameBody*);

    /**
     * Buracracy to do before calling the step method of the PhysicEngine.
     */
    void onStep();

    /**
     * Draws the path again with the same parameters. This is necesary
     * if any of the parameters given in the constructor has changed.
     */
    void retrace();

    /**
     * Make advance just the forward ending point of the ship path.
     * @param delta the time passed in milliseconds.
     * @returns true if the path grows.
     */
    bool grow(int delta);

    /**
     * Move both ending points of the ship path through time.
     * @param delta the time passed in milliseconds.
     */
    void step(int delta);

    /**
     * Deletes a body from the list of influences. It must not be the one
     * being traced nor the point of observation.
     * @param id identifier of the object to delete.
     */
    void deleteBody(int id);

    void evtDeleteBody(Simulation * sim,int id);
    void evtCollision(Simulation * sim,PhysicBody * c1,PhysicBody * c2,const Vector2d & normal);

    void draw(const Matrix2x3 & view);

    void setOrigin(int origin);
    void addFollow(int follow);
    void deleteFollow(int follow);

    /**
     * Sets a timer.
     * @param delta Period of time in miliseconds to wait before calling
     *              callback routine.
     * @param data  Data to pass to the callback function.
     * @param callback  Callback routine. It must expect a PhysicEngine* as second argument.
     */
    template <class Event>
    TimerId set(int delta,const Event & evt) {
        return sim.tevents.set(delta,evt);
    };
    TimerId set(const GeneralTimer & tm) {
        return sim.set(tm);
    };
    void deleteEvent(TimerId tid) {
        sim.tevents.deleteEvent(tid);
    };

    int countEvents() {
        return sim.tevents.countEvents();
    }

    void emptyEventQueue();
    void resetEvents() {
        realsim->tevents.copy(sim.tevents);
    };

    int numNodes();

    int DELTA;

private:
    CollisionHandler ch;
    struct FollowItem {
        Path * path;
        std::list<int> * times;
        int id;
        bool erased;
        FollowItem(int id):
                path(new Path()),times(new std::list<int>()),id(id),erased(false) {};
    };
    std::list<FollowItem> paths;
    bool need_retrace;
    int origin;
    int unstep;
    Simulation sim;
    const Simulation * realsim;
    bool originErased;

    void samplePath(int delta);
    void _trace();
};

#endif // PATHDRAWER_H
