/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "physicbody.h"
#include <math.h>
#include <algorithm>

using namespace std;


const double PhysicEngine::G=6.6726e-17;

PhysicBody::PhysicBody(real m,real x0,real y0,real vx0,real vy0,real w):
        mass(m),w(w),pos(x0,y0),vel(vx0,vy0) {
    angle=0;
    w=0;
    waccel=0;
    lastTestPos=previous_pos=next_pos=pos;
    previous_vel=next_vel=vel;
    id=-1;
    flags=DEFAULTS;
}

PhysicBody::PhysicBody(Vector2d pos,Vector2d vel,real m,real w) :
        mass(m),w(w),pos(pos),vel(vel) {
    angle=0;
    w=0;
    waccel=0;
    lastTestPos=previous_pos=next_pos=pos;
    previous_vel=next_vel=vel;
    id=-1;
    flags=DEFAULTS;
}

void PhysicBody::setVelocity(const Vector2d & vel) {
    this->vel=previous_vel=next_vel=vel;
};
void PhysicBody::setPosition(const Vector2d & pos) {
    lastTestPos=this->pos=previous_pos=next_pos=pos;
};


void PhysicBody::saveState() {
    previous_pos=pos;
    previous_vel=vel;
}


void PhysicBody::testForCollision() {
    lastTestPos=getPosition();
};
Vector2d PhysicBody::lastTestPosition() {
    return lastTestPos;
};

void PhysicBody::stepRotation(int delta) {
    real a=(w+waccel*delta/2)*delta;
    if (a!=0) setAngle(getAngle()+a);
    if (angle>360) angle-=360;
    else if (angle<0) angle+=360;
    w+=waccel*delta;
};

inline Vector2d influence(const Vector2d & dist,real mass) {
    double d=dist.dist();
    d*=d*d;
    return (d>0 ? dist*(PhysicEngine::G*mass/d) : Vector2d());
}

PhysicBody * PhysicBody::clone() {
    return new PhysicBody(*this);
};

void PhysicBody::copy(PhysicBody * dest) const {
    *dest=*this;
};

PhysicBody::~PhysicBody() {};

PhysicEngine::PhysicEngine(const PlayerState & ps) : ps(ps) {
    posFactor=0;
    posTime=0;
    loadtime=1;
    loadcount=6;
    bigStep=160;
    minDist=0;
};

PhysicEngine::PhysicEngine(const PhysicEngine & p): ps(p.ps) {
    *this=p;
};

void PhysicEngine::clear() {
    ps.remainingtrash=0;

    posFactor=0;
    loadtime=1;
    loadcount=6;
    heavyBodies.clear();
    bodies.clear();
    erased.clear();
};

inline bool heavy(const PhysicBody * b) {
    return b->mass>=1e10;
}

void PhysicEngine::addBody(PhysicBody * b) {
    if (heavy(b)) heavyBodies.push_back(b);
    if (erased.empty()) {
        b->id=bodies.size();
        bodies.push_back(b);
    } else {
        b->id=erased.front();
        bodies[b->id]=b;
        erased.pop_front();
    }
};

void PhysicEngine::addBody(int id,PhysicBody * b) {
    if (heavy(b)) heavyBodies.push_back(b);
    if (int(bodies.size())<id) {
        bodies.reserve(id+1);
        bodies.resize(id+1,NULL);
    }
    if (!bodies[id]) erased.remove(id);
    b->id=id;
    bodies[b->id]=b;
};

void PhysicEngine::deleteBody(int id) {
    if (!bodies[id]) return;
    heavyBodies.remove(bodies[id]);
    bodies[id]=NULL;
    erased.push_front(id);
};


void PhysicEngine::refitBody(PhysicBody *b) {
    if (heavyBodies.end()==find(heavyBodies.begin(),heavyBodies.end(),b)) {
        if (b->mass>=1e10) heavyBodies.push_back(b);
    } else if (b->mass<1e10) heavyBodies.remove(b);
};

inline Vector2d PhysicEngine::calculateAcceleration(const PhysicBody & affected) {
    Vector2d acc;
    for (list<PhysicBody*>::iterator i=heavyBodies.begin();i!=heavyBodies.end();i++)
        if (*i!=&affected) acc+=influence((*i)->pos-affected.pos,(*i)->mass);
    return acc;
}


void PhysicEngine::_vstep (int delta) {
#define DELTA_LIMIT 3
#define ERROR_LIMIT 0.001
    minDist=0;
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j) continue;
        (*j)->saveState();
    }

    real error;
    real tempdelta=delta;
    real time=0;
    int n=bodies.size();
    Vector2d accels[n];
    calculateForces(0,accels);
    loadcount++;
    Vector2d temp_pos[n];
    Vector2d temp_vel[n];
    do {
        for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
            if (!*j) continue;
            temp_pos[(*j)->id]=(*j)->pos;
            temp_vel[(*j)->id]=(*j)->vel;
        }
        error = vstepRK5(tempdelta,accels);
        real delta_before=tempdelta;
        if (error>0) {
            if (ERROR_LIMIT>=error && real(loadcount)/real(loadtime)<15)
                tempdelta=tempdelta*0.9*pow(ERROR_LIMIT/error,0.20);
            else 
                tempdelta=tempdelta*0.9*pow(ERROR_LIMIT/error,0.25);
            if (tempdelta<DELTA_LIMIT) 
                tempdelta=DELTA_LIMIT;
        } else 
            tempdelta=delta-time;
        if ((error>ERROR_LIMIT && real(loadcount)/real(loadtime)<15) || tempdelta<DELTA_LIMIT)
            if (tempdelta>DELTA_LIMIT) {
                for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
                    if (!*j) continue;
                    (*j)->pos=temp_pos[(*j)->id];
                    (*j)->vel=temp_vel[(*j)->id];
                }
            } else {
                time+=delta_before;
                if (time<delta) {
                    calculateForces(time,accels);
                    ++loadcount;
                }
            }
        else {
            time+=delta_before;
            if (time<delta) {
                calculateForces(time,accels);
                ++loadcount;
            }
        }
        if (tempdelta>delta-time)
            tempdelta=delta-time;
        loadcount+=5;
    } while (time<delta);
    loadtime++;
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j) continue;
        (*j)->next_pos=(*j)->pos;
        (*j)->next_vel=(*j)->vel;
    }
}


void PhysicEngine::step(int delta) {
    step(delta,null);
}

inline void PhysicEngine::callBodiesStep(int delta) {
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (*j) (*j)->step(delta);
    }
};

void PhysicEngine::_step(int delta,bool best) {
    if (delta<posFactor) {
        for (vector<PhysicBody*>::iterator i=bodies.begin();i!=bodies.end();i++) {
            if (*i && !((*i)->flags & PhysicBody::FIXED)) {
                Vector2d veldiff=delta/real(posTime)*((*i)->next_vel-(*i)->previous_vel);
                (*i)->pos+=delta*((*i)->vel+veldiff/2.0);
                (*i)->vel+=veldiff;
                (*i)->stepRotation(delta);
            }
        }
        posFactor-=delta;
    } else {
        for (vector<PhysicBody*>::iterator i=bodies.begin();i!=bodies.end();i++) {
            if (*i && !((*i)->flags & PhysicBody::FIXED)) {
                (*i)->pos=(*i)->next_pos;
                (*i)->vel=(*i)->next_vel;
                (*i)->stepRotation(posFactor);
            }
        }
        if (posTime>0) {
            callBodiesStep(posTime);
            posTime=0;
        }
        if (delta>posFactor) {
            delta-=posFactor;
            int original_delta=delta;
            if (!best && delta<bigStep) 
                delta=bigStep;

            _vstep(delta);
            if (original_delta<delta) {
                for (vector<PhysicBody*>::iterator i=bodies.begin();i!=bodies.end();i++) {
                    if (*i) {
                        if (!((*i)->flags & PhysicBody::FIXED)) {
                            Vector2d veldiff=original_delta/real(delta)*((*i)->next_vel-(*i)->previous_vel);
                            (*i)->pos=(*i)->previous_pos+original_delta*((*i)->previous_vel+veldiff/2.0);
                            (*i)->vel=(*i)->previous_vel+veldiff;
                        }
                        (*i)->stepRotation(original_delta);
                    }
                }
                posTime=delta;
                posFactor=delta-original_delta;
            } else {
                for (vector<PhysicBody*>::iterator i=bodies.begin();i!=bodies.end();i++) {
                    if (*i)
                        (*i)->stepRotation(original_delta);
                }
                callBodiesStep(original_delta);
                posTime=posFactor=0;
            }
        } else posFactor=0;
    }
}

/*
  Old and lovely Runge-Kutta. (5th order).
 */
real PhysicEngine::vstepRK5 (real delta,Vector2d initaccels[]) {
    real maxdist=0;
    int n=bodies.size();
    Vector2d accels[n];
    Vector2d mipos[n];
    Vector2d k1pos[n];
    Vector2d k1vel[n];
    Vector2d k2pos[n];
    Vector2d k2vel[n];
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        mipos[(*j)->id]=(*j)->pos;
        k1pos[(*j)->id]=(*j)->vel*delta;
        k1vel[(*j)->id]=initaccels[(*j)->id]*delta;
        k2pos[(*j)->id]=((*j)->vel+k1vel[(*j)->id]*0.2)*delta;
        (*j)->pos=mipos[(*j)->id]+k1pos[(*j)->id]*0.2;
    }
    calculateForces(delta*0.2,accels);
    Vector2d k3pos[n];
    Vector2d k3vel[n];
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        k2vel[(*j)->id]=accels[(*j)->id]*delta;
        k3pos[(*j)->id]=((*j)->vel+k1vel[(*j)->id]*(3.0/40.0)+k2vel[(*j)->id]*(9.0/40.0))*delta;
        (*j)->pos=mipos[(*j)->id]+k1pos[(*j)->id]*(3.0/40.0)+k2pos[(*j)->id]*(9.0/40.0);
    }
    calculateForces(delta*0.3,accels);
    Vector2d k4pos[n];
    Vector2d k4vel[n];
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        k3vel[(*j)->id]=accels[(*j)->id]*delta;
        k4pos[(*j)->id]=((*j)->vel+k1vel[(*j)->id]*0.3+k2vel[(*j)->id]*(-0.9)+k3vel[(*j)->id]*1.2)*delta;
        (*j)->pos=mipos[(*j)->id]+k1pos[(*j)->id]*0.3+k2pos[(*j)->id]*(-0.9)+k3pos[(*j)->id]*1.2;
    }
    calculateForces(delta*0.6,accels);
    Vector2d k5pos[n];
    Vector2d k5vel[n];
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        k4vel[(*j)->id]=accels[(*j)->id]*delta;
        k5pos[(*j)->id]=((*j)->vel+k1vel[(*j)->id]*(-11.0/54.0)+k2vel[(*j)->id]*2.5+k3vel[(*j)->id]*(-70.0/27.0)+k4vel[(*j)->id]*(35.0/27.0))*delta;
        (*j)->pos=mipos[(*j)->id]+k1pos[(*j)->id]*(-11.0/54.0)+k2pos[(*j)->id]*2.5+k3pos[(*j)->id]*(-70.0/27.0)+k4pos[(*j)->id]*(35.0/27.0);
    }
    calculateForces(delta,accels);
    Vector2d k6pos[n];
    Vector2d k6vel[n];
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        k5vel[(*j)->id]=accels[(*j)->id]*delta;
        k6pos[(*j)->id]=((*j)->vel+k1vel[(*j)->id]*(1631.0/55296.0)+k2vel[(*j)->id]*(175.0/512.0)+k3vel[(*j)->id]*(575.0/13824.0)+k4vel[(*j)->id]*(44275.0/110592.0)+k5vel[(*j)->id]*(253.0/4096.0))*delta;
        (*j)->pos=mipos[(*j)->id]+k1pos[(*j)->id]*(1631.0/55296.0)+k2pos[(*j)->id]*(175.0/512.0)+k3pos[(*j)->id]*(575.0/13824.0)+k4pos[(*j)->id]*(44275.0/110592.0)+k5pos[(*j)->id]*(253.0/4096.0);
    }
    calculateForces(delta*7.0/8,accels);
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (!*j || ((*j)->flags & PhysicBody::FIXED)) continue;
        k6vel[(*j)->id]=accels[(*j)->id]*delta;

        (*j)->pos=mipos[(*j)->id]+(37.0/378.0)*k1pos[(*j)->id]+(250.0/621.0)*k3pos[(*j)->id]+(125.0/594.0)*k4pos[(*j)->id]+(512.0/1771.0)*k6pos[(*j)->id];
        Vector2d pos2=mipos[(*j)->id]+(2825.0/27648.0)*k1pos[(*j)->id]+(18575.0/48384.0)*k3pos[(*j)->id]+(13525.0/55296.0)*k4pos[(*j)->id]+(277.0/14336.0)*k5pos[(*j)->id]+0.25*k6pos[(*j)->id];
        (*j)->vel+=(37.0/378.0)*k1vel[(*j)->id]+(250.0/621.0)*k3vel[(*j)->id]+(125.0/594.0)*k4vel[(*j)->id]+(512.0/1771.0)*k6vel[(*j)->id];

        real dist=((*j)->pos-pos2).dist();
        if (dist>maxdist) 
            maxdist=dist;
    }

    return maxdist;
}

inline Vector2d PhysicEngine::influenceForce(const Vector2d & dist) {
    double d=dist.dist();
    if (d<minDist && 0<d) 
        minDist=d;
    d*=d*d;
    return (d>100 ? dist*(PhysicEngine::G/d) : Vector2d());
}

void PhysicEngine::calculateForces(real delta,Vector2d accels[]) {
    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (*j)
            accels[(*j)->id]=(*j)->getAcceleration(delta);
    }

    for (list<PhysicBody*>::iterator j=heavyBodies.begin();j!=heavyBodies.end();j++) {
        if (*j) {
            list<PhysicBody*>::iterator i=j;
            for (i++;i!=heavyBodies.end();i++) {
                if (*i && !((*j)->flags & (*i)->flags & PhysicBody::FIXED)) {
                    Vector2d acc=influenceForce((*i)->pos-(*j)->pos);
                    accels[(*i)->id]-=acc*(*j)->mass;
                    accels[(*j)->id]+=acc*(*i)->mass;
                }
            }
        }
    }

    for (vector<PhysicBody*>::iterator j=bodies.begin();j!=bodies.end();j++) {
        if (*j && !heavy(*j)) {
            for (list<PhysicBody*>::iterator i=heavyBodies.begin();i!=heavyBodies.end();i++) {
                if (*i && !((*j)->flags & (*i)->flags & PhysicBody::FIXED)) {
                    Vector2d acc=influenceForce((*i)->pos-(*j)->pos);
                    accels[(*i)->id]-=acc*(*j)->mass;
                    accels[(*j)->id]+=acc*(*i)->mass;
                }
            }
        }
    }
}

int PhysicEngine::unstep() {
    int backTime=0;
    if (posFactor) {
        backTime=posTime-posFactor;
        for (vector<PhysicBody*>::iterator i=bodies.begin();i!=bodies.end();i++) {
            if (!*i) continue;
            (*i)->unstep();
            (*i)->stepRotation(-backTime);
        }
        posTime=posFactor=0;
    }
    if (loadcount>1000000) 
        loadcount=6;
    loadtime=1;

    return backTime;
}

int PhysicEngine::getInterpolatedTime() {
    return posFactor? posTime-posFactor :0;
}


PhysicEngine & PhysicEngine::operator = (const PhysicEngine & pe) {
    minDist=pe.minDist;
    posFactor=pe.posFactor;
    posTime=pe.posTime;
    bigStep=pe.bigStep;
    loadcount=pe.loadcount;
    loadtime=pe.loadtime;
    bodies.clear();
    ps=pe.ps;
    heavyBodies.clear();
    erased.clear();

    bodies.reserve(pe.bodies.size());
    for (vector<PhysicBody*>::const_iterator i=pe.bodies.begin();i!=pe.bodies.end();i++)
        if (*i) addBody((*i)->clone());
        else bodies.push_back(NULL);

    erased=pe.erased;

    return *this;
};

