/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PHYSICBODY_H
#define PHYSICBODY_H

#include "global.h"
#include "vector2d.h"
#include <list>
#include <vector>
#include <inttypes.h>
#include "playerstate.h"

class PhysicEngine;

/**
 * This class represents a body with mass. These bodies generate
 * gravitational fields.
 */
class PhysicBody {
public:
    /** Mass of the body.  */
    real mass;

    /** Angular acceleration and velocity. */
    real waccel,w; 

    int id;

    /** Builds a body with mass.
     * @param m The mass of the body.
     * @param x0 Initial horizontal coordinate.
     * @param y0 Initial vertical coordinate.
     * @param vx0 Initial horizontal velocity.
     * @param vy0 Initial vertical velocity.
     */
    PhysicBody(real x0=0,real y0=0,real vx0=0,real vy0=0,real m=1e3,real w=0);

    /** Builds a body with mass.
     * @param m The mass of the body.
     * @param pos Initial position.
     * @param vel Initial velocity.
     */
    PhysicBody(Vector2d pos,Vector2d vel=Vector2d(0,0),real m=1e16,real w=0);

    virtual void setVelocity(const Vector2d & vel);
    virtual void setPosition(const Vector2d & pos);
    virtual void setPosition(real x,real y) {
        setPosition(Vector2d(x,y));
    };
    Vector2d getVelocity() const {
        return vel;
    };
    Vector2d getPosition() const {
        return pos;
    };
    real getPositionX() const {
        return pos.x;
    };
    real getPositionY() const {
        return pos.y;
    };
    virtual Vector2d getAcceleration(real) {
        return Vector2d();
    };

    /**
     * Rotates the body given the amount of time that has passed
     * since last step.
     * @param delta  The amount of passed time.
     */
    void stepRotation(int delta);

    /** Convenient method for setting a common interface for objects
     *  which needs especial updates every physics step.
     */
    virtual void step(int) {};
    /** Undoes the last step of the object. */
    virtual void unstep() {
        pos=next_pos=previous_pos;
        vel=next_vel=previous_vel;
    };
    virtual void onStep() {};

    /**
     * Convenient method for a polimorphic copy constructor.
     */
    virtual PhysicBody * clone();

    /**
     * Convenient method for a polimorphic assignment.
     */
    virtual void copy(PhysicBody * dest) const;

    virtual ~PhysicBody();

    virtual void setAngle(real a) {
        angle=a;
    };
    virtual real getAngle() const {
        return angle;
    };
    virtual void saveState();

    /**
     * Used for various purposes.
     */
    enum {FIXED=0x01
                ,DEFAULTS=0x00
         };
    int_fast8_t flags;

    /**
     * Tells the body it is beign tested for collision.
     */
    void testForCollision();
    Vector2d lastTestPosition();

    Vector2d lastTestPos;

private:
    Vector2d next_pos, previous_pos,next_vel, previous_vel;
    /**
     * Orientation of the body in decimal degrees.
     */
    real angle;
    Vector2d pos; /** Where's the body in some fixed coordinate system. */
    Vector2d vel; /** Acceleration and velocity. */

    friend class PhysicEngine;
};

/** 
 * An implementation of a physical simulation with forces in the newtonian sense.
 * It makes steps of size bigStep, and interpolates the resulting values to estimates
 * positions between steps. This simulator is based on the 5th order Runge-Kutta 
 * algorithm for solving differential equations.
 * */
class PhysicEngine {
public:

    /* The minimum distance between any pair of collidable objects. */
    real minDist;

    /* user data */
    PlayerState ps;

    std::vector<PhysicBody*> bodies;

    /**
     * Universal gravitational constant (m^3/(ms^2*Kg))
     */
    static const double G;


    /**
     * The engine breaks the step in smalles ones to gain accuracy. This is
     * the maximum size those small steps can have.
     */
    int bigStep;

    /**
     * Builds a engine which can animate the physic bodies in a given range.
     * @param prec size of the steps to take.
     */
    PhysicEngine(const PlayerState & ps);


    /**
     * Makes a copy of an engine cloning all its objects.
     * @param p the engine to clone
     */
    PhysicEngine(const PhysicEngine & p);

    /**
     * Adds a body.
     * @param b the body to add.
     */
    void addBody(PhysicBody * b);
    void addBody(int id,PhysicBody * b);

    /**
     * Deletes a body.
     * @param id identifier of the body to delete.
     */
    void deleteBody(int id);

    /**
     * If changes in a body mass are made after adding the body to
     * the engine, it is told to the engine this way.
     * @param b The body to refit.
     */
    void refitBody(PhysicBody *b);

    /**
     * Makes the bodies move. The interval delta is broken into
     * substeps of size bigStep plus a smaller step with the
     * remaining time.
     * @param delta number of elapsed miliseconds.
     * @param posfunction a function to call after each small step
     *                    passing to it the size of the small step.
     */
    template <class PosUnaryFunction >
    void step(int delta,PosUnaryFunction posfunction);
    /** Same as above with a dummy posfunction. */
    void step(int delta);

    /** 
     * Like step, but asks the simulation to determine with as much
     * accuracy as possible the positions of the bodies after the
     * period of time delta. Function step may estimate the position
     * by interpolating the position at nearby times; best_step 
     * does not make such interpolation.
     */
    template <class PosUnaryFunction >
    void best_step(int delta,PosUnaryFunction posfunction);
    /** Same as above with a dummy posfunction. */
    void best_step(int delta);
    /** 
     * Undoes the last substep of the simulation.
     * Returns the amount of time of the undone step.
     * */
    int unstep();
    int getInterpolatedTime();
    void clear();

    /**
     * Calculates acceleration over a given bodie, considering all the bodies in
     * the engine.
     * @param affected the given body.
     * @return the calculated acceleration.
     */
    Vector2d calculateAcceleration(const PhysicBody & affected);

    /**
     * Clones bodies.
     */
    PhysicEngine & operator = (const PhysicEngine &);


    /**
     * Destroys the physic engine, but leaves intact the bodies inside it.
     */
    virtual ~PhysicEngine() {};

    /** 
     * Period of time of the last step. posTime tells how much in the future is the 
     * last computed position of the bodies.
     * */
    int getLastStepTime() {
        return posTime;
    };

private:

    /** 
     * Steps the simulation a given amount of time.
     * The period of time is not broken into substeps.
     * @param best tells if the position of the bodies after the step
     *        can be estimated by interpolation at nearby times or needs
     *        to be as accurate as possible. 
     * */
    void _step(int delta,bool best=false);

    /** Applyies Runge-Kutta with an adaptative step based on error estimates. */
    void _vstep (int delta);

    void callBodiesStep(int delta);

    /** 
     * Period of time of the last step. posTime tells how much in the future is the 
     * last computed position of the bodies.
     * */
    int posTime;
    /** The period of time between the time of the last interpolated step and posTime. */
    int posFactor;

    std::list<PhysicBody*> heavyBodies;
    int loadcount;
    int loadtime;
    real stepRK5 (PhysicBody & pb,real delta,const Vector2d & accel);
    real vstepRK5 (real delta,Vector2d accels[]);

    Vector2d influenceForce(const Vector2d & dist);
    void calculateForces(real delta,Vector2d accels[]);
    std::list<int> erased;
};

inline void null(int){};

template <class PosUnaryFunction>
inline void PhysicEngine::step(int delta,PosUnaryFunction posfunction) {
    int accd=delta;
    while (bigStep<=accd) {
        _step(bigStep);
        posfunction(bigStep);
        accd-=bigStep;
    }
    if (accd) {
        _step(accd);
        posfunction(accd);
    }
}

template <class PosUnaryFunction >
inline void PhysicEngine::best_step(int delta,PosUnaryFunction posfunction) {
    int accd=delta;
    while (bigStep<=accd) {
        _step(bigStep,true);
        posfunction(bigStep);
        accd-=bigStep;
    }
    // if the remaining time is before the next computed time
    // unstep and perform a smaller step.
    if (posFactor>accd) {
        int t=unstep();
        accd+=t;
        posfunction(-t);
    }
    if (accd) {
        _step(accd,true);
        posfunction(accd);
    }
}

inline void PhysicEngine::best_step(int delta) {
    best_step(delta,null);
}

#endif // PHYSICBODY_H
