/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * This file defines counted references and 
 * */

#ifndef REFERENCE_H
#define REFERENCE_H
#include <stdlib.h>

template <class PointedType>
void del(PointedType*p) {
    delete p;
}
template <class PointedType>
void delarray(PointedType*p) {
    delete [] p;
}

/** Reference-counted smart pointer implementation. */
template <class PointedType,void FreeRoutine(PointedType*) = del<PointedType> >
class Ref {
protected:
    PointedType * p;
    int * refCounter;
public:
    Ref(): p(NULL), refCounter(NULL) {};
    explicit Ref(PointedType * p) : p(p), refCounter() {
        refCounter=p?new int(1):NULL;
    };
    Ref(const Ref<PointedType,FreeRoutine> & r) : p(r.p), refCounter(r.refCounter) {
        if (refCounter) ++*refCounter;
    };

    void clear() {
        if (p && !--*refCounter) {
            FreeRoutine(p);
            delete refCounter;
        }
    };

    Ref<PointedType,FreeRoutine> & operator=(const Ref<PointedType,FreeRoutine> & r) {
        if (this!=&r) {
            clear();
            refCounter=r.refCounter;
            if (refCounter) ++*refCounter;
            p=r.p;
        }
        return *this;
    };

    virtual ~Ref() {
        clear();
    };

    PointedType * operator ->() const {
        return p;
    }
    PointedType & operator *() const {
        return *p;
    }
    bool operator == (const Ref<PointedType,FreeRoutine> & r) const {
        return p==r.p;
    }
    bool operator != (const Ref<PointedType,FreeRoutine> & r) const {
        return *this!=r;
    }
    bool operator < (const Ref<PointedType,FreeRoutine> & r) const {
        return p<r.p;
    }
};

template <class PointedType>
inline PointedType* clone(PointedType*p) {
    return p?new PointedType(*p):NULL;
}

/** Deep copy smart pointer implementation. */
template <class PointedType,
          PointedType* CloneRoutine(PointedType*)=clone<PointedType>,
          void FreeRoutine(PointedType*) = del<PointedType>  >
class OwnRef {
protected:
    PointedType * p;
public:
    OwnRef(): p(NULL) {};
    OwnRef(PointedType * p) : p(p) {};
    OwnRef(const OwnRef<PointedType,CloneRoutine,FreeRoutine> & r) {
        p=CloneRoutine(r.p);
    };

    OwnRef<PointedType,CloneRoutine,FreeRoutine> & operator=(const OwnRef<PointedType,CloneRoutine,FreeRoutine> & r) {
        if (this!=&r) {
            if (p) FreeRoutine(p);
            p=CloneRoutine(r.p);
        }
        return *this;
    };

    OwnRef<PointedType,CloneRoutine,FreeRoutine> & operator=(PointedType * p) {
        if (this->p!=p) {
            if (this->p) FreeRoutine(this->p);
            this->p=CloneRoutine(p);
        }
        return *this;
    };

    virtual ~OwnRef() {
        if (p) FreeRoutine(p);/*cout<<rc<<"\n";*/
    };

    PointedType * operator ->() const {
        return p;
    }
    PointedType & operator *() const {
        return *p;
    }
};

#endif // REFERENCE_H
