/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 
 * This file contains template definitions for resources and resource 
 * loaders. 
 */

#ifndef RESOURCEMANAGER_H
#define RESOURCEMANAGER_H

#include <string>
#include <map>
#include <vector>
#include <list>
#include <SDL.h>
#include "reference.h"
#include "fontrenderer.h"
#include "SDL_mixer.h"
#include <set>

class Parser;

void Custom_FreeChunk(Mix_Chunk * mc);

/** A definition for paletes. */
struct Palette {
    SDL_Color colors[256];
    int ncolors;
    Palette() : ncolors(0) {};
};

/** An implementation for singleton classes. */
template <class T>
class Singleton {
private:
    static Ref<T> i;
public:
    static T * getInstance() {
        if (!&*i) i=Ref<T>(new T());
        return &*i;
    };
    static void destroy() {
        if (&*i) i=Ref<T>();
    };
};
template <class T>
Ref<T> Singleton<T>::i;

/** Helper routine to free resources using a specific resource loader. */
template <class R>
void FreeResource(typename R::ResourceT *p) {
    Singleton<R>::getInstance()->free(p);
}

/** 
 * Resource loaders have a list of loaded resources.
 * This class provides such a list so it is possible to share the resources
 * (instead of loading the same resource more than once).
 * */
template <class R>
class ResourceList {
public:
    typedef Ref<typename R::ResourceT,FreeResource<R> > InternalRef;
    std::map<std::string,InternalRef> l;
};

/** Non template methods of a resource loader. */
class RLoader {
public:
    virtual Parser & parse(Parser & p,std::string * id)=0;
    virtual ~RLoader() {}
};

/** 
 * A class which allows to load resources of type T. 
 * A specific resource loader should subclass this or ResourceLoaderNonRef.
 * */
template <class T>
class ResourceLoader : public RLoader {
public:
    /** Resource type loaded by this resource loader. */
    typedef T ResourceT;
    /** Creates a resource from a given identifier. */
    virtual T * create(const std::string & id)=0;
    /** 
     * Parses resource data to initialize any state needed by the resource loader.
     * This method must be called before attempting creation of any resource.
     * */
    virtual Parser & parse(Parser & p,std::string * id)=0;
    /** Frees  a resource loaded with this resource loader. */
    virtual void free(T *)=0;
    virtual ~ResourceLoader() {};
};

/** 
 * A class which allow to load resources which are not ref counted. 
 * This may be desired if the resources have a cheap representation,
 * like a string, that is not worth the trouble to load on demand
 * (all values are kept in memory).
 * The methods behave the same than those of the ResourceLoader class.
 * */
template <class T>
class ResourceLoaderNonRef : public RLoader {
public:
    typedef T ResourceT;
    virtual T create(const std::string & id)=0;
    virtual Parser & parse(Parser & p,std::string * id)=0;
    virtual void free(T &)=0;
    virtual ~ResourceLoaderNonRef() {};
};

/** A reference (counted) to a resource of the type loaded by resource loader R. */
template <class R>
class Resource : public Ref<typename R::ResourceT,FreeResource<R> > {
private:
    typedef typename ResourceList<R>::InternalRef InternalRef;
    Resource<R> & operator=(typename R::ResourceT * p);
public:
    Resource() {};
    explicit Resource(typename R::ResourceT * p): Ref<typename R::ResourceT,FreeResource<R> >(p) {};
    /** 
     * Asks R to load a resource with identifier id and makes it
     * the resource pointed by this reference. If the resource was already loaded
     * it is not loaded again, but shared through this reference.
     * */
    Resource<R>(const std::string & id);
    /**
     * Unloads the resource if no more references to it remain.
     * */
    ~Resource();
};

template <class R>
Resource<R>::Resource(const std::string & id) {
    ResourceList<R> * rl=Singleton<ResourceList<R> >::getInstance();
    // call to the asignment opertor for counted references, 
    // this is dark C++.
    *static_cast<InternalRef*>(this)=rl->l[id];
    if (!&**this)
        rl->l[id]=*this=Resource<R>(Singleton<R>::getInstance()->create(id));
};
template <class R>
Resource<R>::~Resource() {
    // We must explicitly remove the las reference to the resource from the
    // ResourceList or the resource won't be unloaded. 
    ResourceList<R> * rl=Singleton<ResourceList<R> >::getInstance();
    if (this->refCounter && *this->refCounter==2) {
        for (typename std::map<std::string,InternalRef>::iterator i=rl->l.begin();i!=rl->l.end();i++)
            if (&*(i->second)==&**this) {
                rl->l.erase(i);
                break;
            }
    }
};

template <class R>
Resource<R> & Resource<R>::operator=(typename R::ResourceT * p) {
    // calls the base class assignment operator, more dark C++
    *static_cast<InternalRef*>(this)=InternalRef(p);
    return *this;
};



class ZoomRotator;

/** A loader for images that may be rotated. */
class RotateImageLoader : public ResourceLoader<ZoomRotator> {
public:
    ZoomRotator * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id) {
        return p;
    };
    void free(ZoomRotator * p);
};

typedef Resource<RotateImageLoader> RMROTRef;

/** A loader for music data. */
class MusicLoader : public ResourceLoader<Mix_Music> {
private:
    std::map<std::string,std::string> music_paths;
public:
    Mix_Music * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(Mix_Music * p) {
        Mix_FreeMusic(p);
    };
};

typedef Resource<MusicLoader> RMMUSICRef;

/** A loader for sound effects. */
class SoundLoader : public ResourceLoader<Mix_Chunk> {
private:
    std::map<std::string,std::string> sound_paths;
public:
    Mix_Chunk * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(Mix_Chunk * p) {
        Custom_FreeChunk(p);
    };
};

typedef Resource<SoundLoader> RMSOUNDRef;

/** A loader for color palettes. */
class PaletteLoader : public ResourceLoader<Palette> {
private:
    std::map<std::string,std::string> palette_paths;
public:
    Palette * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(Palette * p) {
        delete p;
    };
};

typedef Resource<PaletteLoader> RMPALETTERef;

/** A resource loader for texts. */
class TextLoader : public ResourceLoaderNonRef<std::string> {
    std::map<std::string,std::string> texts;
public:
    std::string create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(std::string & p) {};
};

typedef TextLoader::ResourceT RMTEXTRef;

inline std::string getText(const std::string & id) {
    return Singleton<TextLoader>::getInstance()->create(id);
}

/** A resource loader that loads a list of level paths. */
class LevelLoader : public ResourceLoaderNonRef<std::vector<std::string>*> {
    std::vector<std::string> levels;
public:
    typedef std::vector<std::string>::iterator level_iterator;
    std::vector<std::string> * create(const std::string & id);
    Parser & parse(Parser & p,std::string * id);
    void free(std::vector<std::string> *& p) {};
};

typedef LevelLoader::ResourceT RMLEVELRef;

inline std::vector<std::string> * getLevels() {
    return Singleton<LevelLoader>::getInstance()->create("");
}

/** Reads resource data in file path for each resource loader in the list rl. */
void readResources(const std::string& path,std::list<RLoader*> & rl);

/** Reads resource data in file path for the resource loader r. */
void readResourceFile(const std::string& path,RLoader* r);

/** Frees a resource loader. */
template <class R>
void FreeLoader(ResourceLoader<typename R::ResourceT> * r) {
    Singleton<ResourceList<R> >::destroy();
    Singleton<R>::destroy();
};

/** Frees a resource loader. */
template <class R>
void FreeLoader(ResourceLoaderNonRef<typename R::ResourceT> * r) {
    Singleton<R>::destroy();
};

/** Frees resource loader R. */
template <class R>
void FreeLoader() {
    FreeLoader<R>(Singleton<R>::getInstance());
}

#define RM_SetError SDL_SetError
#define RM_GetError SDL_GetError


#endif // RESOURCEMANAGER_H
