/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SIMULATION_H
#define SIMULATION_H

#include "physicbody.h"
#include "gamebody.h"
#include "timer.h"
#include "functionexts.h"
#include <list>
#include "eventlistenerlist.h"

template <class Item,class Data>
class CollisionEngine;
class Simulation;
class Collision;

typedef TimerQueue<Simulation*>::TimerRef GeneralTimer;
typedef TimerQueue<Simulation*> Timer;
typedef TimerEvent<Simulation*> Event;
typedef _EmptyEvent<Simulation*> EmptyEvent;
typedef TimerQueue<Simulation*>::TimerId TimerId;


/** A listener for collision events. */
class SimCollisionListener {
public:
    /** 
     * Called whenever a collision occurs. 
     * @param sim is the simulation where the collision occurred 
     * @param c1 one of the coliding bodies
     * @param c2 the other coliding bodies
     * @param normal the normal to the surface of contact between the bodies. 
     * */
    virtual void evtCollision(Simulation * sim,PhysicBody * c1,PhysicBody * c2,const Vector2d & normal)=0;
    virtual ~SimCollisionListener() {};
};

/** A listener for object removals. */
class SimDeletionListener {
public:
    /** 
     * Called whenever an object with identifier id is removed
     * from a simulation sim.
     */
    virtual void evtDeleteBody(Simulation * sim,int id)=0;
    virtual ~SimDeletionListener() {};
};

/** Implementation of a physical simulation of a system with gravitational forces. */
class Simulation {
public:
    mutable EventListenerList<SimDeletionListener*> deletionListeners;
    mutable EventListenerList<SimCollisionListener*> collisionListeners;

    /** The PlayerState may need to be updated whenever an event happens. */
    Simulation(const PlayerState & ps,const std::list<int> &onStep);

    /** Steps the simulation a period of time in milliseconds. */
    virtual void step(int delta);
    /** 
     * Steps the simulation a period of time in milliseconds. 
     * A call to this function may involve various substeps.
     * The template parameter specifies a function object that 
     * is called after each substep in the simulation (passing it
     * the size of the substep in milliseconds).
     * A negative time may be passed to posfunction, which indicates that
     * the simulation will go backwards the given amount
     * of time.
     * */
    template <class PosUnaryFunction >
    void step(int delta,PosUnaryFunction posfunction);

    virtual void clear();
    virtual void purge();

    virtual void addBody(PhysicBody * c);
    void addBody(int id,PhysicBody * c);
    virtual void deleteBody(int id);

    /**
     * Sets a timer.
     * @param tm timer event to set.
     * @returns a reference to the set timer.
     */
    virtual TimerId set(const GeneralTimer & tm) {
        return tevents.set(tm);
    };

    virtual void setPrecision(int p) {
        physic.bigStep=p;
    }
    virtual void setCollisionPrecision(int p) {
        colPrecision=p;
    }

    const PlayerState * getPlayerState() const {
        return &physic.ps;
    }
    PlayerState * getPlayerState() {
        return &physic.ps;
    }

    PhysicEngine physic;
    Timer tevents;

    virtual ~Simulation();

    void operator = (const Simulation & orig);

private:
    CollisionEngine<PhysicBody,Simulation> * ce;
    /** 
     * Pushes into the timer event queue the collisions that happened at the exact 
     * collision times. Collisions are detected after they occurred, therefore
     * the simulation may need to undo a small amount of time so the collisions 
     * are produced at their real times with accuracy. This method unsteps the 
     * simulation if needed and returns the amount of undone time.
     * */
    int handle_collisions();

    const std::list<int> & onStep;
    // how much time to wait between collision tests
    int colPrecision;
    // time remaining before the next collision test
    int deltaCol;
    std::list<std::pair<int,Collision> > coldata;
    template <class PosUnaryFunction >
    /** Steps the simulation without testing for collisions. */
    void _step(int delta,PosUnaryFunction posfunction);
    friend int handle_col(Simulation * sim,PhysicBody * c1,PhysicBody * c2,real t,const Vector2d & normal);
    friend class Collision;
    void deleteBody_without_event(int id);
};


template <class PosUnaryFunction >
void Simulation::step(int delta,PosUnaryFunction posfunction) {
    int first_delta = delta;
    if (physic.minDist<32 && deltaCol>10) {
        // test collisions regardless the remaining time for the 
        // next test
        int t=handle_collisions();
        if (t) {
            posfunction(-t);
            _step(t,posfunction);
        }
        deltaCol=colPrecision;
    }

    if (deltaCol<=delta) {
        if (deltaCol) {
            _step(deltaCol,posfunction);
            delta-=deltaCol;
            int t=handle_collisions();
            if (t) {
                posfunction(-t);
                _step(t,posfunction);
            }
        }
        while (delta>=colPrecision) {
            _step(colPrecision,posfunction);
            delta-=colPrecision;
            int t=handle_collisions();
            if (t) {
                posfunction(-t);
                _step(t,posfunction);
            }
        }
        deltaCol=colPrecision-delta;
    } else 
        deltaCol-=delta;

    // if step(0) was called force evaluation of _step 
    if (!first_delta || delta)
        _step(delta,posfunction);
};


template <class PosUnaryFunction >
void Simulation::_step(int delta,PosUnaryFunction posfunction) {
    for (std::list<int>::const_iterator i=onStep.begin();i!=onStep.end();i++)
        physic.bodies[(*i)]->onStep();
    tevents.step(delta,bind2nd(
                         bind1st3(
                             mem_fun2(&PhysicEngine::best_step<PosUnaryFunction>),
                             &physic),
                     posfunction),
                 bind2nd(
                         bind1st3(
                             mem_fun2(&PhysicEngine::step<PosUnaryFunction>),
                             &physic),
                     posfunction)
                );
};


#endif // SIMULATION_H
