/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TIMER_H
#define TIMER_H

#include "global.h"
#include <set>
#include <map>
#include <functional>
#include "reference.h"
//#include <iostream.h>

template <class Ctx>
class TimerQueue;
struct IdObject {};

/** 
 * Interface for handling timer events. The template parameter 
 * Ctx specifies the state the events may need to query or modify.
 */
template <class Ctx>
class GenericTimer {
private:
    typename TimerQueue<Ctx>::TimerId id;
public:
    GenericTimer() : id(typename TimerQueue<Ctx>::TimerId(new IdObject)) {}
    GenericTimer(const typename TimerQueue<Ctx>::TimerId & id) : id(id) {}
    virtual void fire(Ctx ctx)=0;
    virtual void cancel(Ctx ctx)=0;
    virtual int nextCall() const=0;
    virtual void setQueue(Ctx,const typename TimerQueue<Ctx>::TimerId &)=0;
    virtual GenericTimer<Ctx> * clone() const=0;
    virtual ~GenericTimer<Ctx>() {};
    typename TimerQueue<Ctx>::TimerId getId() const {
        return id;
    };
};

template <class Ctx>
class EmptyTimer : public GenericTimer<Ctx> {
public:
    EmptyTimer(const typename TimerQueue<Ctx>::TimerId & id) : GenericTimer<Ctx>(id) {}
    void fire(Ctx) {};
    void cancel(Ctx) {};
    int nextCall() const {
        return 0;
    };
    void setQueue(Ctx,const typename TimerQueue<Ctx>::TimerId &) {};
    GenericTimer<Ctx> * clone() const {
        return new EmptyTimer<Ctx>(*this);
    };
};

/** Function object to compare two events. */
template <class Ctx>
struct lessGT {
    bool operator () (GenericTimer<Ctx> * t1,GenericTimer<Ctx> * t2) const {
        return t1->getId()<t2->getId();
    }
};

/** A wrapper for untimed events. */
template <class Ctx,class Event>
class EventTimer : public GenericTimer<Ctx> {
private:
    Event evt;
    int setTime;
public:
    EventTimer(int delta,const Event & evt) : evt(evt),setTime(delta) {};
    GenericTimer<Ctx> * clone() const {
        return new EventTimer(*this);
    };
    void fire(Ctx ctx) {
        setTime+=this->evt.fire(ctx);
    }
    void cancel(Ctx ctx) {
        this->evt.cancel(ctx);
    }
    int nextCall() const {
        return setTime;
    };
    void setQueue(Ctx ctx,const typename TimerQueue<Ctx>::TimerId & id) {
        this->evt.setQueue(ctx,id);
    };
};

/** An event without a specific execution time. */
template <class Ctx>
class TimerEvent {
public:
    virtual int fire(Ctx)=0;
    virtual void cancel(Ctx)=0;
    virtual void setQueue(Ctx,const typename TimerQueue<Ctx>::TimerId &)=0;
    virtual ~TimerEvent() {};
};

/** An empty event without a specific execution time.
 * This is usefull to have a simulation stop at a particular 
 * time to be defined later.
 */
template <class Ctx>
class _EmptyEvent {
public:
    virtual int fire(Ctx) {
        return 0;
    };
    virtual void cancel(Ctx) {};
    virtual void setQueue(Ctx,const typename TimerQueue<Ctx>::TimerId &) {};
    virtual ~_EmptyEvent() {};
};

template <class Ctx>
GenericTimer<Ctx> * _clone(GenericTimer<Ctx> * gt) {
    return gt->clone();
}

template void del<IdObject>(IdObject*p);

/**
 * A timer that waits certain amount of miliseconds before calling a
 * callback routine.
 */
template <class Ctx>
class TimerQueue {
public:
    typedef GenericTimer<Ctx> * TimerRef;
private:
    typedef std::pair<int,TimerRef > _CallRef;
    typedef std::map<_CallRef,bool> CallType;
    typedef std::set<TimerRef,lessGT<Ctx> > TimerSet;
public:
    typedef typename TimerSet::iterator iterator;
    typedef Ref<IdObject> TimerId;

    TimerQueue(Ctx ctx) : clock(0),ctx(ctx) {}

    /**
    * Sets an event.
    * @param delta Period of time in miliseconds to wait before calling
    *              callback routine.
    * @param data  Data to pass to the callback function.
    * @param callback  Callback routine.
    */
    template <class Event>
    TimerId set(int delta,const Event & evt) {
        TimerRef tm(new EventTimer<Ctx,Event>(clock+delta,evt));
        calls[std::make_pair(tm->nextCall(),tm)]=false;
        timers.insert(tm);
        tm->setQueue(ctx,tm->getId());
        return tm->getId();
    };
    TimerId set(const TimerRef & tm) {
        TimerRef tmc(tm->clone());
        calls[std::make_pair(tmc->nextCall(),tmc)]=false;
        timers.insert(tmc);
        tmc->setQueue(ctx,tmc->getId());
        return tmc->getId();
    };

    /**
     * Removes passed events, and fires events in the given period of time.
     *
     * @param prefunction is executed prior to calling each event passing it 
     * the amount of time elapsed since last event or the time of the step
     * call if there was no prior even in the given period.
     *
     * @param posfunction is executed after the last event fired in the given 
     * interval, passing it the time elapsed since the last event or the time of
     * the step call if there was no event.
     */
    template <class PreUnaryFunction,class PosUnaryFunction>
    void step(int delta,PreUnaryFunction prefunction,PosUnaryFunction posfunction);

    /**
     * Copies to queue all unfired events.
     */
    void copy(TimerQueue<Ctx> & dest) const;

    /**
     * Removes and physically deletes all timers.
     */
    void clear();

    void erase(const iterator & i) {
        cancel(*i);
        timers.erase(i);
    };
    iterator find(const TimerId & t) {
        EmptyTimer<Ctx> et(t);
        return timers.find(&et);
    };
    iterator end() {
        return timers.end();
    }
    void deleteEvent(const TimerId & t) {
        iterator i=find(t);
        if (i!=end()) erase(i);
    };

    int countEvents () {
        return timers.size();
    }

    int clock;
private:
    bool addCall(const TimerRef & tm) {
        int call;
        if ((call=tm->nextCall())>clock) {
            calls[std::make_pair(call,tm)]=false;
            return true;
        } else return false;
    };
    void cancel(const TimerRef & tm) {
        tm->cancel(ctx);
        calls[std::make_pair(tm->nextCall(),tm)]=true;
        delete tm;
    }

    TimerSet timers;
    CallType calls;
    Ctx ctx;
};

template <class Ctx>
template <class PreUnaryFunction,class PosUnaryFunction>
void TimerQueue<Ctx>::step(int delta,PreUnaryFunction prefunction,PosUnaryFunction posfunction) {
    // past: time advanced during this call
    int past=0;
    typename CallType::iterator i,j;
    i=calls.begin();
    int oldclock=clock;
    while (i!=calls.end() && i->first.first<=oldclock+delta) {
        if (i->second) {
            j=i++;
            calls.erase(j);
            continue;
        }
        int t=i->first.first-clock;
        past+=t;
        prefunction(t);
        clock+=t;
        if (i->second) {
            j=i++;
            calls.erase(j);
            continue;
        }
        i->first.second->fire(ctx);
        if (!addCall(i->first.second)) {
            timers.erase(i->first.second);
            delete i->first.second;
        }
        i++;
    }
    posfunction(delta-past);
    clock+=delta-past;
    calls.erase(calls.begin(),i);
};

template <class Ctx>
void TimerQueue<Ctx>::clear() {
    for (iterator i=timers.begin();i!=timers.end();i++)
        cancel(*i);
    timers.clear();
    calls.clear();
};

template <class Ctx>
void TimerQueue<Ctx>::copy(TimerQueue<Ctx> & dest) const {
    dest.clear();
    dest.clock=clock;
    typename TimerSet::const_iterator i;
    for (i=timers.begin();i!=timers.end();i++) dest.set(*i);
};

#endif // TIMER_H
