/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VECTOR2D_H
#define VECTOR2D_H

#include <math.h>
#include <iosfwd>
#include <iostream>
#include "sincosf.h"

typedef double real;

template <class T>
class vector2d {
public:
T x,y;
vector2d(T t=0):x(t),y(t) {};
vector2d(T x0,T y0):x(x0),y(y0) {};
template <class U>
vector2d(const vector2d<U> & v): x(v.x),y(v.y) {}

vector2d operator - (const vector2d &v2) const {
    return vector2d(x-v2.x,y-v2.y);
}
vector2d operator - () const {
    return vector2d(-x,-y);
}
vector2d operator + (const vector2d &v2) const {
    return vector2d(x+v2.x,y+v2.y);
}
vector2d operator / (T d) const {
    return vector2d(x/d,y/d);
}
vector2d operator * (T f) const {
    return vector2d(x*f,y*f);
}

void operator += (const vector2d & v) {
    x+=v.x;
    y+=v.y;
};
void operator -= (const vector2d & v) {
    x-=v.x;
    y-=v.y;
};
void operator *= (T f) {
    x*=f;
    y*=f;
};
void operator /= (T f) {
    x/=f;
    y/=f;
};
bool operator != (const vector2d &v) const {
    return x!=v.x || y!=v.y;
};
bool operator == (const vector2d &v) const {
    return !(*this!=v);
};
bool operator < (const vector2d &v) const {
    return x<v.x || (x==v.x && y<v.y);
};
bool operator > (const vector2d &v) const {
    return v<*this;
};
bool operator <= (const vector2d &v) const {
    return x<=v.x || (x==v.x && y<=v.y);
};
bool operator >= (const vector2d &v) const {
    return v<=*this;
};

T sqrDist() const {
    return x*x + y*y;
}
real dist() const {
    return hypot(real(x),real(y));
}
vector2d rot90() const {
    return vector2d(-y,x);
}
vector2d flip() const {
    return vector2d(y,x);
}
};

typedef vector2d<real> Vector2d;

template <class T, class U>
inline T dot(const vector2d<T> & v1,const vector2d<U> & v2) {
return v1.x*v2.x+v1.y*v2.y;
}

inline Vector2d operator * (real f,const Vector2d & v) {
return v*f;
}


/**
 * This class represent linear transformations on R^2.
 */
class Matrix2x3 {
public:
vector2d<double> c1, c2, c3;

Matrix2x3(real a11,real a12,real a13,real a21,real a22,
          real a23):c1(a11,a21),c2(a12,a22),c3(a13,a23) {}
Matrix2x3():c1(1,0),c2(0,1),c3(0,0) {}

Matrix2x3(const Vector2d & c1,const Vector2d & c2,
          const Vector2d & c3):c1(c1),c2(c2),c3(c3) {}

/**
 * This constructor builds a matrix which translates each point acording to vector
 * (dx,dy), then it rotates the points according to angle, then scales them with factors
 * xscale and yscale, and finally translates them again according to vector (dx2,dxy2).
 * The sumarized transformation is: T v=S(xscale,yscale)*R(angle)*(v+(dx,dy))
 * @param xscale The scale factor for horizontal coordinate.
 * @param yscale The scale factor for vertical coordinate.
 * @param angle The rotation angle in decimal degrees.
 * @param dx The pre-rotation translation shift on the horizontal axis.
 * @param dy The pre-rotation translation shift on the vertical axis.
 * @param dx2 The pos-scaling translation shift on the horizontal axis.
 * @param dy2 The pos-scaling translation shift on the vertical axis.
 */
Matrix2x3(real xscale,real yscale,real angle,real dx,real dy,real dx2,real dy2) {
    real s, c;
    sincos(angle*M_PI/180,&s,&c);
    c1=Vector2d(c*xscale,s*xscale);
    c2=Vector2d(-s*yscale,c*yscale);
    c3=c1*dx+c2*dy;
    c3+=Vector2d(dx2,dy2);
}

Vector2d operator * (const Vector2d & v) const {
    return c1*v.x+c2*v.y+c3;
}

Matrix2x3 operator * (const Matrix2x3 & m) const {
    return Matrix2x3(c1*m.c1.x+c2*m.c1.y,c1*m.c2.x+c2*m.c2.y,(*this)*m.c3);
}

template <class charT>
friend std::basic_ostream<charT> & operator <<
(std::basic_ostream<charT> &,const Matrix2x3&);
};


template <class charT,class T>
std::basic_ostream<charT> & operator <<
(std::basic_ostream<charT> & o,const vector2d<T>&v) {
o<<o.widen('(')<<v.x<<o.widen(',')<<o.widen(' ')<<v.y<<o.widen(')');
return o;
};

template <class charT>
std::basic_ostream<charT> & operator <<
(std::basic_ostream<charT>& o,const Matrix2x3& m) {
o<<o.widen('|')<<m.c1.x<<o.widen(' ')<<m.c2.x<<o.widen(' ')<<m.c3.x<<o.widen('|')<<o.widen('\n');
o<<o.widen('|')<<m.c1.y<<o.widen(' ')<<m.c2.y<<o.widen(' ')<<m.c3.y<<o.widen('|');
return o;
};

#endif // VECTOR2D_H
