/*
  Copyright (C) 2009 Facundo Domínguez 

  This file is part of Spacejunk.

  Spacejunk is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Foobar is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <list>
#include "graphic.h"
#include <math.h>

#include "resourcemanager.h"
#include "sge.h"
#include "debugmsg.h"
#include "physfs.h"
#include <algorithm>
#include <iterator>
#include "parsercombinators.h"

using namespace std;

typedef vector2d<int> Point;


// Walk.first the observer point, 
// Walk.second the wall point.
typedef std::pair<Point,Point> Walk;
typedef std::list<Point> Poly;

class PPoint {
public:
  Point p;
  bool internal;
  PPoint(const Point & p=0,bool internal=false) : p(p), internal(internal) {};
  operator Point() const { return p; }
};

typedef std::list<PPoint> Poly2;

Walk find_wall(const Graphic & g,const Point & p=Point(0,0)){
  int x,y;
  for(x=p.x;x<g.getWidth();x++)
    if (g.insideImage(x,p.y))
      return Walk(Point(x-1,p.y),Point(1,0));

  for(y=p.y+1;y<g.getHeight();y++)
    for(x=0;x<g.getWidth();x++)
      if (g.insideImage(x,y))
	return Walk(Point(x-1,y),Point(1,0));

  return Walk(Point(p.x-2,p.y),Point());
};

std::pair<Point,Point> follow_wall(const Graphic & g,const Walk & walk){
  Point next=walk.first-walk.second.rot90();
  
  if (g.insideImage(next.x,next.y)) // hit a wall
    if (g.insideImage(next.x+walk.second.x,next.y+walk.second.y)) // not on a corner
      return Walk(walk.first+walk.second,-walk.second.rot90()); // turn 90 degrees
    else return Walk(walk.first,-walk.second.rot90()); // turn 90 degrees
  else if (g.insideImage(next.x+walk.second.x,next.y+walk.second.y)) // not on a corner
    return Walk(next,walk.second); // go on in the same direction
  else return Walk(next+walk.second,walk.second.rot90()); // turn -90 degrees on a corner
};

bool is_corner(const Graphic & g,const Walk & walk){
  Point next=walk.first+walk.second.rot90();
  return !g.insideImage(next.x,next.y) && !g.insideImage(next.x+walk.second.x,next.y+walk.second.y);
}

Poly * getOutline(const Graphic & g){
  Poly *l=new Poly();
  Walk init=find_wall(g);
  if (init.first.x==-2) return l;
  Walk current=init;

  do {
    l->push_back(current.first+current.second);
    Walk next=follow_wall(g,current);
    int count=0;
    while (next!=init 
	   && (!is_corner(g,next) || l->back()==next.first+next.second)) {
      if (count>=2)
	if (next.second==current.second 
	    || (!is_corner(g,next) && !g.insideImage(next.first.x+next.second.x-current.second.x*2,next.first.y+next.second.y-current.second.y*2))) count=0;
	
      current=next;
      next=follow_wall(g,current);
      count++;
    }
/*    if (count>1 && next.first==current.first && l->back()!=current.first+current.second) 
      l->push_back(current.first+current.second);*/
    current=next;

    // While the observer does not walk throug the same point 
    // looking at the same wall.
  } while (init!=current);
  
  if (l->back()==l->front() && ++l->begin()!=l->end()) l->pop_back();
  return l;
};

real polygonArea(const Poly & poly){
  if (poly.empty()) return 0;

  Poly::const_iterator i=poly.begin();
  Point prev=*i++;
  real area=0;
  while(i!=poly.end()){
    area+=(prev.y+i->y)*(i->x-prev.x)/2.0;
    prev=*i++;
  }
  area+=(prev.y+poly.front().y)*(poly.front().x-prev.x)/2.0;

  return area;
};

real polygonArea(const Poly2 & poly){
  if (poly.empty()) return 0;

  Poly2::const_iterator i=poly.begin();
  Point prev=*i++;
  real area=0;
  while(i!=poly.end()){
    area+=(prev.y+i->p.y)*(prev.x-i->p.x)/2.0;
    prev=*i++;
  }
  area+=(prev.y+poly.front().p.y)*(prev.x-poly.front().p.x)/2.0;

  return area;
};


bool checkThreshold(const Point & prev,const Point & next,Poly & nodes,real threshold){

  // look at area
  nodes.push_front(prev);
  nodes.push_front(next);
  real area=polygonArea(nodes);
  nodes.pop_front();
  nodes.pop_front();
  Vector2d d1=next-nodes.back();
  Vector2d d2=nodes.back()-prev;
  //  CHERROR<<area<<" "<<Vector2d(prev)<<" "<<Vector2d(nodes.back())<<" "<<Vector2d(next)<<ENDL;
  return (dot(d1,d2)>=0 || dot(d1.rot90(),d2)!=0) && area<threshold;
}

bool checkThresholdAngle(const Point & prev,const Point & next,Poly & nodes,real threshold){
  Vector2d v=next-prev;
  if (v==0) return false;
/*  Point oldsegment=nodes.back()-prev;
  real d=dot(v.rot90(),oldsegment);
  return (d*d<=threshold*v.sqrDist());
}*/


  // look at angles
  for(Poly::const_iterator i=nodes.begin();i!=nodes.end();i++){
    Point oldsegment=*i-prev;
    real d=dot(v.rot90(),oldsegment);
    if (d*d>=threshold*v.sqrDist()){
      return false;
    }
  }
  return true;
}

void optimizePolygonLines(Poly & poly,real threshold){
  if (poly.empty()) return;
  Poly nodes;

  Poly::iterator i=poly.begin();
  Point prev=*i++;
  if (i!=poly.end()){
    Poly::iterator orig=i++;
    if (i!=poly.end()){
      poly.push_back(prev);
      
      do {
	nodes.push_back(*orig);
	if (checkThreshold(prev,*i,nodes,threshold))
	    poly.erase(orig);
	else {
	  prev=*orig;
	  nodes.clear();
	}
	orig=i;
	i++;
      } while (i!=poly.end());

      poly.pop_back();
      i=poly.begin();
      i++;
      nodes.push_back(poly.front());
      if (checkThreshold(prev,*i,nodes,threshold))
	poly.pop_front();
    }
  }
};

void optimizePolygonLineAngles(Poly & poly,real threshold){
  if (poly.empty()) return;
  Poly nodes;

  Poly::iterator i=poly.begin();
  Point prev=*i++;
  if (i!=poly.end()){
    Poly::iterator orig=i++;
    if (i!=poly.end()){
      poly.push_back(prev);
      
      do {
	nodes.push_back(*orig);
	if (checkThresholdAngle(prev,*i,nodes,threshold))
	    poly.erase(orig);
	else {
	  prev=*orig;
	  nodes.clear();
	}
	orig=i;
	i++;
      } while (i!=poly.end());

      poly.pop_back();
      i=poly.begin();
      i++;
      nodes.push_back(poly.front());
      if (checkThresholdAngle(prev,*i,nodes,threshold))
	poly.pop_front();
    }
  }
};

real messure(Poly::iterator i,Poly & nodes){
  Point prev=i!=nodes.begin() ? (*(--i)++) : nodes.back();
  Point next=++i!=nodes.end() ? (*i) : nodes.front();
  i--;
  Vector2d v=next-prev;
  if (v==0) return 0;
  Point oldsegment=*i-prev;
  real d=dot(v.rot90(),oldsegment);
  return fabs(d*d/v.sqrDist());
}

real messureB(Poly::iterator i,Poly & nodes){
  Point prev=i!=nodes.begin() ? (*(--i)++) : nodes.back();
  Point next=++i!=nodes.end() ? (*i) : nodes.front();
  i--;

  Poly mynodes;
  
  // look at area
  mynodes.push_front(prev);
  mynodes.push_front(next);
  mynodes.push_front(*i);
  return fabs(polygonArea(mynodes));
}

class cmp {
public:
  bool operator()(const pair<real,Poly::iterator> & p0,const pair<real,Poly::iterator> & p1){
    return p0.first<p1.first;
  }
};

int optimizePolygonVertex(Poly & poly,real threshold){
  if (poly.empty()) return 0;
  vector<pair<real,Poly::iterator> > nodes(poly.size());
  nodes.clear();
  for(Poly::iterator i=poly.begin();i!=poly.end();i++)
    nodes.push_back(make_pair(messure(i,poly),i));

  //nth_element(nodes.begin(),nodes.begin()+int(threshold),nodes.end(),cmp());
  sort(nodes.begin(),nodes.end(),cmp());

  if (nodes.front().first<threshold) {
    poly.erase(nodes.front().second);
    return 1;
  }
  else return 0;
};

void optimize(Poly & poly,real threshold){
  while(optimizePolygonVertex(poly,threshold));
}

bool convex(const Vector2d & p0,const Vector2d & p1,const Vector2d & p2){
  return dot((p1-p0).rot90(),p2-p1)>=0;
}


bool segment_line_intersect(Point p0,Point p1,Point q0,Point q1){
  // (q1.y-q0.y)(q.x-q0.x)-(q1.x-q0.x)(q.y-q0.y)=0
  return ((q1.y-q0.y)*(p0.x-q0.x)-(q1.x-q0.x)*(p0.y-q0.y))*((q1.y-q0.y)*(p1.x-q0.x)-(q1.x-q0.x)*(p1.y-q0.y))<0;
}

bool line_line_intersect(const Vector2d & p0,const Vector2d & d0,const Vector2d & p1,const Vector2d & d1,Vector2d & n){
  real s=dot(d0.rot90(),d1);
  if (s) n=p1+d1*dot(d0,(p1-p0).rot90())/s;
  return s!=0;
}

bool segment_segment_intersect(const Point & p0,const Point & p1,const Point & q0,const Point & q1,Vector2d & n){
  real s=dot((p1-p0).rot90(),q1-q0);
  if (s) s=dot(p1-p0,(q0-p0).rot90())/s;
  else return false;
  if (0<s && s<1) {
    real t=dot((q1-q0).rot90(),p1-p0);
    if (t) t=dot(q1-q0,(p0-q0).rot90())/t;
    else return false;
    if (0<t && t<1) {
      n=Vector2d(q0)+Vector2d(q1-q0)*s;
      return true;
    }
  }
  
  return false;
}

template <class List>
typename List::iterator nextit(typename List::iterator i,List & p){
  if (++i==p.end()) return p.begin();
  else return i;
};

template <class List>
typename List::iterator previt(typename List::iterator i,List & p){
  if (i==p.begin()) return --p.end();
  else return --i;
};

void polygon_intersect(list<Poly2::iterator> & ps,Poly2 & nodes,Poly2::iterator begin,Poly2::iterator end,Point p0,Point p1){
  if (begin==end) return;
  Poly2::iterator i=begin;
  Poly2::iterator prev;
  Vector2d n;
  do {
    prev=i;
    i=nextit(i,nodes);
    if (segment_segment_intersect(*prev,*i,p0,p1,n)) ps.push_front(prev);
  } while (i!=end);
}


class Circle { 
 public:
  real r;
  Vector2d center;
  Circle() { r=0; }
  Circle(Vector2d center,real r) : r(r), center(center) {}
  template <class T>
  bool contains(const vector2d<T> & p) const {
    return (center-p).sqrDist()<=r*r;
  };
  real area() const { return r*r*M_PI; }
};

Circle find_circle(const list<Point> & border){
  list<Point>::const_iterator first=border.begin();
  if (first==border.end()) return Circle();
  list<Point>::const_iterator second=first;
  second++;
  if (second==border.end()) return Circle(*first,0);
  list<Point>::const_iterator third=second;
  third++;
  Vector2d a0=Vector2d(*second-*first)/2.0;
  if (third==border.end()) return Circle(a0+*first,a0.dist());

  Vector2d a1=Vector2d(*third-*second)/2.0;
  Vector2d n;
  
  if ((a1+*second-*first).sqrDist()<=a1.sqrDist())
    return Circle(a1+*second,a1.dist());
  else if ((a0+*first-*third).sqrDist()<=a0.sqrDist())
    return Circle(a0+*first,a0.dist());
  else if ((Vector2d(*first+*third)/2.0-*second).sqrDist()<=(*first-*third).sqrDist()/4.0)
    return Circle(Vector2d(*first+*third)/2.0,(*first-*third).dist()/2.0);

  if (line_line_intersect(a1+*second,a1.rot90(),a0+*first,a0.rot90(),n))
    return Circle(n,(n-*first).dist());
  else {
    CHERROR<<"find_circle: Something impossible happened: "<<*first<<" "<<*second<<" "<<*third<<ENDL;
    exit(1);
  }
}

Circle minimumDiskWithBorder(list<Point>::iterator begin,list<Point>::iterator end,list<Point> & border){
  if (begin==end || border.size()==3)
    return find_circle(border);
  else {
    list<Point>::iterator second=begin;
    second++;
    Circle c=minimumDiskWithBorder(second,end,border);
    if (c.contains(*begin)) return c;
    
    border.push_front(*begin);
    c=minimumDiskWithBorder(second,end,border);
    border.pop_front();
    return c;
  }
}

Circle minimumDisk(list<Point>::iterator begin,list<Point>::iterator end){
  list<Point> border;
  return minimumDiskWithBorder(begin,end,border);
};

real circle_intersection_area(const Circle & c0,const Circle & c1){
  // Test if one circumference surrounds the other.
  real r=fabs(c1.r-c0.r);
  if ((c1.center-c0.center).sqrDist()<=r*r) 
    if (c1.r>c0.r) return c0.area();
    else return c1.area();

  // Test if they intersect at all
  r=c0.r+c1.r;
  if (r*r<=(c1.center-c0.center).sqrDist()) return 0;
  
  // Intersections
  real cdist=(c1.center-c0.center).dist();
  real xi0=(c1.r*c1.r-c0.r*c0.r-cdist*cdist)/(2*cdist);
  real xi1=(c1.r*c1.r-c0.r*c0.r+cdist*cdist)/(2*cdist);

  real ai0=c0.r*c0.r*(M_PI-asin(fabs(xi0)/c0.r))+fabs(xi0*sqrt(c0.r*c0.r-xi0*xi0));
  real ai1=c1.r*c1.r*(M_PI-asin(fabs(xi1)/c1.r))+fabs(xi1*sqrt(c1.r*c1.r-xi1*xi1));

  if (xi0*xi1>=0)
    if (c1.r>c0.r) return ai0+c1.area()-ai1;
    else return ai1+c0.area()-ai0;
  else return c0.area()-ai0+c1.area()-ai1;
}

real messureDivision(Poly2 & nodes,Poly2::iterator pos,Poly2::iterator current){
  list<Point> component;
  for(Poly2::iterator i=pos;i!=current;i=nextit(i,nodes))
    component.push_back(*i);
  component.push_back(*current);
  Circle c0=minimumDisk(component.begin(),component.end());

  component.clear();
  for(Poly2::iterator i=current;i!=pos;i=nextit(i,nodes))
    component.push_back(*i);
  component.push_back(*pos);
  Circle c1=minimumDisk(component.begin(),component.end());

  real a=c0.area()+c1.area()-circle_intersection_area(c0,c1);
  if (!a) {
    CHERROR<<"Circled area == 0"<<ENDL;
    exit(1);
  } else return polygonArea(nodes)/a;
};

/**
 * How to break a polygon into convex components:
 *    Among all the posible convex subdivisions, 
 *    choose that which minimizes the difference of
 *    the sum of the area of minimum enclosing disks and 
 *    the area of the original poligon.
 */

Poly2::iterator chooseSubdivision(Poly2::iterator pos,Poly2 & nodes,bool relax){
  if (nodes.size()<=3) return nodes.end();

  Poly2::iterator maxpos=nodes.end();
  real max=0;

  list<Poly2::iterator> pinters;

  for(Poly2::iterator i=nextit(pos,nodes);i!=pos; i=nextit(i,nodes) ) {

    pinters.clear();
    polygon_intersect(pinters,nodes,nodes.begin(),--nodes.end(),*i,*nextit(i,nodes));
    if (pinters.empty()) {
      if (convex(i->p,pos->p,nextit(pos,nodes)->p) && 
	  (relax || convex(previt(pos,nodes)->p,pos->p,i->p))&&
	  (dot((nextit(pos,nodes)->p-pos->p).rot90(),i->p-pos->p) || dot((nextit(pos,nodes)->p-pos->p),i->p-pos->p)<0)) {

	real temp=messureDivision(nodes,pos,i);
	if (max<temp) {
	  max=temp;
	  maxpos=i;
	}
      }
    }
  };
  if (max) return maxpos;
  else return nodes.end();
}

void convexComponent(Poly2::iterator pos,Poly2 & nodes,Poly2 & newcomponent){
  if (nodes.size()<=3) return;

  if (pos->internal) {
    pos=nextit(pos,nodes);
    nodes.erase(previt(pos,nodes));
  }

  Poly2::iterator maxpos=chooseSubdivision(pos,nodes,false);
  if (maxpos==nodes.end()) maxpos=chooseSubdivision(pos,nodes,true);

  if (maxpos==nodes.end()) {
    CHERROR<<" polyconvex: ";
    for(Poly2::iterator i=newcomponent.begin();i!=newcomponent.end();i++)
      CHERROR<<"("<<i->p.x<<" "<<i->p.y<<") ";
    CHERROR<<ENDL;

    CHERROR<<"polyconvex: ";
    for(Poly2::iterator i=nodes.begin();i!=nodes.end();i++)
      CHERROR<<"("<<i->p.x<<" "<<i->p.y<<") ";
    CHERROR<<ENDL;

    CHERROR<<"Polygon decomposition not found!"<<ENDL;
    exit(1);
  }

  if (!pos->internal) {
    newcomponent.push_back(*pos);
    pos->internal=true;
  }
  Poly2::iterator j=pos;
  j=nextit(j,nodes);
  while (j!=maxpos) {
    Poly2::iterator k=j;
    j=nextit(j,nodes);
    newcomponent.push_back(*k);
    nodes.erase(k);
  } 
  if (!newcomponent.back().internal) newcomponent.push_back(PPoint(maxpos->p,true));

  if (maxpos->internal && nextit(pos,nodes)!=maxpos && previt(pos,nodes)!=maxpos) 
    nodes.erase(maxpos);
  if (nextit(pos,nodes)->internal) nodes.erase(nextit(pos,nodes));
  if (previt(pos,nodes)->internal) nodes.erase(pos);
}

void breakinconvex(list<Poly2*> & ps,Poly p){
  bool done;
  Poly2 * p2=new Poly2();
  for(Poly::iterator i=p.begin();i!=p.end();i++) p2->push_back(*i);
  ps.push_back(p2);
  for(list<Poly2*>::iterator j=ps.begin();j!=ps.end();j++)
    do {
      done=true;
      for(Poly2::iterator i=(*j)->begin();i!=(*j)->end();i++)
	if (!convex(previt(i,**j)->p,i->p,nextit(i,**j)->p)) {
	  Poly2 * t=new Poly2();
	  convexComponent(i,**j,*t);
	  if (!t->empty()) ps.push_back(t);
	  else delete t;
	  done=false;
	  break;
	}
    } while (!done);
}


void polygonBBOX(const Poly & poly,SDL_Rect * r){
  r->x=r->y=r->w=r->h=0;
  if (poly.empty()) return;
  Poly::const_iterator i=poly.begin();
  r->x=i->x;r->y=i->y;

  for (;i!=poly.end();i++){
    if (i->x<r->x) {
      r->w+=r->x-i->x;
      r->x=i->x;
    }
    if (i->y<r->y) {
      r->h+=r->y-i->y;
      r->y=i->y;
    }
    if (r->x+r->w-1<i->x) r->w=i->x-r->x+1;
    if (r->y+r->h-1<i->y) r->h=i->y-r->y+1;
  }
};

typedef std::vector<PPoint> Polygon;

void Poly2Polygon(const Poly2 & pin,Polygon & pout){
  pout.resize(pin.size());
  int count=0;
  for(Poly2::const_iterator i=pin.begin();i!=pin.end();i++)
    pout[count++]=*i;
}

ostream & operator << (ostream & o,const Circle & c){
  o<<"c "<<c.center<<" "<<c.r<<"\n";
  return o;
};

ostream & operator << (ostream & o,const Polygon & p){
  o<<"p "<<p.size();
  for(Polygon::const_iterator i=p.begin();i!=p.end();i++)
    o<<" "<<i->p<<" "<<(i->internal?1:0);
  o<<"\n";
  return o;
};

Lexer & parseCircle(Lexer & p,Circle * c){
  PARSEbegin(Lexer,p);
    LS(L"c") LVECTOR(&c->center) LFLOAT(&c->r);
  PARSEend;
  return p;
};

Lexer & parsePolygon(Lexer & p,Polygon * pol){
  pol->clear();
  PARSEbegin(Lexer,p);
    Vector2d v;
    int i;
    LS(L"p");LINT(&i);
    pol->reserve(i);
    MANYbegin;
      LVECTOR(&v);LINT(&i);
      pol->push_back(PPoint(v,i));
    MANYend;
  PARSEend;
  return p;
};

Lexer & parseMask(Lexer & p,std::list<pair<Polygon*,Circle> > & m){
  Polygon * pol;
  PARSEbegin(Lexer,p);
    MANYbegin;
      Circle circ;
      pol=new Polygon();
      P1(parsePolygon,pol);
      P1(parseCircle,&circ);
      m.push_back(make_pair(pol,circ));
      pol=false;
    MANYend;
  PARSEend;
  if (pol) delete pol;
  return p;
};


#define mod(x,m) ((x)<(m) ? (x) : (x-m))
#define prom_mod(i,j,m) ((i)<=(j) ? ((i)+(j))/2 : mod(((i)+(j)+(m))/2,m))
#define increase(i,pol) ((i)==int(pol.size()-1)?0:(i)+1)
#define decrease(i,pol) ((i)?(i)-1:int(pol.size()-1))

class MKPolygon {
private:
  Point * points;
  Uint16 * byangle;  
  real * angles;
  Uint16 mysize;
  real shift;
  int dir;
public:
  typedef Point * iterator;
  MKPolygon(int size) : mysize(size),shift(0),dir(1) {
    assert(size>1);
    points=new Point[mysize];
    byangle=new Uint16[mysize];
    angles=new real[mysize];
  }
  iterator begin() const { return points; }
  iterator end() const { return points+mysize; }
  Point & front() const { return points[0]; }

  Uint16 getIndex(real angle) const {
    return byangle[getAngleIndex(angle)];
  }
  Uint16 getIndexI(Uint16 angleindex) const {
    return byangle[angleindex];
  }
  real getAngle(Uint16 index) const {
    return angles[index];
  }
  Uint16 getAngleIndex(real angle) const {
    while (angle<0) angle+=2*M_PI;
    while (angle>=2*M_PI) angle-=2*M_PI;
    return Uint16(trunc(mysize*angle/(2*M_PI)));
  }
  Point & operator[] (Uint16 index) const {
    return points[index];
  }
  Uint16 size() const { return mysize; }
  void compute();
};

class PolygonTransformer {
protected:
  const MKPolygon*pol;
public:
  PolygonTransformer(const MKPolygon*pol): pol(pol){};
  virtual Uint16 getIndex(real angle) const { return pol->getIndex(angle); };
  virtual real getAngle(Uint16 index) const { return pol->getAngle(index); };
  virtual Uint16 getAngleIndex(real angle) const { return pol->getAngleIndex(angle); };
  virtual Vector2d operator[] (Uint16 index) const { return (*pol)[index]; };
  Uint16 getIndexI(Uint16 angleindex) const { return pol->getIndexI(angleindex); }
  virtual ~PolygonTransformer(){};
  Uint16 size() const { return pol->size(); }
};

class PolygonReverser : public PolygonTransformer {
private:
  Matrix2x3 m;
public:
  PolygonReverser(const Matrix2x3 & m,const MKPolygon*pol) : PolygonTransformer(pol), m(m) {};
  Uint16 getIndex(real angle) const { return pol->getIndex(M_PI+angle); };
  real getAngle(Uint16 index) const { return mod(M_PI+pol->getAngle(index),2*M_PI); };
  Uint16 getAngleIndex(real angle) const { return pol->getAngleIndex(M_PI+angle); };
  Vector2d operator[] (Uint16 index) const { return -(m*(*pol)[index]); };  
  void print(){
    CHERROR<<"reverser:"<<ENDL;
    for(int i=0;i<size();i++) {
      CHERROR<<i<<" "<<(*this)[i]<<" "<<getAngle(i)<<" "<<pol->getAngle(i)<<" "<<getIndex(getAngle(i))<<ENDL;
    }
  }
};

inline bool between(int x,int y,int z){
  if (x<=z) return x<y && y<=z;
  else return y<=z || x<y;
}

void MKPolygon::compute(){
  // fill angles
  for(int i=1;i<mysize;i++){
    Point p=points[i]-points[i-1];
    assert(p!=0);
    real a=atan2(p.y,p.x);
    if (a<0) a+=2*M_PI;
    angles[i-1]=a;
  };
  Point p=points[0]-points[mysize-1];
  assert(p!=0);
  real a=atan2(p.y,p.x);
  if (a<0) a+=2*M_PI;
  angles[mysize-1]=a;

  // fill byangle
  int j=0;
  int aindex=getAngleIndex(angles[j]);
  int prev=aindex;
  for(int i=0;i<mysize;i++){
    while(!between(prev,i,aindex)) {
      j=mod(j+1,mysize);
      prev=aindex;
      aindex=getAngleIndex(angles[j]);
    }
    byangle[i]=j;
  }
};

class MKPoint {
public:
  Uint16 i0,i1;
  MKPoint(Uint16 i0,Uint16 i1) : i0(i0), i1(i1) {};
  MKPoint() : i0(0), i1(0) {}
  bool operator != (const MKPoint & p) const { return i0!=p.i0 || i1!=p.i1; }
};

inline int cmp_angles(real angle0,real angle1){
  real diff=angle0-angle1;
  if ((0<diff && diff<M_PI) || diff<=-M_PI) return -1;
  else if (diff) return 1;
  else return 0;
}

int search_MKPolygonAngle(const PolygonTransformer & pol,real angle,int left,int right){
  while (left!=right){
    int mid=prom_mod(left,right,pol.size());
    if (cmp_angles(pol.getAngle(mid),angle)==1) left=increase(mid,pol);
    else right=mid;
  }
  return left;
}

class MKSum {
  const MKPolygon & pol0;
  const PolygonTransformer & pol1;
  real angleshift;
public:
  MKSum(const MKPolygon * pol0,real angleshift,const PolygonTransformer * pol1) : 
    pol0(*pol0),pol1(*pol1),angleshift(angleshift) {};

  MKPoint MKnext(const MKPoint & p) const;
  MKPoint MKprev(const MKPoint & p) const;
  MKPoint MKprom(const MKPoint & p0,const MKPoint & p1) const;
  Vector2d operator[] (const MKPoint & p) const {
    return Vector2d(pol0[p.i0])+pol1[p.i1];
  };
  MKPoint pickNode(int i0) const;
  MKPoint pickNode1(int i1) const;
};

MKPoint MKSum::pickNode(int i0) const {
  real a=pol0.getAngle(i0)+angleshift;
  int angle=pol1.getAngleIndex(a);
  return MKPoint(i0,search_MKPolygonAngle(pol1,a,pol1.getIndexI(angle),pol1.getIndexI(increase(angle,pol1))));
};

MKPoint MKSum::pickNode1(int i1) const {
  real a=pol1.getAngle(i1)-angleshift;
  int angle=pol0.getAngleIndex(a);
  PolygonTransformer pt(&pol0);
  return MKPoint(search_MKPolygonAngle(pt,a,pol0.getIndexI(angle),pol0.getIndexI(increase(angle,pol0))),i1);
};

MKPoint MKSum::MKnext(const MKPoint & p) const {
  switch(cmp_angles(pol0.getAngle(p.i0)+angleshift,pol1.getAngle(p.i1))){
  case -1: return MKPoint(p.i0,increase(p.i1,pol1));break;
  case 1: return MKPoint(increase(p.i0,pol0),p.i1);break;
  default: return MKPoint(increase(p.i0,pol0),increase(p.i1,pol1));
  }
}
  
MKPoint MKSum::MKprev(const MKPoint & p) const {
  int np0=decrease(p.i0,pol0);
  int np1=decrease(p.i1,pol1);
  switch(cmp_angles(pol0.getAngle(np0)+angleshift,pol1.getAngle(np1))){
  case -1: return MKPoint(np0,p.i1);break;
  case 1: return MKPoint(p.i0,np1);break;
  default: return MKPoint(np0,np1);
  }
}

MKPoint MKSum::MKprom(const MKPoint & p0,const MKPoint & p1) const {
  if (p0.i0==p1.i0) {
    int n=prom_mod(p0.i1,p1.i1,pol1.size());
    if (n==p0.i1) return p0;
    else return pickNode1(n);
  } else {
    int n=prom_mod(p0.i0,p1.i0,pol0.size());
    if (n==p0.i0) {
      MKPoint temp=pickNode(n);
      if (temp!=p0) return temp;
      else {
	temp=MKnext(temp);
	if (temp!=p1) return temp;
	else return p0;
      }
    } else return pickNode(n);
  }
}

real linedist(const Vector2d &p,const Vector2d &d,const Vector2d & x) { 
  return dot(x-p,d.rot90());
};

bool line_line_intersect(const Vector2d & p,const Vector2d & dp,
			 const Vector2d & q,const Vector2d & dq,
			 real & s,real & t){
  s=dot(dp,dq.rot90());
  if (s) {
    real r=dot(q-p,dq.rot90());
    //qpos=r>0?1:(r<0?-1:0);
    s=r/s;
    t=dot(dp.rot90(),dq);
    r=dot(p-q,dp.rot90());
    //    ppos=r>0?1:(r<0?-1:0);
    t=r/t;
  } else return false;
  return true;
}

/* Returns a node which is on the line or has an edge that touches the line. */
bool find_midle_0(const Vector2d &p,const Vector2d &d,const MKSum & sum, 
		  MKPoint left,MKPoint right,MKPoint & res){
  real dleft=linedist(p,d,sum[left]);
  real dright=linedist(p,d,sum[right]);
  bool ascending;
  if (dright<dleft) ascending=true;
  else if (dleft<dright) ascending=false;
  else ascending=dleft<linedist(p,d,sum[sum.MKnext(left)]);

  MKPoint mid;
  real dmid;
  MKPoint pnext;
  real dnext;
  if (ascending) {
    do {
      mid=sum.MKprom(left,right);
      dmid=linedist(p,d,sum[mid]);
      pnext=sum.MKnext(mid);
      dnext=linedist(p,d,sum[pnext]);
      if (dmid<=dnext)
	if (dleft<dnext) left=pnext;
	else right=mid;
      else if (dnext<0) right=mid;
      else left=pnext;
    } while(left!=right);
  } else {
    do {
      mid=sum.MKprom(left,right);
      dmid=linedist(p,d,sum[mid]);
      pnext=sum.MKnext(mid);
      dnext=linedist(p,d,sum[pnext]);
      if (dmid>=dnext)
	if (dnext<dleft) left=pnext;
	else right=sum.MKprev(mid);
      else if (0<dnext) right=mid;
      else left=pnext;
    } while(left!=right);
  }
  pnext=sum.MKnext(left);
  if (linedist(p,d,sum[left])*linedist(p,d,sum[pnext])<=0){
    res=left;
    return true;
  }

  return false;
}

bool intersect_MKSum_line(const Vector2d &p,const Vector2d &d,const MKSum & sum,real & t0,real & t1){
  MKPoint left=sum.pickNode(0);
  MKPoint ip0;
  if (find_midle_0(p,d,sum,left,sum.MKprev(left),ip0)){
    left=sum.MKnext(ip0);
    MKPoint ip1;
    if (find_midle_0(p,d,sum,left,ip0,ip1)) {
      Vector2d p0=sum[ip0];
      MKPoint pnext=sum.MKnext(ip0);
      real s;
      line_line_intersect(p,d,p0,sum[pnext]-p0,t0,s);
      Vector2d q0=sum[ip1];
      MKPoint qnext=sum.MKnext(ip1);
      line_line_intersect(p,d,q0,sum[qnext]-q0,t1,s);
      if (t1<t0) swap(t0,t1);
      return true;
    };
  }
  return false;
}


template <class Function> 
MKPoint find_min_extreme(const Function & F,const MKSum & pol){
  MKPoint left=pol.pickNode(0);
  if (F.stop()) return left;

  real ini=F(left,pol);
  MKPoint right=pol.MKprev(left);
  if (F.stop()) return right;

  real end=F(right,pol);
  real pnext;
  bool ascending;
  MKPoint midnext;
  if (end<ini) ascending=true;
  else if (ini<end) ascending=false;
  else {
    midnext=pol.MKnext(left);
    pnext=F(midnext,pol);
    if (F.stop()) return midnext;
    if (ini<pnext) return left;
    else ascending=false;
  }

  MKPoint mid;
  if (ascending) {
    do {
      mid=pol.MKprom(left,right);
      midnext=pol.MKnext(mid);
      real pmid=F(mid,pol);
      if (F.stop()) return mid;

      real pmidnext=F(midnext,pol);
      if (F.stop()) return midnext;
      
      if (pmid<=pmidnext) 
	if (pmidnext<=end) right=mid;
        else left=midnext;
      else left=midnext;
    } while(left!=right);
  } else {
    do {
      mid=pol.MKprom(left,right);
      midnext=pol.MKnext(mid);
      real pmid=F(mid,pol);
      if (F.stop()) return mid;
      real pmidnext=F(midnext,pol);
      if (F.stop()) return midnext;
      
      if (pmid>pmidnext) 
	if (pmidnext>=end) right=pol.MKprev(mid);
        else left=midnext;
      else right=mid;
    } while(left!=right);    
  }
  return left;
}


class pointangle {
public:
  Vector2d p;
  mutable bool mystop;
  pointangle(const Vector2d & p) : p(p), mystop(false) {}
  real operator ()(const MKPoint & i,const MKSum &pol) const {
    Vector2d p1=pol[pol.MKnext(i)];
    Vector2d p0=pol[i];
    real r=dot(p-p0,(p1-p0).rot90());
    if (r<0) mystop=true;
    return r; 
  };
  bool stop() const { return mystop; };
};

bool interior(const Point & p,const MKSum & pol){
  pointangle pa(p);
  find_min_extreme(pa,pol);
  return !pa.stop();
}

Point mousePosition;

bool process_events(Matrix2x3 & m,Point & p,real & angle){
  SDL_Event e;
  while (SDL_PollEvent(&e)){
    switch(e.type){
    case SDL_KEYDOWN:        
      switch (e.key.keysym.sym){
      case SDLK_PLUS:
	angle+=1;
	break;
      case SDLK_MINUS:
	angle-=1;
	break;
      case SDLK_LEFT:
	p.x-=1;
	break;
      case SDLK_RIGHT:
	p.x+=1;
	break;
      case SDLK_UP:
	p.y-=1;
	break;
      case SDLK_DOWN:
	p.y+=1;
	break;
      case SDLK_ESCAPE: return false;
      default:;
      }break;
    case SDL_MOUSEMOTION:
      mousePosition.x=e.motion.x;
      mousePosition.y=e.motion.y;
      break;
    }
  }
  return true;
}

int W=9;
int H=9;

void drawGrid(const Matrix2x3 & m,const Graphic & g){
  Uint32 graycolor=SDL_MapRGB(Graphic::getScreen()->format,128,128,128);
  for(int y=0;y<g.getHeight();y++)
    for(int x=0;x<g.getWidth();x++)
      if (g.insideImage(x,y)) {
	Vector2d xy=m*Vector2d(x,y);
	Vector2d xy1=m*Vector2d(x+1,y+1);
	Graphic::drawRect(int(xy.x),int(xy.y),int(xy1.x),int(xy1.y),graycolor);
      }
};

void drawVertexes(const Matrix2x3 & m,MKPolygon* p){
  MKPolygon::iterator i=p->begin();
  Vector2d prev=*i++;
  for(;i!=p->end();i++){
    MKPolygon::iterator j=i;
    Vector2d next= ++j==p->end() ? p->front() : *j;
    Vector2d p0=m*prev;
    Vector2d p1=m**i;
    Vector2d p11=m*(*i+1);
    Vector2d p2=m*next;
    sge_FilledRect(Graphic::getScreen(),int(p1.x),int(p1.y),
		   int(p11.x),int(p11.y),
		   convex(prev,*i,next) ? SDL_MapRGB(Graphic::getScreen()->format,255,255,255) :
		   SDL_MapRGB(Graphic::getScreen()->format,0,255,0));
    prev=*i;
  }
}

void drawLines(const Matrix2x3 & m,MKPolygon* p,Uint32 color){
  MKPolygon::iterator i=p->begin();
  Vector2d prev=*i;
  //  bool internal=i->internal;
  i++;
  for(;i!=p->end();i++){
    //    if (!internal) {
      Vector2d p0=m*(prev+0.5);
      Vector2d p1=m*(Vector2d(*i)+0.5);
      Graphic::drawArrow(int(p0.x),int(p0.y),int(p1.x),int(p1.y),color);
      //}
    prev=*i;
    //internal=i->internal;
  }
  //  if (!internal) {
  Vector2d p0=m*(prev+0.5);
  Vector2d p1=m*(Vector2d(p->front())+0.5);
  Graphic::drawArrow(int(p0.x),int(p0.y),int(p1.x),int(p1.y),color);
    //  }
}

void drawPolygons(const Matrix2x3 & m,list<MKPolygon*> & lp,Uint32 color){

    for(list<MKPolygon*>::iterator p=lp.begin();p!=lp.end();p++){
      drawVertexes(m,*p);

      drawLines(m,*p,color);

      Poly points;
      for(MKPolygon::iterator i=(*p)->begin();i!=(*p)->end();i++) points.push_back(*i);
      Circle c=minimumDisk(points.begin(),points.end());
      Vector2d cxy=m*(c.center+0.5);
      Graphic::drawCircle(cxy,int(c.r*W),SDL_MapRGB(Graphic::getScreen()->format,0,255,0));
    }
};

void makeMKPolygons(list<MKPolygon*> & lp,const list<Poly2*> & lpp){
  for(list<Poly2*>::const_iterator i=lpp.begin();i!=lpp.end();i++){
    MKPolygon * mkp=new MKPolygon((*i)->size());
    int count=0;
    for(Poly2::iterator j=(*i)->begin();j!=(*i)->end();j++)
      (*mkp)[count++]=*j;
    mkp->compute();
    lp.push_back(mkp);
  }
};

MKPolygon * makeFromSum(const MKSum &sum){
  list<Point> l;
  MKPoint first=sum.pickNode(0);
  MKPoint next=first,prev;
  //int cc=0;
  do {
    //    CHERROR<<next.i0<<" "<<next.i1<<ENDL;
    l.push_back(sum[next]);
    prev=next;
    next=sum.MKnext(next);
    //    CHERROR<<"next "<<next.i0<<" "<<next.i1<<ENDL;
    //    CHERROR<<"prev "<<sum.MKprev(next).i0<<" "<<sum.MKprev(next).i1<<ENDL;
    assert(prev.i0==sum.MKprev(next).i0);
    assert(prev.i1==sum.MKprev(next).i1);
  } while(next!=first);
  MKPolygon * mkp=new MKPolygon(l.size());
  int count=0;
  for(list<Point>::iterator j=l.begin();j!=l.end();j++)
    (*mkp)[count++]=*j;
  mkp->compute();
  
  return mkp;
}

void drawInters(real angle,const Matrix2x3& m,const Vector2d & bluelinep,const Point & bluelined,
		const MKPolygon * pol0,const MKPolygon * pol1,const Vector2d &pos0,const Vector2d & pos1,
		const Graphic & g1){
  Point sumpos(Graphic::getScreen()->w/2,Graphic::getScreen()->h/2);
  Vector2d p0,p1;
  real t0,t1;
  
  PolygonReverser pr(Matrix2x3(1,1,angle,-g1.getWidth()/2,-g1.getHeight()/2,
			       pos1.x-pos0.x+g1.getWidth()/2,pos1.y-pos0.y+g1.getHeight()/2),pol1);
  MKSum sum(pol0,-M_PI*angle/180,&pr);
  MKPolygon * sumpol=makeFromSum(sum);
  drawLines(Matrix2x3(1,1,0,0,0,sumpos.x,sumpos.y)*m,sumpol,SDL_MapRGB(Graphic::getScreen()->format,255,0,0));


/*  MKPoint first=sum.pickNode(0);
  MKPoint next=first,prev;
  CHERROR<<"sum: "<<ENDL;
  do {
    CHERROR<<sum[next]<<ENDL;
    next=sum.MKnext(next);
  } while(next!=first);
  pr.print();*/

  // draw origin
  p0=sumpos;
  p1=sumpos;
  p1+=W;
  Graphic::drawFilledRect(int(p0.x),int(p0.y),int(p1.x),int(p1.y),
			  SDL_MapRGB(Graphic::getScreen()->format,255,0,0));    

  if (intersect_MKSum_line(bluelinep-sumpos/W,bluelined,sum,t0,t1)) {
    //if (intersect_MKSum_line(Vector2d(30,31),Point(1,1),sum,t0,t1)) {
    //CHERROR<<"inters "<<t0<<" "<<t1<<ENDL;
    p0=m*(bluelinep+t0*bluelined);
    Vector2d p01=m*(bluelinep+t0*bluelined+1);
    p1=m*(bluelinep+t1*bluelined);
    Vector2d p11=m*(bluelinep+t1*bluelined+1);
    Graphic::drawFilledRect(int(p0.x),int(p0.y),int(p01.x),int(p01.y),
			    SDL_MapRGB(Graphic::getScreen()->format,0,0,255));    

    Graphic::drawFilledRect(int(p1.x),int(p1.y),int(p11.x),int(p11.y),
			    SDL_MapRGB(Graphic::getScreen()->format,0,0,255));    

    if (interior((mousePosition-sumpos+1)/W,sum)){
      p0=mousePosition;
      p1=mousePosition+W;
      Graphic::drawFilledRect(int(p0.x),int(p0.y),int(p1.x),int(p1.y),
	       SDL_MapRGB(Graphic::getScreen()->format,255,0,0));    
    }
  };

  p0=m*(bluelinep+0.5);
  p1=m*(bluelinep+bluelined);
  Graphic::drawArrow(int(p0.x),int(p0.y),int(p1.x),int(p1.y),SDL_MapRGB(Graphic::getScreen()->format,0,0,255));
  delete sumpol;
}


void test2(real threshold,char * file0,char * file1){
  Graphic g0(file0,false);
  Graphic g1(file1,false);

  Poly * l0=getOutline(g0);
  Poly * l1=getOutline(g1);
  //SDL_Rect r;
  //polygonBBOX(*l1,&r);
  W=640/max(g0.getWidth(),g1.getWidth())/4;
  H=480/max(g0.getHeight(),g1.getHeight())/4;
  if (H<W) W=H;
  else H=W;

  optimize(*l0,threshold);
  optimize(*l1,threshold);
  CHERROR<<"poly: ";
  for(Poly::iterator i=l1->begin();i!=l1->end();i++)
    CHERROR<<"("<<i->x<<" "<<i->y<<") ";
  CHERROR<<ENDL;

  real t0,t1;
  assert(line_line_intersect(Vector2d(0,0),Vector2d(1,0),Vector2d(0,-1),Vector2d(0,1),t0,t1));
  assert(t0==0);
  assert(t1==1);
  assert(line_line_intersect(Vector2d(-0.5,0),Vector2d(1,0),Vector2d(0,-1.5),Vector2d(0,1),t0,t1));
  assert(t0==0.5);
  assert(t1==1.5);

  Vector2d tt;
  assert(line_line_intersect(Vector2d(0,0),Vector2d(1,0),Vector2d(0,0),Vector2d(0,1),tt));
  assert(tt==0);
  assert(line_line_intersect(Vector2d(0,0),Vector2d(1,1),Vector2d(0,1),Vector2d(1,-1),tt));
  assert(tt==0.5);
  assert(circle_intersection_area(Circle(0,1),Circle(0,1))==M_PI);
  assert(circle_intersection_area(Circle(Vector2d(-1,0),sqrt(2)),Circle(Vector2d(1,0),sqrt(2)))==M_PI-2);

  Polygon pol;// (8 1) (3 5) (3 9) (11 26) (17 26) (21 23) (22 19) (14 3)
  pol.push_back(PPoint(Point(8,1),false));
  pol.push_back(PPoint(Point(3,5),false));
  pol.push_back(PPoint(Point(3,9),false));
  pol.push_back(PPoint(Point(11,26),false));
  pol.push_back(PPoint(Point(17,26),false));
  pol.push_back(PPoint(Point(21,23),false));
  pol.push_back(PPoint(Point(22,19),false));
  pol.push_back(PPoint(Point(14,3),false));
/*  ycoord y;
  for(unsigned int i=0;i<pol.size();i++){
    real pini=y(Matrix2x3(),0,pol);
    real pend=y(Matrix2x3(),pol.size()-1,pol);
    assert(find_min_extreme(y,pol,Matrix2x3(),pini,pend)==int(i));
    PPoint pp=pol[pol.size()-1];
    for(unsigned int j=pol.size()-1;j>0;j--) pol[j]=pol[j-1];
    pol[0]=pp;
  }*/
//  assert(interior(Vector2d(3,5),Matrix2x3(),pol));
//  assert(interior(Vector2d(3,6),Matrix2x3(),pol));
//  assert(!interior(Vector2d(3,4),Matrix2x3(),pol));


  //  printf("optimized area: %f\n",polygonArea(*l1));

  list<Poly2*> lpp0;
  breakinconvex(lpp0,*l0);
  list<MKPolygon*> lp0;
  makeMKPolygons(lp0,lpp0);

  list<Poly2*> lpp1;
  breakinconvex(lpp1,*l1);
  list<MKPolygon*> lp1;
  makeMKPolygons(lp1,lpp1);

  Matrix2x3 m(W,H,0,0,0,0,0);
  Point pos1;
  Vector2d pos0((Graphic::getScreen()->w-g0.getWidth())/2/W,(Graphic::getScreen()->h-g0.getHeight())/2/H);

  Point bluelined=Point(Graphic::getScreen()->w/4/W,Graphic::getScreen()->h/4/H);
  Vector2d bluelinep;
  real angle=0;
  Uint32 whitecolor=SDL_MapRGB(Graphic::getScreen()->format,255,255,255);

  UPDATEbegin;
  while(process_events(m,pos1,angle)){
    bluelinep=Vector2d(real(mousePosition.x)/W,real(mousePosition.y)/H);
    Graphic::clearScreen(SDL_MapRGB(Graphic::getScreen()->format,0,0,0));
    drawGrid(m*Matrix2x3(Vector2d(1,0),Vector2d(0,1),pos0),g0);
    drawGrid(m*Matrix2x3(1,1,angle,-g1.getWidth()/2,-g1.getHeight()/2,pos1.x+g1.getWidth()/2,pos1.y+g1.getHeight()/2),g1);

    drawPolygons(m*Matrix2x3(Vector2d(1,0),Vector2d(0,1),pos0),lp0,whitecolor);
    drawPolygons(m*Matrix2x3(1,1,angle,-g1.getWidth()/2,-g1.getHeight()/2,pos1.x+g1.getWidth()/2,pos1.y+g1.getHeight()/2),lp1,whitecolor);

    drawInters(angle,m,bluelinep,bluelined,lp0.front(),lp1.front(),pos0,pos1,g1);

    SDL_Delay(20);
    Graphic::flip(false);
  }
  UPDATEend;

  printf("number of components lp0: %d\n",lp0.size());
  for(list<Poly2*>::iterator p=lpp0.begin();p!=lpp0.end();p++) {
    printf("vertexes: %d\n",(*p)->size());
    delete *p;
  }
  for(list<MKPolygon*>::iterator p=lp0.begin();p!=lp0.end();p++) delete *p;

  printf("number of components lp1: %d\n",lp1.size());
  for(list<Poly2*>::iterator p=lpp1.begin();p!=lpp1.end();p++) {
    printf("vertexes: %d\n",(*p)->size());
    delete *p;
  }
  for(list<MKPolygon*>::iterator p=lp1.begin();p!=lp1.end();p++) delete *p;

  delete l0;
  delete l1;
}

int main(int argc,char**argv) {

  real threshold=0;
  if (argc>1) 
    threshold=strtod(argv[1],NULL);
  if (argc<4) {
    CHERROR<<"USAGE: testoutline n filename filename"<<ENDL;
    exit(1);
  }

  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE)!=0) {
    CHERROR<<SDL_GetError()<<ENDL;
    exit(1);
  }

  ResourceManager rm;

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, 10);

  Graphic::set_video_mode();
  
  test2(threshold,argv[2],argv[3]);

  SDL_Quit();
  return 0;
}
